﻿using System;
using Bestia_Magister.GUI;
using Bestia_Magister.Managers;
using Bestia_Magister.Misc;
using Bestia_Magister.Rotations;
using Bestia_Magister.Settings;
using Styx;
using Styx.CommonBot;
using Styx.CommonBot.Routines;
using Styx.TreeSharp;
using Styx.WoWInternals.DBC;
using Styx.WoWInternals.WoWObjects;
using Action = Styx.TreeSharp.Action;

namespace Bestia_Magister
{
    #region Nested: LocationContextEventArg
    public class WoWContextEventArg : EventArgs
    {
        public readonly WoWContext CurrentWoWContext;
        public readonly WoWContext PreviousWoWContext;

        public WoWContextEventArg(WoWContext currentWoWContext, WoWContext prevWoWContext)
        {
            CurrentWoWContext = currentWoWContext;
            PreviousWoWContext = prevWoWContext;
        }
    }

    #endregion

    public class Main : CombatRoutine
    {
        #region Namings
        private static WoWUnit Pet { get { return StyxWoW.Me.Pet; } }
        private static WoWUnit Focus { get { return StyxWoW.Me.FocusedUnit; } }
        private static WoWUnit Target { get { return StyxWoW.Me.CurrentTarget; } }
        private static LocalPlayer Me { get { return StyxWoW.Me; } }
        private static MainSettings MainSettings { get { return MainSettings.Instance; } }
        #endregion

        #region Delegates
        public delegate T Selection<out T>(object context);
        public delegate WoWPoint LocationRetriever(object context);
        public delegate WoWUnit UnitSelectionDelegate(object context);
        public delegate bool SimpleBooleanDelegate(object context);
        public delegate string SimpleStringDelegate(object context);
        public delegate int SimpleIntDelegate(object context);
        #endregion

        #region Overrides : Combat Routine

        public Main()
        {
            Instance = this;
            BotEvents.OnBotStarted += BotEvents_OnBotStarted;
            BotEvents.Player.OnMapChanged += e =>
            {
                // Don't run this handler if we're not the current routine!
                if (RoutineManager.Current.Name != Name)
                    return;

                //Set Rotation
                RotationSelector.RebuildBehaviors();
                // Only ever update the context. All our internal handlers will use the context changed event
                // so we're not reliant on anything outside of ourselves for updates.
                UpdateContext();
            };
        }

        public static Main Instance { get; private set; }

        /// <summary>
        /// Version information
        /// </summary>
        internal static readonly Version Version = new Version(2, 0, 0);

        /// <summary>
        /// Returns the Combat Routine's name and version
        /// </summary>
        public override string Name
        {
            get { return "The Mark I " + Version; }
        }

        /// <summary>
        /// Combat Routines Class
        /// </summary>
        public override WoWClass Class
        {
            get { return WoWClass.Hunter; }
        }

        /// <summary>
        /// Do we want a GUI or not
        /// </summary>
        public override bool WantButton
        {
            get { return true; }
        }

        /// <summary>
        /// Opens up the GUI on button press
        /// </summary>
        public override void OnButtonPress()
        {
            var gui = new ConfigurationForm();
            var guiSV = new ConfigurationFormSV();
            switch (Me.Specialization)
            {
                case WoWSpec.HunterSurvival:
                    guiSV.ShowDialog();
                    break;
                case WoWSpec.HunterBeastMastery:
                    gui.ShowDialog();
                    break;
            }
        }

        /// <summary>
        /// Called when this CC is selected as the current CC.
        /// </summary>
        public override void Initialize()
        {
            base.Initialize();

            Logger.DebugLog("Character Level: {0}", Me.Level);
            Logger.DebugLog("Character Faction: {0}", Me.IsAlliance ? "Alliance" : "Horde");
            Logger.DebugLog("Character Race: {0}", Me.Race);
            Logger.DebugLog("Character Mapname: {0}", Me.MapName);
            Logger.InitLog("LocatonContext: {0}", CurrentWoWContext.ToString());
            Logger.InitLog("Thankyou for choosing Mark I as your Combat Routine");
            Logger.InitLog("Current BotBase is {0}", BotManager.Current.Name);
            Logger.InitLog("Max DPS Setup:");
            Logger.CombatLog("Botbase: Tyrael");
            Logger.CombatLog("TPS: 40");
            Logger.CombatLog("Framelock: On");
            Logger.CombatLog("Plugins: Off");
            Logger.InitLog("Current Routine Version:");
            Logger.CombatLog("- Mark I ~ Hunter v" + Version); // Developer
            Logger.CombatLog("- A Hunter Routine");
            Logger.CombatLog("- by Falldown & Mirabis");
            Logger.InitLog("Your Latency: " + Helpers.MyLatency + " ms");
          //  Logger.CombatLog("Remember to set your Custom Lag Tolerance to: 400ms"); BUG: Not needed
            //Logger.CombatLog(" - Accessed in: Game Menu -> Interface -> Combat"); BUG: Not needed
          //  Logger.CombatLog("");
            Logger.DebugLog("Press [F9] to calculate Tree Performance (New)");
            Logger.DebugLog("Press [F6] to show current DPS");
            //Logger.CombatLog("");
            try
            {
                //Get Talents
                TalentManager.Update();
                //Talents for Debugging
                TalentManager.Dumpglyphs();
                //Set Rotation
                RotationSelector.RebuildBehaviors();
                
                // NOTE: Hook these events AFTER the context update.
                OnWoWContextChanged += (orig, ne) =>  Logger.DebugLog(" Context changed, switching rotation to: {0}", CurrentWoWContext.ToString());
            }
            catch (Exception e)
            {
                 Logger.DebugLog(e.ToString());
                BotManager.Current.Stop();
            }

        }

        #region BotEvents_OnBotStarted
        private void BotEvents_OnBotStarted(EventArgs args)
        {
            if (RoutineManager.Current.Name != Name) return;
            switch (BotManager.Current.Name)
            {
                case "Tyrael":
                case "LazyRaider":
                case "Raid Bot":
                case "CombatBot":
                case "Tyrael - BE":
                    break;
                default:
                    TreeRoot.Stop(string.Format("Wrong Botbase! {0} is not supported", BotManager.Current.Name));
                    return;
            }
            DpsMeter.Initialize();
        }

        public override void ShutDown()
        {
            DpsMeter.Shutdown();
            HotkeyManager.RemoveAllKeys();
        }

        #endregion

        #region Pulse
        public override void Pulse()
        { // No pulsing if we're loading or out of the game.
            if (!StyxWoW.IsInGame || !StyxWoW.IsInWorld || !StyxWoW.Me.IsValid)
                return;
            UpdateContext();
            // Double cast shit
            CastManager.PulseDoubleCastEntries();
            /*//Register keys uhm why pulse BUG: Already checked at start of rotation
            if (!HotkeyManager.Registered)
            {
                HotkeyManager.RegisterKeys(); //Hotkeys Registered
            }*/
            PetManager.Pulse();

            if (StyxWoW.Me.CurrentTarget != null)
                DpsMeter.Update();
        }


        #endregion
        #endregion

        #region Overrides : Behavior
        public override Composite CombatBehavior
        {
            get { return RotationSelector.CreateCombat; } //BUG: ...was here.. no point making all static but it is now... ileave it 
        }

        public override Composite RestBehavior
        {
            get { return RotationSelector.CreatePreCombat; }
        }
        #endregion

        #region Combat Selector
        /*
        private static Composite CombatSelector()
        {
            return new PrioritySelector(  
   
            Helpers.TreePerformance("CombatSelector"),

            new Decorator(ret => !HotkeyManager.Registered, new Action(ret => HotkeyManager.RegisterKeys())), //Hotkeys Registered

                new Decorator(ret => Me.CurrentMap.IsArena,
                new Throttle(1, TimeSpan.FromSeconds(1), Helpers.AcquireTargets)),

                new Decorator(ret => MainSettings.DisablePetActions && Me.Specialization == WoWSpec.HunterBeastMastery,
                new Action(delegate
                {
                    if (MainSettings.GrowlPets)
                    {
                        Lua.DoString("RunMacroText('/petautocastoff Growl');");
                    }
                    if (PetSettings.SpiritMend != SpiritMend.Never)
                    {
                        Lua.DoString("RunMacroText('/petautocastoff Spirit Mend');");
                    }
                    if (PetSettings.RoarOfSacrifice != RoarOfSacrifice.Never)
                    {
                        Lua.DoString("RunMacroText('/petautocastoff Roar of Sacrifice');");
                    }
                    return RunStatus.Failure;
                })),

                PetManager.CreateHunterCallPetBehavior(),
                Helpers.UseItems,  // Trinkets, Potions, etc 

        new Decorator(ret => Me.Specialization == WoWSpec.HunterBeastMastery,
            new PrioritySelector(
                HandleDefensiveCooldowns(),  // Heals, Defensive Buffs, etc 
                HandleCrowdControl(), // Freeze trap, etc
                HandleHotkeyCCs(), // Hotkey Crowd Controlling Rotation

                Hold(),

        new Decorator(ret => ((!Helpers.IsKeyDown(MainSettings.Hotkeys.Instance.CCModifierChoice) || MainSettings.Hotkeys.Instance.CCModifierChoice == Keys.None) && !Helpers.IsKeyDown(MainSettings.Hotkeys.Instance.CCModeKeyChoice)) &&
            (Me.CurrentPendingCursorSpell == null || CurrentWoWContext == WoWContext.PvE) && (Target == null || !Target.HasAura("Scatter Shot")) && ((Unit.ValidFocus(true, false) && !Focus.HasAura("Scatter Shot")) || !Unit.ValidFocus(true, false) || Focus.HasAura("Freezing Trap") || Helpers.GetSpellCooldown("Freezing Trap").TotalSeconds > 1),
            new PrioritySelector(
                // Attack with pet
                new Decorator(ret => PetSettings.PetAttack && Unit.ValidUnit(Target, true, false) && !Helpers.SelfControl(Target) && Me.GotAlivePet && (Pet.CurrentTarget == null || Pet.CurrentTarget != Me.CurrentTarget),
                    new Action(delegate
                    {
                        PetManager.CastPetAction("Attack");
                        return RunStatus.Failure;
                    })),

                HandleInterrupts(), // Silence, Stuns, etc 

                new Decorator(ret => HotkeyManager.RegenMode,
                HandleFocusRegen()), // Focus Saving Rotation

        new Decorator(ret => !HotkeyManager.RegenMode,
            new PrioritySelector(
                HandleOffensiveCooldowns(), // Damage Increasing Buffs, Aspect of the Hawk
                HandleTraps(),// All traps
                new Decorator(ret => Helpers.AddCount(25) >= MainSettings.AddCount || HotkeyManager.AoEMode,
                HandleAoE()), // Area of Effect Rotation
                HandleSingleTarget()) // Single Target Rotation  
            ))
        ))
        ),

        new Decorator(ret => Me.Specialization == WoWSpec.HunterSurvival || HotkeyManager.AoEMode,
            new PrioritySelector(
                HandleDefensiveCooldownsSV(),  // Heals, Defensive Buffs, etc 
                HandleCrowdControlSV(), // Control adds etc,

                // Attack with pet
                new Decorator(ret => PetSettingsSV.PetAttack && Unit.ValidUnit(Target, true, false) && !Helpers.SelfControl(Target) && Me.GotAlivePet && (Pet.CurrentTarget == null || Pet.CurrentTarget != Me.CurrentTarget),
                    new Action(delegate
                    {
                        PetManager.CastPetAction("Attack");
                        return RunStatus.Failure;
                    })),

                HandleInterruptsSV(), // Silence, Stuns, etc 

                new Decorator(ret => HotkeyManager.RegenMode,
                HandleFocusRegen()), // Focus Saving Rotation
        new Decorator(ret => !HotkeyManager.RegenMode,
            new PrioritySelector(
                HandleOffensiveCooldownsSV(), // Damage Increasing Buffs, Aspect of the Hawk
                HandleTrapsSV(),// All traps
                new Decorator(ret => HotkeyManager.AoEMode || Helpers.AddCount(25) >= MainSettingsSV.AddCount,
                HandleAoESV()), // Area of Effect Rotation
                HandleSingleTargetSV()))

            ) // Single Target Rotation  
        )  
    );

        }
        */
        #endregion

        #region Hold
        /// <summary>
        /// Holds the rotation for custom actions
        /// </summary>
        /// <returns>RunStatus.Success</returns>
        
        public static Composite Hold()
        {
            return
                new Decorator(
                    ret => Me.HasAnyCachedAura(new[] {"Drink", "Food", "First Aid", "Feign Death"}) || !Unit.ValidUnit(Target, true, true) || Lists.ShouldHold(Target),
                    new Action(delegate { return RunStatus.Success; }));
        }

        #endregion

        /* Old Behavior
      
        #region HandleDefensiveCooldowns

        #region Beast Mastery
        private static Composite HandleDefensiveCooldowns()
        {
            return new PrioritySelector(

            #region PvE
    new Decorator(ret => CurrentWoWContext == WoWContext.PvE,
        new PrioritySelector(
            Spell.BuffSelf("Exhilaration", ret => MainSettings.Talent45 && TalentManager.HasTalent(7) && NeedExhilaration()),
            Spell.BuffSelf("Feign Death", ret => ShouldFeignDeath()),
            //Spell.BuffSelf("Detterence", ret => ShouldDetterence() ),
            Spell.BuffSelf("Master's Call", ret => MainSettings.MastersCallPvE && Me.GotAlivePet && Lists.Snared(Me) >= 250 || Lists.Rooted(Me) >= 250),
            PetManager.CreateCastPetActionOn("Spirit Mend", ret => Me, ret => Me.HealthPercent < PetSettings.SpiritHealMe && (PetSettings.SpiritMend == SpiritMend.Me || PetSettings.SpiritMend == SpiritMend.Me_Focus || PetSettings.SpiritMend == SpiritMend.Me_Focus_Pet)),
            PetManager.CreateCastPetActionOn("Spirit Mend", ret => Focus, ret => Unit.ValidFocus(false, true) && Focus.HealthPercent < PetSettings.SpiritHealFocus && (PetSettings.SpiritMend == SpiritMend.Focus || PetSettings.SpiritMend == SpiritMend.Me_Focus || PetSettings.SpiritMend == SpiritMend.Me_Focus_Pet)),
            PetManager.CreateCastPetActionOn("Spirit Mend", ret => Pet, ret => Pet.HealthPercent < PetSettings.SpiritHealPet && (PetSettings.SpiritMend == SpiritMend.Pet || PetSettings.SpiritMend == SpiritMend.Me_Pet || PetSettings.SpiritMend == SpiritMend.Me_Focus_Pet)),
            Spell.BuffSelf("Mend Pet", ret => PetSettings.MendPet && Me.GotAlivePet && Pet.HealthPercent <= PetSettings.PetHealth && !Pet.HasMyAura("Mend Pet")),
            PetManager.CreateCastPetAction("Heart of the Phoenix", ret => PetSettings.RevivePet && (Me.Pet == null || (Me.Pet != null && !Pet.IsAlive))),
            Spell.BuffSelf("Will of The Forsaken", ret => Helpers.isForsaken(Me).TotalSeconds > 1),
            Spell.BuffSelf("Stoneform", ret => CooldownSettings.Racial && Me.Race == WoWRace.Dwarf && StyxWoW.Me.GetAllAuras().Any(a => a.Spell.Mechanic == WoWSpellMechanic.Bleeding || a.Spell.DispelType == WoWDispelType.Disease || a.Spell.DispelType == WoWDispelType.Poison)) // Defensive cooldown
            )),
            #endregion

            #region PvP
            new Decorator(ret => CurrentWoWContext == WoWContext.PvP,
                new PrioritySelector(
                    Spell.BuffSelf("Gift of the Naaru", ret => CooldownSettings.Racial && Me.Race == WoWRace.Draenei && Me.HealthPercent < 68), // Dranei racial Heal
                    Spell.BuffSelf("Exhilaration", ret => TalentManager.HasTalent(7) && MainSettings.Talent45 && (Me.HealthPercent < 67 || (Me.GotAlivePet && Pet.HealthPercent < 15))),
                    Spell.BuffSelf("Mend Pet", ret => PetSettings.MendPet && Me.GotAlivePet && !Pet.HasAura("Mend Pet") && Pet.HealthPercent < PetSettings.PetHealth),
                    Spell.BuffSelf("Master's Call", ret => Me.GotAlivePet && (MainSettings.MastersCall == MastersCall.Me || MainSettings.MastersCall == MastersCall.Me_Partner) && Lists.Snared(Me) >= 250 || Lists.Rooted(Me) >= 250),
                    Spell.BuffSelf("Deterrence", ret => MainSettings.DeterrenceReflect && TalentManager.HasGlyph("Mirrored Blades") && Helpers.UnitsCastingReflectableSpell()),
                    Spell.Buff("Master's Call", ret => Helpers.ArenaPartner, ret => Me.CurrentMap.IsArena && Helpers.ArenaPartner != null && Me.GotAlivePet && (MainSettings.MastersCall == MastersCall.Partner || MainSettings.MastersCall == MastersCall.Me_Partner) && Lists.Snared(Helpers.ArenaPartner) >= 250 || Lists.Rooted(Helpers.ArenaPartner) >= 250),
                    PetManager.CreateCastPetActionOn("Roar of Sacrifice", ret => Me, ret => (Me.HealthPercent < PetSettings.RoSMe || Lists.CCd(Me) > 500) && (PetSettings.RoarOfSacrifice == RoarOfSacrifice.Me || PetSettings.RoarOfSacrifice == RoarOfSacrifice.Me_Partner)),
                    PetManager.CreateCastPetActionOn("Roar of Sacrifice", ret => Helpers.ArenaPartner, ret => Me.CurrentMap.IsArena && Helpers.ArenaPartner != null && (Helpers.ArenaPartner.HealthPercent < PetSettings.RoSPartner || Lists.CCd(Helpers.ArenaPartner) > 500) && (PetSettings.RoarOfSacrifice == RoarOfSacrifice.Partner || PetSettings.RoarOfSacrifice == RoarOfSacrifice.Me_Partner)),
                    PetManager.CreateCastPetActionOn("Spirit Mend", ret => Me, ret => Me.HealthPercent < PetSettings.SpiritHealMe && (PetSettings.SpiritMend == SpiritMend.Me || PetSettings.SpiritMend == SpiritMend.Me_Focus || PetSettings.SpiritMend == SpiritMend.Me_Focus_Pet)),
                    PetManager.CreateCastPetActionOn("Spirit Mend", ret => Focus, ret => Unit.ValidFocus(false, true) && Focus.HealthPercent < PetSettings.SpiritHealFocus && (PetSettings.SpiritMend == SpiritMend.Focus || PetSettings.SpiritMend == SpiritMend.Me_Focus || PetSettings.SpiritMend == SpiritMend.Me_Focus_Pet)),
                    PetManager.CreateCastPetActionOn("Spirit Mend", ret => Pet, ret => Pet.HealthPercent < PetSettings.SpiritHealPet && (PetSettings.SpiritMend == SpiritMend.Pet || PetSettings.SpiritMend == SpiritMend.Me_Pet || PetSettings.SpiritMend == SpiritMend.Me_Focus_Pet)),
                    PetManager.CreateCastPetActionOn("Spirit Mend", ret => Helpers.ArenaPartner, ret => Me.CurrentMap.IsArena && Helpers.ArenaPartner != null && Helpers.ArenaPartner.HealthPercent < PetSettings.SpiritHealPartner && (PetSettings.SpiritMend == SpiritMend.Partner || PetSettings.SpiritMend == SpiritMend.Me_Partner || PetSettings.SpiritMend == SpiritMend.Pet_Partner || PetSettings.SpiritMend == SpiritMend.Me_Pet_Partner)),
                    Spell.BuffSelf("Feign Death", ret => ShouldFeignDeathPvP()),
                    Spell.BuffSelf("Will of The Forsaken", ret => Me.Race == WoWRace.Undead && CooldownSettings.Racial && Helpers.isForsaken(Me).TotalSeconds > 1),
                    Spell.BuffSelf("Stoneform", ret => CooldownSettings.Racial && Me.Race == WoWRace.Dwarf && StyxWoW.Me.GetAllAuras().Any(a => a.Spell.Mechanic == WoWSpellMechanic.Bleeding || a.Spell.DispelType == WoWDispelType.Disease || a.Spell.DispelType == WoWDispelType.Poison)) // Defensive cooldown

            #endregion

            )));
        }
        #endregion

        #region Survival
        private static Composite HandleDefensiveCooldownsSV()
        {
            return new PrioritySelector(
            #region PvE
            new Decorator(ret => CurrentWoWContext == WoWContext.PvE,
                new PrioritySelector(
                    Spell.BuffSelf("Exhilaration", ret => TalentManager.HasTalent(7) && MainSettingsSV.Talent45 && NeedExhilaration()),
                    Spell.BuffSelf("Feign Death", ret => ShouldFeignDeath()),
                    Spell.BuffSelf("Master's Call", ret => MainSettingsSV.MastersCallPvE && Me.GotAlivePet && Lists.Snared(Me) >= 250 || Lists.Rooted(Me) >= 250),
                //Spell.BuffSelf("Detterence", ret => ShouldDetterence() ),
                    Spell.BuffSelf("Mend Pet", ret => PetSettingsSV.MendPet && Me.GotAlivePet && Pet.HealthPercent <= PetSettingsSV.PetHealth && !Pet.HasMyAura("Mend Pet")),
                    PetManager.CreateCastPetAction("Heart of the Phoenix", ret => PetSettingsSV.RevivePet && (Me.Pet == null || (Me.Pet != null && !Pet.IsAlive))),
                    Spell.BuffSelf("Will of The Forsaken", ret => Me.Race == WoWRace.Undead && CooldownSettingsSV.Racial && Helpers.isForsaken(Me).TotalSeconds > 1),
                    Spell.BuffSelf("Stoneform", ret => CooldownSettingsSV.Racial && Me.Race == WoWRace.Dwarf && StyxWoW.Me.GetAllAuras().Any(a => a.Spell.Mechanic == WoWSpellMechanic.Bleeding || a.Spell.DispelType == WoWDispelType.Disease || a.Spell.DispelType == WoWDispelType.Poison)) // Defensive cooldown
                    )));
            #endregion 
        }
        #endregion

        #endregion

        #region Hotkey CC's

        #region Beast Mastery
        private static Composite HandleHotkeyCCs()
        {
            return new PrioritySelector(

            #region PVE
new Decorator(ret => CurrentWoWContext == WoWContext.PvE,
                new PrioritySelector(

                )),

            #endregion

            #region PvP
 new Decorator(ret => CurrentWoWContext == WoWContext.PvP && (Helpers.IsKeyDown(MainSettings.Hotkeys.Instance.CCModifierChoice) || MainSettings.Hotkeys.Instance.CCModifierChoice == Keys.None) && Helpers.IsKeyDown(MainSettings.Hotkeys.Instance.CCModeKeyChoice),
                new PrioritySelector(      
                    //Focused Unit
                new Decorator(ret => Unit.ValidFocus(true, false),
                    new PrioritySelector(
                    Spell.CreateTrapBehavior("Freezing Trap", true, ret => Focus, ret => CanBeTrapped(Focus) && (SafeToTrap(Focus) || Helpers.IsKeyDown(MainSettings.Hotkeys.Instance.ForceTrapChoice)) && Focus.GetAuraTimeLeft("Freezing Trap", true).TotalSeconds <= 1),
                    Spell.Buff("Wyvern Sting", ret => Focus, ret => TalentManager.HasTalent(5) && !Focus.HasAura("Freezing Trap") && Helpers.GetSpellCooldown("Freezing Trap").TotalSeconds > 5),
                    Spell.Cast("Scatter Shot", ret => Focus, ret => !Helpers.SelfControl(Focus) && Lists.CCd(Focus) < 500 && Lists.Rooted(Focus) < 500),
                    PetManager.CreateCastPetActionOn("Lullaby", ret => Focus, ret => PetSettings.Lullaby && !Focus.HasAnyAura("Lullaby", "Scatter Shot") && ((Helpers.GetSpellCooldown("Scatter Shot").TotalSeconds < 27 && Helpers.GetSpellCooldown("Scatter Shot").TotalSeconds > 1)) && Pet.Location.Distance(Focus.Location) <= 20 && Lists.CCd(Focus) < 500 && Lists.Rooted(Focus) < 500)

                    )),
                    //Current Target
            new Decorator(ret => Unit.ValidUnit(Target, true, false) && !Unit.ValidFocus(true, false),
                new PrioritySelector(
                    Spell.CreateTrapBehavior("Freezing Trap", true, ret => Target, ret => CanBeTrapped(Target) && (SafeToTrap(Target) || Helpers.IsKeyDown(MainSettings.Hotkeys.Instance.ForceTrapChoice)) && Target.GetAuraTimeLeft("Freezing Trap", true).TotalSeconds <= 1),
                    Spell.Buff("Wyvern Sting", ret => Target, ret => TalentManager.HasTalent(5) && !Target.HasAura("Freezing Trap") && Helpers.GetSpellCooldown("Freezing Trap").TotalSeconds > 5),
                    Spell.Cast("Scatter Shot", ret => Target, ret => !Helpers.SelfControl(Target) && Lists.CCd(Target) < 500 && Lists.Rooted(Target) < 500),
                    PetManager.CreateCastPetActionOn("Lullaby", ret => Target, ret => PetSettings.Lullaby && !Target.HasAnyAura("Lullaby", "Scatter Shot") && ((Helpers.GetSpellCooldown("Scatter Shot").TotalSeconds < 27 && Helpers.GetSpellCooldown("Scatter Shot").TotalSeconds > 1)) && Pet.Location.Distance(Target.Location) <= 20 && Lists.CCd(Target) < 500 && Lists.Rooted(Target) < 500)
                    ))
                ))
                );
            #endregion
        }
        #endregion

        #region Survival
        private static Composite HandleHotKeyCCsSV()
        {
            return new PrioritySelector(

            #region PVE
new Decorator(ret => CurrentWoWContext == WoWContext.PvE,
                new PrioritySelector(
            )));
            #endregion
        }

        #endregion

        #endregion

        #region HandleCrowdControl

        #region Beast Mastery
        private static Composite HandleCrowdControl()
        {
            return new PrioritySelector(

            #region PVE
        new Decorator(ret => CurrentWoWContext == WoWContext.PvE && Unit.ValidUnit(Target, true, false),
                new PrioritySelector(   
                    Spell.CastOnGround("Binding Shot", ret => Target.Location, ret => MainSettings.Talent30_BS && Helpers.AddCount(15) >= MainSettings.BindingCount),
                    Spell.Buff("Distracting Shot",ret => Target, ret => MainSettings.DistractingShot && Me.GotAlivePet && TalentManager.HasGlyph("Distracting Shot") && Target.CurrentTargetGuid == Me.Guid),
                    Spell.Cast("Blink Strike", ret => MainSettings.Talent75_BS && Me.GotAlivePet && Pet.Location.Distance(Target.Location) > 25 && Pet.InLoSSS(Target)),
                    PetManager.CreateCastPetActionOn("Lullaby", ret => Focus, ret => PetSettings.Lullaby && Unit.ValidFocus(true, false) && !Focus.HasAnyAura("Lullaby", "Scatter Shot") && Pet.Location.Distance(Focus.Location) <= 20 && Lists.CCd(Focus) < 250),
                    Spell.Cast("Misdirection", ret => MDTarget(), ret => !Me.HasAura("Misdirection") && ShouldMisdirect())
                )),

            #endregion

            #region PvP
            new Decorator(ret => CurrentWoWContext == WoWContext.PvP,
                new PrioritySelector(
                    //Current Target
                    new Decorator(ret => Unit.ValidUnit(Target, true, false) && Me.CurrentPendingCursorSpell == null,
                        new PrioritySelector(
                    Spell.Cast("Tranquilizing Shot", ret => Target, ret => MainSettings.Tranquilizing && Lists.CanBeTranquilized(Target, 2) && (Helpers.PlayerPower > 60 || Me.HasAnyAura("The Beast Within", "Thrill of the Hunt") || TalentManager.HasGlyph("Tranquilizing Shot")) && Target.Distance <= 40),
                    PetManager.CreateCastPetActionOn("Growl", ret => Helpers.EnemyPet, ret => MainSettings.GrowlPets && Helpers.EnemyPet != null),
                    Spell.Cast("Arcane Shot", ret => Helpers.ShamanTotem, ret => Me.CurrentMap.IsArena && Helpers.ShamanTotem != null && MainSettings.KillTotems)
                    )),
                    

                    //Focused Unit
                    new Decorator( ret => Unit.ValidFocus(true, false),
                        new PrioritySelector(

                   new Decorator(ret => MainSettings.CCFreezingTrap && CanBeTrapped(Focus) && Helpers.GetSpellCooldown("Freezing Trap").TotalSeconds < 1,
                    Spell.CreateTrapBehavior("Freezing Trap", true, ret => Focus, ret => SafeToTrap(Focus) && Focus.GetAuraTimeLeft("Freezing Trap", true).TotalSeconds <= 1)),
    
                    Spell.Buff("Wyvern Sting", ret => Focus, ret => MainSettings.CCWyvernSting && TalentManager.HasTalent(5) && !Focus.HasAura("Freezing Trap")),
                    Spell.Cast("Scatter Shot", ret => Focus, ret => MainSettings.FocusScatter && !Helpers.SelfControl(Focus) && Lists.CCd(Focus) < 500 && Lists.Rooted(Focus) < 500),
                    PetManager.CreateCastPetActionOn("Lullaby", ret => Focus, ret => PetSettings.Lullaby && !Focus.HasAnyAura("Lullaby", "Scatter Shot") && ((Helpers.GetSpellCooldown("Scatter Shot").TotalSeconds < 27 && Helpers.GetSpellCooldown("Scatter Shot").TotalSeconds > 1) || !MainSettings.FocusScatter) && Pet.Location.Distance(Focus.Location) <= 20 && Lists.CCd(Focus) < 500),
                    Spell.Cast("Tranquilizing Shot", ret => Focus, ret => MainSettings.FocusTranq && Lists.CanBeTranquilized(Focus, 2) && (Helpers.PlayerPower > 60 || Me.HasAnyAura("The Beast Within", "Thrill of the Hunt") || TalentManager.HasGlyph("Tranquilizing Shot")) && Focus.Distance <= 40 && Me.CurrentPendingCursorSpell == null)

                    ))))
                );
            #endregion
        }
        #endregion

        #region Survival
        private static Composite HandleCrowdControlSV()
        {
            return new PrioritySelector(

            #region PVE
            new Decorator(ret => CurrentWoWContext == WoWContext.PvE,
                new PrioritySelector(
                    Spell.CastOnGround("Binding Shot", ret => Target.Location, ret => MainSettingsSV.Talent30_BS && Helpers.AddCount(15) >= MainSettingsSV.BindingCount),
                    Spell.Buff("Distracting Shot", ret => Target, ret => MainSettingsSV.DistractingShot && Me.GotAlivePet && TalentManager.HasGlyph("Distracting Shot") && Target.CurrentTargetGuid == Me.Guid),
                    Spell.Cast("Blink Strike", ret => MainSettingsSV.Talent75_BS && Me.GotAlivePet && Pet.Location.Distance(Target.Location) > 25 && Pet.InLoSSS(Target)),
                    PetManager.CreateCastPetActionOn("Lullaby", ret => Focus, ret => PetSettingsSV.Lullaby && Unit.ValidFocus(true, false) && !Focus.HasAnyAura("Lullaby", "Scatter Shot") && Pet.Location.Distance(Focus.Location) <= 20 && Lists.CCd(Focus) < 250),
                    Spell.Cast("Misdirection", ret => MDTarget(), ret => !Me.HasAura("Misdirection") && ShouldMisdirect())
            )));
            #endregion
        }

        #endregion

        #endregion

        #region HandleInterrupts

        #region Beast Mastery
        private static Composite HandleInterrupts()
        {
            return new PrioritySelector(

            #region PvP
       new Decorator(ret => CurrentWoWContext == WoWContext.PvP,
            new PrioritySelector( 
                new Decorator(ret => Unit.ValidFocus(true, false),
                    new PrioritySelector(
                        Spell.Cast("Silencing Shot", ret => Focus, ret => TalentManager.HasTalent(4) && MainSettings.FocusSilencingShot && ShouldSilencePvP(Focus) && Focus.HealthPercent < MainSettings.SilenceHealth && Lists.BreakOnDamage(Focus) < 250 && Lists.Silenced(Focus) < 250 && Lists.CCd(Focus) < 250),
                        Spell.Cast("Wyvern Sting", ret => Focus, ret => MainSettings.CCWyvernSting && !Focus.HasAura("Freezing Trap") && Focus.Distance <= 35),
                        Spell.Cast("Tranquilizing Shot", ret => Focus, ret => MainSettings.FocusTranq && Lists.CanBeTranquilized(Focus,2) && (Helpers.PlayerPower > 60 || Me.HasAura("The Beast Within") && Focus.Distance <= 40 )),
                        Spell.Buff("Concussive Shot", ret => Focus, ret => MainSettings.FocusConcussive && Lists.Snared(Focus) <= 500 && Focus.Distance <= 40 && !Helpers.SpellOnGCD("Kill Command") && Me.CurrentPendingCursorSpell == null && Lists.BreakOnDamage(Focus) < 250),
                        Spell.BuffSelf("War Stomp", ret => CooldownSettings.Racial && Me.Race == WoWRace.Tauren && Focus.Distance <= 8 && Focus.IsCasting && Focus.CurrentCastTimeLeft.TotalMilliseconds > 850 && Lists.CCd(Focus) < 750))),
               new Decorator(ret => Unit.ValidUnit(Target, true, false),
                    new PrioritySelector(
                        Spell.Cast("Tranquilizing Shot", ret => Target, ret => MainSettings.Tranquilizing && Lists.CanBeTranquilized(Target, 2) && (Helpers.PlayerPower > 60 || Me.HasAura("The Beast Within") || Target.HasAura("Freezing Trap")) && Target.Distance <= 40),
                        Spell.Cast("Intimidation", ret => ShouldIntimidate() && Lists.BreakOnDamage(Target) < 250 && Lists.CCd(Target) < 500),
                        Spell.Cast("Silencing Shot", ret => Target, ret => TalentManager.HasTalent(4) && MainSettings.Talent30_SS && ShouldSilencePvP(Target) && Target.HealthPercent < MainSettings.SilenceHealth && Lists.BreakOnDamage(Target) < 250 && Lists.Silenced(Target) < 250),
                        Spell.Buff("Concussive Shot", ret => Target, ret => MainSettings.ConcussiveShot && Lists.BreakOnDamage(Target) < 250 && Lists.Snared(Target) <= 0 && Target.Distance <= 40 && !Helpers.SpellOnGCD("Kill Command")),
                        Spell.BuffSelf("War Stomp", ret => CooldownSettings.Racial && Me.Race == WoWRace.Tauren && Target.Distance <= 8 && Lists.BreakOnDamage(Target) < 250 && Lists.CCd(Target) < 750 && Target.IsCasting && Target.CurrentCastTimeLeft.TotalMilliseconds > 850),
                        Spell.Cast("Arcane Torrent", ret => Target, ret => CooldownSettings.Racial && Me.Race == WoWRace.BloodElf && Lists.BreakOnDamage(Target) < 250 && Lists.Silenced(Target) < 250 && Me.CurrentTarget.IsPlayer && (Lists.InterruptCastNoChannel(Target) > 200 || Lists.InterruptCastChannel(Target) > 200))))
                    )),
            #endregion

            #region PvE
        new Decorator(ret => CurrentWoWContext == WoWContext.PvE && Unit.ValidUnit(Target,true,false),
            new PrioritySelector(
                Spell.Cast("Scatter Shot", ret => ShouldScatterShot()),
                Spell.Cast("Silencing Shot", ret => TalentManager.HasTalent(4) && MainSettings.Talent30_SS && (Lists.InterruptCastNoChannel(Target) > 200 || Lists.InterruptCastChannel(Target) > 200)),
                Spell.Cast("Wyvern Sting", ret => TalentManager.HasTalent(5) && MainSettings.Talent30_WS && (Lists.InterruptCastNoChannel(Target) > 200 || Lists.InterruptCastChannel(Target) > 200)),
                Spell.Buff("Concussive Shot", ret => Target, ret => MainSettings.ConcussiveShot && Lists.Snared(Target) <= 0 && !Lists.TargetIsBoss(Target) && Target.Distance <= 40),
                Spell.Cast("Intimidation", ret => ShouldIntimidate() && Me.Specialization == WoWSpec.HunterBeastMastery),
                PetManager.CreateCastPetActionOn("Lullaby", ret => Target, ret => PetSettings.LullabyInt && !Target.HasAnyAura("Lullaby", "Scatter Shot") 
                    && ((Helpers.GetSpellCooldown("Scatter Shot").TotalSeconds < 27 && Helpers.GetSpellCooldown("Scatter Shot").TotalSeconds > 1) 
                    || (MainSettings.ScatterShot == ScatterShot.Never || MainSettings.ScatterShot == ScatterShot.Defend))
                    && Pet.Location.Distance(Target.Location) <= 20 && (Lists.InterruptCastNoChannel(Target) > 200 || Lists.InterruptCastChannel(Target) > 200))
            ))
            #endregion
         );
        }
        #endregion

        #region Survival
        private static Composite HandleInterruptsSV()
        {
        #region PvE
        return new Decorator(ret => CurrentWoWContext == WoWContext.PvE && Unit.ValidUnit(Target, true, false),
            new PrioritySelector(
                Spell.Cast("Scatter Shot", ret => ShouldScatterShot()),
                Spell.Cast("Silencing Shot", ret => TalentManager.HasTalent(4) && MainSettingsSV.Talent30_SS && (Lists.InterruptCastNoChannel(Target) > 200 || Lists.InterruptCastChannel(Target) > 200)),
                Spell.Cast("Wyvern Sting", ret => TalentManager.HasTalent(5) && MainSettingsSV.Talent30_WS && (Lists.InterruptCastNoChannel(Target) > 200 || Lists.InterruptCastChannel(Target) > 200)),
                Spell.Buff("Concussive Shot", ret => Target, ret => MainSettingsSV.ConcussiveShot && Lists.Snared(Target) <= 0 && !Lists.TargetIsBoss(Target) && Target.Distance <= 40),
                PetManager.CreateCastPetActionOn("Lullaby", ret => Target, ret => PetSettingsSV.LullabyInt && !Target.HasAnyAura("Lullaby", "Scatter Shot")
                    && ((Helpers.GetSpellCooldown("Scatter Shot").TotalSeconds < 27 && Helpers.GetSpellCooldown("Scatter Shot").TotalSeconds > 1)
                    || (MainSettingsSV.ScatterShot == ScatterShot.Never || MainSettingsSV.ScatterShot == ScatterShot.Defend))
                    && Pet.Location.Distance(Target.Location) <= 20 && (Lists.InterruptCastNoChannel(Target) > 200 || Lists.InterruptCastChannel(Target) > 200))
            ));
        #endregion
        }
        #endregion

        #endregion

        #region Focus Regen

        #region Beast Mastery
        private static Composite HandleFocusRegen()
        {
            return new PrioritySelector(
            #region PvP
new Decorator(ret => CurrentWoWContext == WoWContext.PvP && Unit.ValidUnit(Target, true, false) && Lists.BreakOnDamage(Target) < 250,
                new PrioritySelector(
            #region Hold
                //Wait out GCD, do not traverse if can't cast
                     Helpers.WaitForCastOrChannel(),
                     new Decorator(ret => !Helpers.IsGlobalCooldown(405),
                        new PrioritySelector(
            #endregion
                        Spell.CancelCast("Cobra Shot", ret => Helpers.PlayerPower > 100 & Me.CurrentCastTimeLeft.TotalMilliseconds > 500, RunStatus.Failure),
                        Spell.Cast("Kill Shot", ret => Target.HealthPercent <= 20),
                        Spell.Cast("Dire Beast", ret => MainSettings.Talent60_DB && TalentManager.HasTalent(11) && Spell.PredicFocus() < 70),
                        Spell.Cast("Blink Strike", ret => MainSettings.Talent75_BS && TalentManager.HasTalent(14)),
                        Spell.Cast("Glaive Toss", ret => MainSettings.Talent90_GT && TalentManager.HasTalent(16)),
                        Spell.CastCobraShot(ret => Target, ret => Spell.PredicFocus() < 100),
                        Spell.Cast("Arcane Shot", ret => MainSettings.ArcaneShot && Spell.PredicFocus() >= 100)
                    ))
                )),
            #endregion

            #region PvE
 new Decorator(ret => CurrentWoWContext == WoWContext.PvE && Unit.ValidUnit(Target, true, false),
            new PrioritySelector(
            #region Hold
                //Wait out GCD, do not traverse if can't cast
                Helpers.WaitForCastOrChannel(),
                new Decorator(ret => !Helpers.IsGlobalCooldown(405),
            #endregion
                    new PrioritySelector(
                        Spell.CancelCast("Cobra Shot", ret => Helpers.PlayerPower > 100 & Me.CurrentCastTimeLeft.TotalMilliseconds > 500, RunStatus.Failure),
                        Spell.Cast("Kill Shot", ret => Target.HealthPercent <= 20),
                        Spell.Cast("Dire Beast", ret => MainSettings.Talent60_DB && TalentManager.HasTalent(11) && Spell.PredicFocus() < 70),
                        Spell.Cast("Blink Strike", ret => MainSettings.Talent75_BS && TalentManager.HasTalent(14)),
                        Spell.Cast("Glaive Toss", ret => MainSettings.Talent90_GT && TalentManager.HasTalent(16)),
                        Spell.CastCobraShot(ret => Target, ret => Spell.PredicFocus() < 100),
                        Spell.Cast("Arcane Shot", ret => MainSettings.ArcaneShot && Spell.PredicFocus() >= 100)
                            ))
                    )));
            #endregion
        }
        #endregion

        #region Survival
        private static Composite HandleFocusRegenSV()
        {

            #region PvE
            return new Decorator(ret => CurrentWoWContext == WoWContext.PvE && Unit.ValidUnit(Target, true, false),
                new PrioritySelector(
            #region Hold
                //Wait out GCD, do not traverse if can't cast
                    Helpers.WaitForCastOrChannel(),
                    new Decorator(ret => !Helpers.IsGlobalCooldown(405),
            #endregion
                        new PrioritySelector(
                            Spell.CancelCast("Cobra Shot", ret => Helpers.PlayerPower > 80 & Me.CurrentCastTimeLeft.TotalMilliseconds > 500, RunStatus.Failure),
                            Spell.Cast("Kill Shot", ret => Target.HealthPercent <= 20),
                            Spell.Cast("Explosive Shot", ret => MainSettingsSV.ExplosiveShot && Me.ActiveAuras.ContainsKey("Lock and Load")),
                            Spell.Cast("Dire Beast", ret => MainSettings.Talent60_DB && TalentManager.HasTalent(11) && Spell.PredicFocus() < 60),
                            Spell.Cast("Blink Strike", ret => MainSettings.Talent75_BS && TalentManager.HasTalent(14)),
                            Spell.Cast("Glaive Toss", ret => MainSettings.Talent90_GT && TalentManager.HasTalent(16)),
                            Spell.CastCobraShot(ret => Target, ret => Spell.PredicFocus() < 80),
                            Spell.Cast("Arcane Shot", ret => MainSettings.ArcaneShot && Spell.PredicFocus() >= 80)
                        ))
                    ));

        }
            #endregion

        #endregion

        #endregion


        #region HandleOffensiveCooldowns

        #region Beast Mastery
        private static Composite HandleOffensiveCooldowns()
        {
            return new PrioritySelector(

            #region PvP
            new Decorator(ret => CurrentWoWContext == WoWContext.PvP && Lists.BreakOnDamage(Target) < 250 && Unit.ValidUnit(Target, true, false),
                new PrioritySelector(
                Spell.BuffSelf("Focus Fire", ret => Me.GetAuraStackCount("Frenzy") >= 5 && !Me.HasAura("The Beast Within")),
                Spell.BuffSelf("Aspect of the Hawk", ret => MainSettings.AspectHawk && !Me.IsMoving && !Me.HasAura(109260) && !Me.HasAura(13165)),
                Spell.BuffSelf("Berserking", ret => CooldownSettings.Racial && Me.Race == WoWRace.Troll),
                Spell.BuffSelf("Blood Fury", ret => CooldownSettings.Racial && Me.Race == WoWRace.Orc),
                Spell.BuffSelf("Fervor", ret => SpellManager.HasSpell("Fervor") && MainSettings.Talent60_FV && (Helpers.PlayerPower < 60 || Me.HasAura("The Beast Within") && Helpers.PlayerPower < 40)),
                Spell.BuffSelf("Rapid Fire", ret => CooldownSettings.RapidFire && ((Me.GotTarget && Target.CurrentHealth > 22000 && Target.IsPlayer) || Lists.TrainingDummies.Contains(Target.Entry))),
                Spell.Cast("Stampede", ret => CooldownSettings.Stampede && Target.Distance <= 40 && ((Target.CurrentHealth > 22000 && Target.IsPlayer) || Lists.TrainingDummies.Contains(Target.Entry))),
                Spell.BuffSelf("Bestial Wrath", ret => CooldownSettings.BestialWrath && NeedBestialWrathPvP()),
                Spell.BuffSelf("Readiness", ret => NeedReadinessPvP()),
                Spell.BuffSelf("Lifeblood", ret => CooldownSettings.Lifeblood))),
            #endregion

            #region PvE
        new Decorator(ret => CurrentWoWContext == WoWContext.PvE && Unit.ValidUnit(Target, true, false),
            new PrioritySelector(
                Spell.BuffSelf("Focus Fire", ret => ((ShouldFocusFire() && Me.GetAuraStackCount("Frenzy") >= 5) || (Me.GetAuraStackCount("Frenzy") >= 1 && Me.GetAuraTimeLeft("Frenzy", false).TotalSeconds <= 2))),
                Spell.BuffSelf("Aspect of the Hawk", ret => MainSettings.AspectHawk && !Me.IsMoving && !Me.HasAura(109260) && !Me.HasAura(13165)),
                PetManager.CreateCastPetActionOn("Rabid", ret => Me, ret => PetSettings.Rabid && Lists.TargetIsBoss(Target)),
                Spell.BuffSelf("Berserking", ret => CooldownSettings.Racial && Me.Race == WoWRace.Troll),
                Spell.BuffSelf("Blood Fury", ret => CooldownSettings.Racial && Me.Race == WoWRace.Orc),
                Spell.Buff("Hunter's Mark",ret => Target, ret => MainSettings.HuntersMark && (Target.HealthPercent > 20 || Lists.TargetIsBoss(Target))),
                Spell.BuffSelf("Fervor", ret => SpellManager.HasSpell("Fervor") && MainSettings.Talent60_FV && (Helpers.PlayerPower < 60 || Helpers.PlayerPower < 40 && Me.HasAura("The Beast Within"))),
                Spell.BuffSelf("Rapid Fire", ret => CooldownSettings.RapidFire && Unit.ValidUnit(Target, true, false) && (Lists.TargetIsBoss(Target) && (Helpers.SpellOnCooldown("Bestial Wrath") || !CooldownSettings.BestialWrath) && !Me.HasAnyAura("The Beast Within", "Bloodlust", "Heroism", "Time Warp", "Ancient Hysteria")) || Lists.TrainingDummies.Contains(Target.Entry)),
                Spell.Cast("Stampede", ret => CooldownSettings.Stampede && Lists.TargetIsBoss(Target) && (Target.CurrentHealth > 1000000 || Lists.TrainingDummies.Contains(Target.Entry))),
                Spell.Cast("Bestial Wrath", ret => CooldownSettings.BestialWrath && NeedBestialWrathPvE()),
                Spell.BuffSelf("Readiness", ret => Lists.TargetIsBoss(Target) && NeedReadinessPvE() && CooldownSettings.Readiness),
                Spell.BuffSelf("Lifeblood", ret => CooldownSettings.Lifeblood && Target.MaxHealth > Me.MaxHealth)
             ))
            #endregion
    );
                
        }
        #endregion

        #region Survival
        private static Composite HandleOffensiveCooldownsSV()
        {
            #region PvE
        return new Decorator(ret => CurrentWoWContext == WoWContext.PvE && Unit.ValidUnit(Target, true, false),
            new PrioritySelector(
                Spell.BuffSelf("Aspect of the Hawk", ret => MainSettingsSV.AspectHawk && !Me.IsMoving && !Me.HasAura(109260) && !Me.HasAura(13165)),
                PetManager.CreateCastPetActionOn("Rabid", ret => Me, ret => PetSettingsSV.Rabid && Lists.TargetIsBoss(Target)),
                Spell.BuffSelf("Berserking", ret => CooldownSettingsSV.Racial && Me.Race == WoWRace.Troll),
                Spell.BuffSelf("Blood Fury", ret => CooldownSettingsSV.Racial && Me.Race == WoWRace.Orc),
                Spell.Buff("Hunter's Mark", ret => Target, ret => MainSettingsSV.HuntersMark && (Target.HealthPercent > 20 || Lists.TargetIsBoss(Target))),
                Spell.BuffSelf("Fervor", ret => SpellManager.HasSpell("Fervor") && MainSettingsSV.Talent60_FV && Helpers.PlayerPower < 40),
                Spell.BuffSelf("Rapid Fire", ret => CooldownSettingsSV.RapidFire && (Lists.TargetIsBoss(Target) && !Me.HasAnyAura("Bloodlust", "Heroism", "Time Warp", "Ancient Hysteria")) || Lists.TrainingDummies.Contains(Target.Entry)),
                Spell.Cast("Stampede", ret => Target != null && CooldownSettingsSV.Stampede && Lists.TargetIsBoss(Target) && (Target.CurrentHealth > 1000000 || Lists.TrainingDummies.Contains(Target.Entry))),
                Spell.BuffSelf("Readiness", ret => Lists.TargetIsBoss(Target) && NeedReadinessPvESV() && CooldownSettingsSV.Readiness),
                Spell.BuffSelf("Lifeblood", ret => CooldownSettingsSV.Lifeblood && Target.MaxHealth > Me.MaxHealth)
             ));
            #endregion
        }
        #endregion

        #endregion

        #region HandleTraps

        #region Beast Mastery
        private static Composite HandleTraps()
        {
            return new PrioritySelector(

            #region PvP
            new Decorator(ret => CurrentWoWContext == WoWContext.PvP && Lists.BreakOnDamage(Target) < 250 && Unit.ValidUnit(Target, true, false),
                new PrioritySelector(
                    new Decorator(ret => Target.Distance > 5 && Target.Distance < 40,
                        new PrioritySelector(
                            Spell.CreateTrapBehavior("Ice Trap", true, ret => Target, ret => MainSettings.IceTrapLauncher && Unit.GetClassSpec(Target) == ClassType.Melee),
                            Spell.CreateTrapBehavior("Snake Trap", true, ret => Target, ret => MainSettings.SnakeTrapLauncher && Unit.GetClassSpec(Target) == ClassType.Melee),
                            Spell.CreateTrapBehavior("Snake Trap", true, ret => Target, ret => MainSettings.SnakeTrapLauncher && Unit.GetClassSpec(Target) == ClassType.MRanged),
                            Spell.CreateTrapBehavior("Freezing Trap", true, ret => Target, ret => MainSettings.FreezingTrapLauncher && Unit.GetClassSpec(Target) == ClassType.PRanged),
                            Spell.CreateTrapBehavior("Explosive Trap", true, ret => Target, ret => MainSettings.ExplosiveTrapLauncher && Unit.GetClassSpec(Target) == ClassType.Melee)
                        )
                    ),
                    new Decorator(ret => Target.Guid == Me.Guid && Target.Distance <= 5,
                        new PrioritySelector(
                            Spell.CreateTrapBehavior("Ice Trap", true, ret => Target, ret => MainSettings.IceTrap && Unit.GetClassSpec(Target) == ClassType.Melee),
                            Spell.CreateTrapBehavior("Snake Trap", true, ret => Target, ret => MainSettings.SnakeTrap && Unit.GetClassSpec(Target) == ClassType.Melee),
                            Spell.CreateTrapBehavior("Snake Trap", true, ret => Target, ret => MainSettings.SnakeTrap && Unit.GetClassSpec(Target) == ClassType.MRanged),
                            Spell.CreateTrapBehavior("Freezing Trap", true, ret => Target, ret => MainSettings.FreezingTrap && Unit.GetClassSpec(Target) == ClassType.PRanged),
                            Spell.CreateTrapBehavior("Explosive Trap", true, ret => Target, ret => MainSettings.ExplosiveTrap && Unit.GetClassSpec(Me.CurrentTarget) == ClassType.Melee)
                        )
                    )
                )
            ),
            #endregion

            #region PvE
        new Decorator(ret => CurrentWoWContext == WoWContext.PvE && Unit.ValidUnit(Target, true, false),
            new PrioritySelector(
                new Decorator(ret => Target.Distance > 5 && Target.Distance < 40,
                    new PrioritySelector(
                        Spell.CreateTrapBehavior("Explosive Trap", true, ret => Target, ret => MainSettings.STEXPLTrapLauncher),
                        Spell.CreateTrapBehavior("Explosive Trap",true,ret => Target, ret => MainSettings.ExplosiveTrapLauncher && Helpers.AddCount(10) > 2),
                        Spell.CreateTrapBehavior("Ice Trap", true, ret => Target, ret => MainSettings.IceTrapLauncher),
                        Spell.CreateTrapBehavior("Snake Trap", true, ret => Target, ret => MainSettings.SnakeTrapLauncher),
                        Spell.CreateTrapBehavior("Freezing Trap", true, ret => Target, ret => MainSettings.FreezingTrapLauncher)
                        )),
                new Decorator(ret=> Target.Distance <= 5 ,
                    new PrioritySelector(
                        Spell.CreateTrapBehavior("Explosive Trap", false, ret => Target, ret => MainSettings.STEXPLTrapLauncher),
                        Spell.CreateTrapBehavior("Explosive Trap", false, ret => Target, ret => MainSettings.ExplosiveTrap && Helpers.AddCount(10) > 2),
                        Spell.CreateTrapBehavior("Ice Trap", false, ret => Target, ret => MainSettings.IceTrap),
                        Spell.CreateTrapBehavior("Snake Trap", false, ret => Target, ret => MainSettings.SnakeTrap),
                        Spell.CreateTrapBehavior("Freezing Trap", false, ret => Target, ret => MainSettings.FreezingTrap)
                ))
            ))
            #endregion
);
        }
        #endregion

        #region Survival
        private static Composite HandleTrapsSV()
        {
        #region PvE
       return new Decorator(ret => CurrentWoWContext == WoWContext.PvE && Unit.ValidUnit(Target, true, false),
            new PrioritySelector(
                new Decorator(ret => Target.Distance > 5,
                    new PrioritySelector(
                        Spell.CreateTrapBehavior("Explosive Trap", true, ret => Target, ret => MainSettingsSV.STEXPLTrapLauncher),
                        Spell.CreateTrapBehavior("Explosive Trap", true, ret => Target, ret => MainSettingsSV.ExplosiveTrapLauncher && Helpers.AddCount(10) > 2),
                        Spell.CreateTrapBehavior("Ice Trap", true, ret => Target, ret => MainSettingsSV.IceTrapLauncher),
                        Spell.CreateTrapBehavior("Snake Trap", true, ret => Target, ret => MainSettingsSV.SnakeTrapLauncher),
                        Spell.CreateTrapBehavior("Freezing Trap", true, ret => Target, ret => MainSettingsSV.FreezingTrapLauncher)
                        )),
                new Decorator(ret => Target.Distance <= 5,
                    new PrioritySelector(
                        Spell.CreateTrapBehavior("Explosive Trap", false, ret => Target, ret => MainSettingsSV.STEXPLTrapLauncher),
                        Spell.CreateTrapBehavior("Explosive Trap", false, ret => Target, ret => MainSettingsSV.ExplosiveTrap && Helpers.AddCount(10) > 2),
                        Spell.CreateTrapBehavior("Ice Trap", false, ret => Target, ret => MainSettingsSV.IceTrap),
                        Spell.CreateTrapBehavior("Snake Trap", false, ret => Target, ret => MainSettingsSV.SnakeTrap),
                        Spell.CreateTrapBehavior("Freezing Trap", false, ret => Target, ret => MainSettingsSV.FreezingTrap)
                ))
            ));
            #endregion
        }
        #endregion

        #endregion

        #region HandleAoE

        #region Beast Mastery
        private static Composite HandleAoE()
        {
            return new PrioritySelector(

            #region PvP
        new Decorator(ret => CurrentWoWContext == WoWContext.PvP,
                    new PrioritySelector()),
            #endregion

            #region PvE
        new Decorator(ret => CurrentWoWContext == WoWContext.PvE,
            new PrioritySelector(
                Spell.CreateTrapBehavior("Explosive Trap", true, ret => Target, ret => Target.Distance >= 5 && MainSettings.AoEExpTrap),
                Spell.CreateTrapBehavior("Explosive Trap", false, ret => Target, ret => Target.Distance < 5 && MainSettings.AoEExpTrap),
                Spell.BuffSelf("Focus Fire", ret => ((ShouldFocusFire() && Pet.GetAuraStackCount("Frenzy") == 5) || (Pet.GetAuraStackCount("Frenzy") >= 1 && Pet.GetAuraTimeLeft("Frenzy",false).TotalSeconds < 2 ))),
                Spell.Cast("Bestial Wrath", ret => CooldownSettings.BestialWrath && Pet.Location.Distance(Target.Location) <= 25 && !Me.HasAura("The Beast Within")),
                Spell.Cast("Kill Shot", ret => MainSettings.KillShot && Target.HealthPercent < 20 && Spell.PredicFocus() < 40),
                Spell.Cast("Glaive Toss", ret => MainSettings.Talent90_GT && TalentManager.HasTalent(16)),
                Spell.Cast("Barrage", ret => MainSettings.Talent90_BRG && TalentManager.HasTalent(18)),
                Spell.Cast("Multi-Shot", ret => MainSettings.MultiShot && (!Pet.HasAura("Beast Cleave") || Pet.GetAuraTimeLeft("Beast Cleave", false).TotalSeconds < 1 || Spell.PredicFocus() >= 70 || Me.HasAnyAura("The Beast Within", "Thrill of the Hunt") || (TalentManager.HasTalent(10) && !Helpers.SpellOnCooldown("Fervor")))),
                Spell.Cast("Dire Beast", ret => MainSettings.AoEDireBeast && Spell.PredicFocus() < 80),
                Spell.Cast("Blink Strike", ret => MainSettings.Talent75_BS && TalentManager.HasTalent(14) && Me.GotAlivePet && Spell.PredicFocus() < 40),
                Spell.Buff("Lynx Rush", true, ret => Target, ret => MainSettings.AoELynxRush && Me.GotAlivePet && Spell.PredicFocus() < 40 && Pet.Location.Distance(Target.Location) <= 25 && !Me.HasAura("The Beast Within") && DpsMeter.GetCombatTimeLeft(Target, 3).TotalSeconds > 15),
                PetManager.CreateCastPetAction("Froststorm Breath", ret => Me.GotAlivePet && PetSettings.FrostromBreath && Pet.Location.Distance(Target.Location) < 10),
                PetManager.CreateCastPetAction("Burrow Attack", ret => Me.GotAlivePet && PetSettings.Burrow  && Pet.Location.Distance(Target.Location) < 10),
                new Decorator(ret => Spell.PredicFocus() < 20 || (!Me.HasAnyAura("Thrill of the Hunt", "The Beast Within") && (Spell.PredicFocus() < 40 || (Spell.PredicFocus() < 70 && Pet.GetAuraTimeLeft("Beast Cleave", false).TotalSeconds > 1))),
                Spell.Cast("Steady Shot", ret => !TalentManager.HasTalent(10) || Helpers.SpellOnCooldown("Fervor"))) 
            ))
            #endregion
            );
                
        }
        #endregion

        #region Survival
        private static Composite HandleAoESV()
        {
            #region PvE
        return new Decorator(ret => CurrentWoWContext == WoWContext.PvE,
           new PrioritySelector(
                Spell.CreateTrapBehavior("Explosive Trap", true, ret => Target, ret => Target.Distance >= 5 && MainSettingsSV.AoEExpTrap),
                Spell.CreateTrapBehavior("Explosive Trap", false, ret => Target, ret => Target.Distance < 5 && MainSettingsSV.AoEExpTrap),
                Spell.Cast("Kill Shot", ret => MainSettingsSV.KillShot && Target.HealthPercent < 20 && Spell.PredicFocus() < 40),
                Spell.Cast("Dire Beast", ret => MainSettingsSV.AoEDireBeast && Spell.PredicFocus() < 70),
                Spell.Cast("Glaive Toss", ret => MainSettingsSV.Talent90_GT && TalentManager.HasTalent(16)),
                Spell.PreventDoubleCast("Explosive Shot", 1, ret => Target, ret => MainSettingsSV.ExplosiveShot && Me.ActiveAuras.ContainsKey("Lock and Load")),
                Spell.Cast("Barrage", ret => MainSettingsSV.Talent90_BRG && TalentManager.HasTalent(18)),
                Spell.Cast("Multi-Shot", ret => MainSettingsSV.MultiShot && (Me.CurrentFocus >= 40 || (Me.HasAura("Thrill of the Hunt") && Me.CurrentFocus >= 20) || (TalentManager.HasTalent(10) && !Helpers.SpellOnCooldown("Fervor")))),
                Spell.Cast("Blink Strike", ret => MainSettingsSV.Talent75_BS && TalentManager.HasTalent(14) && Me.GotAlivePet && Spell.PredicFocus() < 40),
                Spell.Buff("Lynx Rush", true, ret => Target, ret => MainSettingsSV.AoELynxRush && Me.GotAlivePet && Spell.PredicFocus() < 40 && DpsMeter.GetCombatTimeLeft(Target, 3).TotalSeconds > 15),
                new Decorator(ret => Spell.PredicFocus() < 20 || (!Me.HasAura("Thrill of the Hunt") && Spell.PredicFocus() < 40),
                Spell.Cast("Steady Shot", ret => !TalentManager.HasTalent(10) || Helpers.SpellOnCooldown("Fervor")))
                ));
            #endregion


        }
        #endregion

        #endregion

        #region HandleSingleTarget

        #region Beast Mastery
        private static Composite HandleSingleTarget()
        {
            return new PrioritySelector(
            #region PvP
            new Decorator(ret => CurrentWoWContext == WoWContext.PvP && Unit.ValidUnit(Target, true, false) && Lists.BreakOnDamage(Target) < 250,
                new PrioritySelector( 
                    #region Hold
                    //Wait out GCD, do not traverse if can't cast
                     Helpers.WaitForCastOrChannel(),
                     new Decorator(ret => !Helpers.IsGlobalCooldown(MainSettings.GCDTimeLeft),
                        new PrioritySelector(
            #endregion

                    PetManager.CreateCastPetActionOn("Venom Web Spray", ret => Target, ret => Target.CurrentTargetGuid == Me.Guid && Lists.Rooted(Target) <= 0),
                    Spell.Cast("Hunter's Mark",ret => MainSettings.HuntersMark && !Target.HasAura("Hunter's Mark")),
                    Spell.Cast("Kill Shot", ret => Target.HealthPercent <= 20),

                new Decorator(ret => ShouldCobraShotPvP(),
                    new PrioritySelector(
                    Spell.CastCobraShot(ret => Target, ret => MainSettings.KillCommand && Me.GotAlivePet && Spell.PredicFocus() < 56 && Helpers.PlayerPower > 40 && !Me.HasAura("The Beast Within") && Helpers.GetSpellCooldown("Kill Command").TotalMilliseconds > 1000 && Helpers.GetSpellCooldown("Kill Command").TotalMilliseconds < 3000),
                    Spell.CastCobraShot(ret => Target, ret => MainSettings.KillCommand && Me.GotAlivePet && (Spell.PredicFocus() < 18 || (!Me.HasAura("The Beast Within") && Spell.PredicFocus() < 38)) && Helpers.GetSpellCooldown("Kill Command").TotalMilliseconds < 3000)
                    )),
                new Decorator(ret => Me.GotAlivePet,
                    new PrioritySelector(// Pet dependand, check only once for speed
                        Spell.Cast("Kill Command", ret => Pet.Location.Distance(Target.Location) <= 25 && Pet.InLoSSS(Target)),
                        Spell.Cast("Blink Strike", ret => MainSettings.Talent75_BS && TalentManager.HasTalent(14) && Pet.Location.Distance(Target.Location) <= 40 && Pet.InLoSSS(Target)),
                        Spell.Cast("Lynx Rush", ret => MainSettings.Talent75_LXR && TalentManager.HasTalent(15) && Pet.Location.Distance(Target.Location) <= 10 && Target.GetAuraTimeLeft("Lynx Rush", true).TotalSeconds < 4 && Pet.InLoSSS(Target))
                    )),

                    Spell.PreventDoubleCast("Serpent Sting", 2, ret => Target, ret => ShouldSerpentStingPvP() && Target.GetAuraTimeLeft("Serpent Sting", true).TotalSeconds < 1),

                    Spell.Cast("Dire Beast", ret => MainSettings.Talent60_DB && TalentManager.HasTalent(15)),
                    Spell.Buff("A Murder of Crows", ret => Target, ret => MainSettings.Talent75_AMOC && TalentManager.HasTalent(13) && Target.IsPlayer && !Target.HasMyAura("A Murder of Crows") && (!CooldownSettings.BestialWrath || Me.HasAura("The Beast Within") || Helpers.GetSpellCooldown("Bestial Wrath").TotalSeconds >= 10 || !Me.GotAlivePet || HasSpeedBuff())),
                    Spell.Buff("Widow Venom", true, ret => Target, ret => Spell.PredicFocus() >= 55 && Target.GetAuraTimeLeft("Widow Venom", true).TotalSeconds <= 1 && (!MainSettings.KillCommand || Helpers.GetSpellCooldown("Kill Command").TotalSeconds > 2), "Wild Strike", "Mortal Strike", "Wound Poison"),
                    Spell.Cast("Glaive Toss", ret => MainSettings.Talent90_GT && TalentManager.HasTalent(16) && (!Me.CurrentMap.IsArena || !Unit.ValidFocus(true, false) || Lists.BreakOnDamage(Focus) < 500 || Target.Location.Distance(Focus.Location) > 7)),
                    Spell.Cast("Powershot", ret => MainSettings.Talent90_PWRS && TalentManager.HasTalent(17) && Me.HasAnyAura("Rapid Fire", "Focus Fire", "Bloodlust", "Heroism", "Ancient Hysteria", "Time Warp") && (Helpers.GetSpellCooldown("Kill Command").TotalMilliseconds > 2500 || !Me.GotAlivePet || !MainSettings.KillCommand)),
                    Spell.Cast("Barrage", ret => MainSettings.Talent90_BRG && TalentManager.HasTalent(18)),

                    Spell.CastCobraShot(ret => Target, ret => Spell.PredicFocus() < 90 && Lists.HasteBuffs() >= 50
                    && (!TalentManager.HasTalent(10) || (Helpers.GetSpellCooldown("Fervor").TotalMilliseconds > 750 && Helpers.GetSpellCooldown("Fervor").TotalSeconds < 29))),
                        
                    Spell.Cast("Arcane Shot", ret => MainSettings.ArcaneShot && (Helpers.PlayerPower > 62 || (Me.HasAura("The Beast Within") && Helpers.PlayerPower > 42) || Me.HasAura("Thrill of the Hunt")
                    || (TalentManager.HasTalent(10) && Helpers.GetSpellCooldown("Fervor").TotalMilliseconds < 500) || !Me.GotAlivePet || !MainSettings.KillCommand || !Pet.InLoSSS(Target))),

                    new Decorator(ret => ShouldCobraShotPvP(),
                    Spell.CastCobraShot(ret => Target, ret => Spell.PredicFocus() < MainSettings.FocusAmount || (Me.HasAura("The Beast Within") && Spell.PredicFocus() < MainSettings.FocusAmount / 2)))
                    )))),
            #endregion

            #region PvE
        new Decorator(ret => CurrentWoWContext == WoWContext.PvE && Unit.ValidUnit(Target,true,false),
            new PrioritySelector(
                #region Hold
                //Wait out GCD, do not traverse if can't cast
                Helpers.WaitForCastOrChannel(),
                new Decorator(ret => !Helpers.IsGlobalCooldown(MainSettings.GCDTimeLeft),
                #endregion
                    new PrioritySelector(
                        //Cancel Spell Casts
                        /*new Decorator(ret => Target.HasMyAura("Serpent Sting") && Target.GetAuraTimeLeft("Serpent Sting", true).TotalSeconds < 5,
                            new Sequence(
                            Spell.CancelCast(ret => Me.CastingSpell.Name != "Cobra Shot" && CLua.PlayerPower < 90, RunStatus.Failure),
                            Spell.CastCobraShot(ret => Target, ret => CLua.PlayerPower < 90))), 
                        Spell.CancelCast("Cobra Shot", ret => Helpers.PlayerPower > 90 & Me.CurrentCastTimeLeft.TotalMilliseconds > 700, RunStatus.Failure),
                        //Single Target Rotation
                        Spell.Cast("Kill Shot", ret => Target.HealthPercent <= 20),
                        Spell.Cast("Kill Command", ret => Me.GotAlivePet),
                        Spell.PreventDoubleCast("Serpent Sting", 2, ret => Target, ret => ShouldSerpentStingPvE() && Target.GetAuraTimeLeft("Serpent Sting", true).TotalSeconds < 1),

                new Decorator(ret => ShouldCobraShotPvE(),
                    new PrioritySelector(
                        Spell.CastCobraShot(ret => Target, ret => MainSettings.KillCommand && Me.GotAlivePet && !Me.HasAura("The Beast Within") && Spell.PredicFocus() < 70 && Helpers.PlayerPower > 40 && Helpers.GetSpellCooldown("Kill Command").TotalMilliseconds > 1000 && Helpers.GetSpellCooldown("Kill Command").TotalMilliseconds < 3000),
                        Spell.CastCobraShot(ret => Target, ret => MainSettings.KillCommand && Me.GotAlivePet && (Spell.PredicFocus() < 18 || (!Me.HasAura("The Beast Within") && Spell.PredicFocus() < 38)) && Helpers.GetSpellCooldown("Kill Command").TotalMilliseconds < 3000)
                    )),

                        Spell.Buff("A Murder of Crows", ret => Target, ret => TalentManager.HasTalent(13) && MainSettings.Talent75_AMOC && !Target.HasMyAura("A Murder of Crows") && (Lists.TargetIsBoss(Target) || DpsMeter.GetCombatTimeLeft(Target, 5).TotalSeconds > 60) && (!CooldownSettings.BestialWrath || Me.HasAura("The Beast Within") || Helpers.GetSpellCooldown("Bestial Wrath").TotalSeconds >= 10 || !Me.GotAlivePet || HasSpeedBuff())),
                        Spell.Cast("Dire Beast", ret => MainSettings.Talent60_DB && TalentManager.HasTalent(11) && (Lists.TargetIsBoss(Target) || Spell.PredicFocus() < 60 || Target.CurrentHealth > 150000)),
                        Spell.Cast("Glaive Toss", ret => MainSettings.Talent90_GT && TalentManager.HasTalent(16)),
                        Spell.Cast("Blink Strike", ret => MainSettings.Talent75_BS && TalentManager.HasTalent(14) && Me.GotAlivePet),
                        Spell.Cast("Lynx Rush", ret => Target, ret => MainSettings.Talent75_LXR && TalentManager.HasTalent(15) && Me.GotAlivePet && Target.GetAuraTimeLeft("Lynx Rush", true).TotalSeconds < 4 && (Helpers.GetSpellCooldown("Bestial Wrath").TotalSeconds >= 10 || Me.HasAura("The Beast Within")) && (Lists.TargetIsBoss(Target) || DpsMeter.GetCombatTimeLeft(Target, 3).TotalSeconds > 20)),
                        Spell.Cast("Powershot", ret => MainSettings.Talent90_PWRS && TalentManager.HasTalent(17) && Me.HasAnyAura("Rapid Fire", "Focus Fire", "Bloodlust", "Heroism", "Ancient Hysteria", "Time Warp") && (Helpers.GetSpellCooldown("Kill Command").TotalMilliseconds > 2500 || !Me.GotAlivePet || !MainSettings.KillCommand)),
                        Spell.Cast("Barrage", ret => MainSettings.Talent90_BRG && TalentManager.HasTalent(18) && Lists.TargetIsBoss(Target)),

                        Spell.CastCobraShot(ret => Target, ret => Spell.PredicFocus() < 80 && Lists.HasteBuffs() >= 40
                        && (!TalentManager.HasTalent(10) || (Helpers.GetSpellCooldown("Fervor").TotalMilliseconds > 750 && Helpers.GetSpellCooldown("Fervor").TotalSeconds < 29))),

                        Spell.Cast("Arcane Shot", ret => MainSettings.ArcaneShot && (Helpers.PlayerPower > 62 || (Me.HasAura("The Beast Within") && Spell.PredicFocus() > 42) || Me.HasAura("Thrill of the Hunt") 
                        || (TalentManager.HasTalent(10) && Helpers.GetSpellCooldown("Fervor").TotalMilliseconds < 500) || !Me.GotAlivePet || !MainSettings.KillCommand || !Pet.InLoSSS(Target))),

                        new Decorator(ret => ShouldCobraShotPvE(),
                            Spell.CastCobraShot(ret => Target, ret => Spell.PredicFocus() < MainSettings.FocusAmount || (Me.HasAura("The Beast Within") && Helpers.PlayerPower < MainSettings.FocusAmount / 2)))
                            ))
                    )));
            #endregion
        }
        #endregion

        #region Survival
        private static Composite HandleSingleTargetSV()
        {

     #region PvE
     return new Decorator(ret => CurrentWoWContext == WoWContext.PvE && Unit.ValidUnit(Target,true,false),
            new PrioritySelector(
                #region Hold
                //Wait out GCD, do not traverse if can't cast
                Helpers.WaitForCastOrChannel(),
                new Decorator(ret => !Helpers.IsGlobalCooldown(MainSettingsSV.GCDTimeLeft),
                #endregion

            new PrioritySelector(
                //Cancel Spell Casts
                Spell.CastCobraShot(ret => Target, ret => !Me.IsCasting && Target.HasMyAura("Serpent Sting") && Target.GetAuraTimeLeft("Serpent Sting", true).TotalSeconds < 4 && Helpers.PlayerPower < 80),
                Spell.CancelCast("Cobra Shot", ret => Helpers.PlayerPower > 90 & Me.CurrentCastTimeLeft.TotalMilliseconds > 700, RunStatus.Failure),
                //Single Target Rotation
                Spell.Cast("Kill Shot", ret => Target.HealthPercent <= 20),

                new Decorator(ret => (!TalentManager.HasTalent(10) || (Helpers.GetSpellCooldown("Fervor").TotalMilliseconds > 750 && Helpers.GetSpellCooldown("Fervor").TotalSeconds < 29)),
                    new PrioritySelector(
                        Spell.CastCobraShot(ret => Target, ret => Helpers.GetSpellCooldown("Black Arrow").TotalMilliseconds < 3000 && MainSettingsSV.BlackArrow && Spell.PredicFocus() < 33),
                        Spell.CastCobraShot(ret => Target, ret => Helpers.GetSpellCooldown("Explosive Shot").TotalMilliseconds < 3000 && MainSettingsSV.ExplosiveShot && Spell.PredicFocus() < 23)
                        )),

                        Spell.Cast("Explosive Shot", ret => MainSettingsSV.ExplosiveShot && Me.ActiveAuras.ContainsKey("Lock and Load")),
                        Spell.Cast("Black Arrow", ret => MainSettingsSV.BlackArrow && Target.GetAuraTimeLeft("Black Arrow", true).TotalSeconds < 1 && (DpsMeter.GetCombatTimeLeft(Target, 3).TotalSeconds > 20 || Lists.TargetIsBoss(Target))),
                        Spell.Cast("A Murder of Crows", ret => Target, ret => TalentManager.HasTalent(13) && MainSettingsSV.Talent75_AMOC && !Target.HasMyAura("A Murder of Crows") && (Lists.TargetIsBoss(Target) || DpsMeter.GetCombatTimeLeft(Target, 5).TotalSeconds > 60)),
                        Spell.Cast("Explosive Shot", ret => MainSettingsSV.ExplosiveShot),
                        Spell.PreventDoubleCast("Serpent Sting", 2, ret => Target, ret => ShouldSerpentStingPvE() && Target.GetAuraTimeLeft("Serpent Sting", true).TotalSeconds < 1),
                        Spell.Cast("Dire Beast", ret => MainSettingsSV.Talent60_DB && TalentManager.HasTalent(11) && (Lists.TargetIsBoss(Target) || Spell.PredicFocus() < 60 || Target.CurrentHealth > 150000)),
                        Spell.Cast("Glaive Toss", ret => MainSettingsSV.Talent90_GT && TalentManager.HasTalent(16)),
                        Spell.Cast("Blink Strike", ret => MainSettingsSV.Talent75_BS && TalentManager.HasTalent(14)),
                        Spell.Cast("Lynx Rush", ret => Target, ret => MainSettingsSV.Talent75_LXR && TalentManager.HasTalent(15) && Me.GotAlivePet && Target.GetAuraTimeLeft("Lynx Rush", true).TotalSeconds < 4 && Pet.Location.Distance(Target.Location) < 25 && (Lists.TargetIsBoss(Target) || DpsMeter.GetCombatTimeLeft(Target, 3).TotalSeconds > 20)),
                        Spell.Cast("Powershot", ret => MainSettingsSV.Talent90_PWRS && TalentManager.HasTalent(17) && Me.HasAnyAura("Rapid Fire", "Bloodlust", "Heroism", "Ancient Hysteria", "Time Warp") && (Helpers.GetSpellCooldown("Explosive Shot").TotalMilliseconds > 2500 || !MainSettingsSV.ExplosiveShot)),
                        Spell.Cast("Barrage", ret => MainSettingsSV.Talent90_BRG && TalentManager.HasTalent(18) && Lists.TargetIsBoss(Target)),

                        Spell.CastCobraShot(ret => Target, ret => Spell.PredicFocus() < 70 && Lists.HasteBuffs() >= 40
                        && (!TalentManager.HasTalent(10) || (Helpers.GetSpellCooldown("Fervor").TotalMilliseconds > 750 && Helpers.GetSpellCooldown("Fervor").TotalSeconds < 29))),

                        Spell.Cast("Arcane Shot", ret => MainSettingsSV.ArcaneShot && !Helpers.SpellOnGCD("Explosive Shot") && !Me.HasAura("Thrill of the Hunt") && (Helpers.PlayerPower >= 60 || (Helpers.GetSpellCooldown("Explosive Shot").TotalSeconds > 2.5 && Helpers.PlayerPower >= 50))),
                        Spell.Cast("Arcane Shot", ret => MainSettingsSV.ArcaneShot && Me.HasAura("Thrill of the Hunt") && (!MainSettingsSV.ExplosiveShot || !Helpers.SpellOnGCD("Explosive Shot"))),
                        new Decorator(ret => ShouldCobraShotPvESV(),
                            Spell.CastCobraShot(ret => Target, ret => Spell.PredicFocus() < MainSettingsSV.FocusAmount)
                        )
            ))));
        }
        #endregion

        #endregion

        #endregion

        #region PvE Booleans
        /*
        private static bool ShouldDetterence()
        {
            if (Target == null)
                return false;
            switch (MainSettings.Deterrence)
            {
                case Deterrence.Defend:
                    case Deterrence.Health_Defend:
                    case Deterrence.Health_Reflect:
                    case Deterrence.Health_Reflect_Defend:
                    case Deterrence.Low_Health:
                    case Deterrence.Reflect:
                    case Deterrence.Reflect_Defend:
                    case Deterrence.Never:
                    return false;
            }
            return false;
        }
       
        private static bool ShouldFeignDeath()
        {
            if (Me.Specialization != WoWSpec.HunterSurvival && Unit.ValidUnit(Target, true, false))
            {
                switch (MainSettings.FeignDeath)
                {
                    case FeignDeath.Aggro_Health:
                        return Target.ThreatInfo.RawPercent > 90 && Target.CurrentTargetGuid == Me.Guid && (Target.IsCasting || Target.Distance < 10) && Me.HealthPercent < MainSettings.FeignHealth;
                    case FeignDeath.HighThreat:
                        return Target.ThreatInfo.RawPercent > 90;
                    case FeignDeath.LowHealth:
                        return Target.CurrentTargetGuid == Me.Guid && Me.HealthPercent < MainSettings.FeignHealth;
                    case FeignDeath.OnAggro:
                        return Target.ThreatInfo.RawPercent > 90 && Target.CurrentTargetGuid == Me.Guid && (Target.IsCasting || Target.Distance < 10);
                    case FeignDeath.Threat_Health:
                        return Target.ThreatInfo.RawPercent > 90 && Me.HealthPercent < MainSettings.FeignHealth;
                    case FeignDeath.Never:
                        return false;
                }
                return false;
            }
            if (Me.Specialization == WoWSpec.HunterSurvival && Unit.ValidUnit(Target, true, false))
            {
                switch (MainSettingsSV.FeignDeath)
                {
                    case FeignDeath.Aggro_Health:
                        return Target.ThreatInfo.RawPercent > 90 && Target.CurrentTargetGuid == Me.Guid && (Target.IsCasting || Target.Distance < 10) && Me.HealthPercent < MainSettingsSV.FeignHealth;
                    case FeignDeath.HighThreat:
                        return Target.ThreatInfo.RawPercent > 90;
                    case FeignDeath.LowHealth:
                        return Target.CurrentTargetGuid == Me.Guid && Me.HealthPercent < MainSettingsSV.FeignHealth;
                    case FeignDeath.OnAggro:
                        return Target.ThreatInfo.RawPercent > 90 && Target.CurrentTargetGuid == Me.Guid && (Target.IsCasting || Target.Distance < 10);
                    case FeignDeath.Threat_Health:
                        return Target.ThreatInfo.RawPercent > 90 && Me.HealthPercent < MainSettingsSV.FeignHealth;
                    case FeignDeath.Never:
                        return false;
                }
                return false;
            }
            return false;
        }
        private static bool ShouldIntimidate()
        {
            if (Target == null || !Me.GotAlivePet)
                return false;
            switch (MainSettings.Intimidation)
            {
                    case Intimidate.Interrupt:
                    return Pet.Location.Distance(Target.Location) < 9 && (Lists.InterruptCastNoChannel(Target) > 200 || Lists.InterruptCastChannel(Target) > 200);
                    case Intimidate.Defend:
                    return Pet.Location.Distance(Target.Location) < 9 && Target.Distance <= 20 && Target.CurrentTargetGuid == Me.Guid;
                    case Intimidate.Interrupt_Defend:
                    return (Target.Distance <= 20 && Target.CurrentTargetGuid == Me.Guid || (Lists.InterruptCastNoChannel(Target) > 200 || Lists.InterruptCastChannel(Target) > 200));
                    case Intimidate.Interrupt_Defend_Health:
                    return Target.HealthPercent < MainSettings.IntimidationHealth && (Target.Distance <= 20 && Target.CurrentTargetGuid == Me.Guid || (Lists.InterruptCastNoChannel(Target) > 200 || Lists.InterruptCastChannel(Target) > 200));
                    case Intimidate.Interrupt_Health:
                    return Target.HealthPercent < MainSettings.IntimidationHealth && (Lists.InterruptCastNoChannel(Target) > 200 || Lists.InterruptCastChannel(Target) > 200);
                    case Intimidate.LowHealth:
                    return Target.HealthPercent < MainSettings.IntimidationHealth;
                    case Intimidate.Never:
                    return false;
            }
            return false;
        }
        private static bool ShouldScatterShot()
        {
            if (Target == null)
                return false;
            if (Me.Specialization != WoWSpec.HunterSurvival)
            {
                switch (MainSettings.ScatterShot)
                {
                    case ScatterShot.Defend:
                        return Target.Distance <= 20 && Target.CurrentTargetGuid == Me.Guid;
                    case ScatterShot.Interrupt:
                        return Target.Distance <= 20 && (Lists.InterruptCastNoChannel(Target) > 200 || Lists.InterruptCastChannel(Target) > 200);
                    case ScatterShot.Interrupt_Defend:
                        return Target.Distance <= 20 && (Target.CurrentTargetGuid == Me.Guid || (Lists.InterruptCastNoChannel(Target) > 200 || Lists.InterruptCastChannel(Target) > 200));
                    case ScatterShot.Never:
                        return false;
                }
                return false;
            }
            else if (Me.Specialization == WoWSpec.HunterSurvival)
            {
                switch (MainSettingsSV.ScatterShot)
                {
                    case ScatterShot.Defend:
                        return Target.Distance <= 20 && Target.CurrentTargetGuid == Me.Guid;
                    case ScatterShot.Interrupt:
                        return Target.Distance <= 20 && (Lists.InterruptCastNoChannel(Target) > 200 || Lists.InterruptCastChannel(Target) > 200);
                    case ScatterShot.Interrupt_Defend:
                        return Target.Distance <= 20 && (Target.CurrentTargetGuid == Me.Guid || (Lists.InterruptCastNoChannel(Target) > 200 || Lists.InterruptCastChannel(Target) > 200));
                    case ScatterShot.Never:
                        return false;
                }
                return false;
            }
            return false;
        }
        private static bool ShouldSerpentStingPvE()
        {
            if (Target == null)
                return false;
            if (Me.Specialization != WoWSpec.HunterSurvival)
            {
                switch (MainSettings.SerpentSting)
                {
                    case SerpentSting.Always:
                        return true;
                    case SerpentSting.Sometimes:
                        return Lists.TargetIsBoss(Target) || DpsMeter.GetCombatTimeLeft(Target, 2).TotalSeconds > 12;
                    case SerpentSting.Never:
                        return false;

                }
                return false;
            }
            if (Me.Specialization == WoWSpec.HunterSurvival)
            {
                switch (MainSettingsSV.SerpentSting)
                {
                    case SerpentSting.Always:
                        return true;
                    case SerpentSting.Sometimes:
                        return Lists.TargetIsBoss(Target) || DpsMeter.GetCombatTimeLeft(Target, 2).TotalSeconds > 12;
                    case SerpentSting.Never:
                        return false;

                }
                return false;
            }
            return false;
        }
        private static bool ShouldFocusFire()
        {
            if (Target == null)
                return false;
            return Me.GotAlivePet && !Me.HasAura("The Beast Within") &&
                   (Helpers.GetSpellCooldown("Bestial Wrath").TotalSeconds > 9 ||
                    !Helpers.SpellOnCooldown("Bestial Wrath") &&
                    (Target.CurrentHealth <= 200000 || !Me.HasAnyAura("Rapid Fire","Bestial Wrath","Heroism","Bloodlust", "Time Warp", "Ancient Hysteria")));
        }
        private static bool HasSpeedBuff()
        {
            return Me.HasAnyAura("Bloodlust", "Heroism", "Ancient Hysteria", "Time Warp", "Rapid Fire", "Focus Fire");
        }
        private static bool NeedExhilaration()
        {
            if (Target == null)
                return false;
            return Me.HealthPercent < 68 || (Me.GotAlivePet && Pet.HealthPercent < 15 && PetManager.CanCastPetAction("Heart of the Phoenix") && Helpers.GetSpellCooldown("Heart of the Phoenix").TotalSeconds > 10) ||
                      (Me.GotAlivePet && Pet.HealthPercent < 15 && !PetManager.CanCastPetAction("Heart of the Phoenix"));

        }
        private static bool NeedBestialWrathPvE()
        {
            if (Target == null)
                return false;
            return Me.GotAlivePet && BestialExtensionPvE() && !HasSpeedBuff() && !Me.HasAura("The Beast Within") && (Target.CurrentHealth > 100000 || Lists.TargetIsBoss(Target));

        }
        private static bool BestialExtensionPvE()
        {
            if (Target == null)
                return false;
            return (!TalentManager.HasTalent(15) || Helpers.GetSpellCooldown("Lynx Rush").TotalSeconds >= 10 ||
                    Helpers.GetSpellCooldown("Lynx Rush").TotalMilliseconds < 1500) &&
                   (!TalentManager.HasTalent(13) || Helpers.GetSpellCooldown("A Murder of Crows").TotalSeconds >= 10 ||
                    Helpers.GetSpellCooldown("A Murder of Crows").TotalMilliseconds < 1500);

        }
        private static bool NeedReadinessPvE()
        {
            if (Target == null)
                return false;
            return Helpers.GetSpellCooldown("Rapid Fire").TotalSeconds > 20 &&
                   Helpers.GetSpellCooldown("Bestial Wrath").TotalSeconds > 10 &&
                   (Helpers.GetSpellCooldown("Kill Command").TotalSeconds > 2 || !MainSettings.KillCommand) &&
                   ReadinessExtensionPvE();

        }
        private static bool ReadinessExtensionPvE()
        { 
            return (TalentManager.HasTalent(15) && Helpers.GetSpellCooldown("Lynx Rush").TotalSeconds > 9 ||!TalentManager.HasTalent(15))
                    && (TalentManager.HasTalent(16) && Helpers.GetSpellCooldown("Glaive Toss").TotalSeconds > 2 ||!TalentManager.HasTalent(16))
                    && (TalentManager.HasTalent(17) && Helpers.GetSpellCooldown("Powershot").TotalSeconds > 5 || !TalentManager.HasTalent(17))
                    && (TalentManager.HasTalent(18) && Helpers.GetSpellCooldown("Barrage").TotalSeconds > 5 || !TalentManager.HasTalent(18))
                    && (TalentManager.HasTalent(10) && Helpers.GetSpellCooldown("Fervor").TotalSeconds > 9 || !TalentManager.HasTalent(10))
                    && (TalentManager.HasTalent(13) && Helpers.GetSpellCooldown("A Murder of Crows").TotalSeconds > 9 || !TalentManager.HasTalent(13))
                    && (TalentManager.HasTalent(14) && Helpers.GetSpellCooldown("Blink Strike").TotalSeconds > 2 || !TalentManager.HasTalent(14))
                    && (TalentManager.HasTalent(11) && Helpers.GetSpellCooldown("Dire Beast").TotalSeconds > 4 || !TalentManager.HasTalent(11));
        }
        private static bool ShouldCobraShotPvE()
        {
            if (Target == null)
                return false;
            return /*(Helpers.GetSpellCooldown("Kill Command").TotalSeconds > Math.Round(((Spell.GetSpellCastTime("Cobra Shot").TotalSeconds / ((100 + Helpers.HasteBuffs()) / 100)) - 0.4), 2) || CLua.PlayerPower < 20 || (!Me.HasAura("The Beast Within") && CLua.PlayerPower < 40))
                    && (!TalentManager.HasTalent(10) || (Helpers.GetSpellCooldown("Fervor").TotalMilliseconds > 750 && Helpers.GetSpellCooldown("Fervor").TotalSeconds < 29));

        }

        private static bool ShouldCobraShotPvESV()
        {
            if (Target == null)
                return false;
            return (!TalentManager.HasTalent(10) || (Helpers.GetSpellCooldown("Fervor").TotalMilliseconds > 750 && Helpers.GetSpellCooldown("Fervor").TotalSeconds < 29));

        }
        private static bool NeedReadinessPvESV()
        {
            if (Target == null)
                return false;
            return Helpers.GetSpellCooldown("Rapid Fire").TotalSeconds > 20 &&
                   (Helpers.GetSpellCooldown("Black Arrow").TotalSeconds > 5 || !MainSettingsSV.BlackArrow) &&
                   (Helpers.GetSpellCooldown("Explosive Shot").TotalSeconds > 2 || !MainSettingsSV.ExplosiveShot) &&
                   ReadinessExtensionPvE();

        }
        #endregion

        #region PvP Booleans
        private static bool ShouldSerpentStingPvP()
        {
            if (Target == null)
                return false;
            switch (MainSettings.SerpentSting)
            {
                case SerpentSting.Always:
                    return true;
                case SerpentSting.Sometimes:
                    return Target.HealthPercent > 25;
                case SerpentSting.Never:
                    return false;

            }
            return false;
        }
        private static bool NeedReadinessPvP()
        {
            if (Target == null)
                return false;
            return CooldownSettings.Readiness &&
                    Helpers.GetSpellCooldown("Rapid Fire").TotalSeconds > 10 &&
                   Helpers.GetSpellCooldown("Bestial Wrath").TotalSeconds > 8 &&
                   ReadinessExtensionPvP();

        }
        private static bool ReadinessExtensionPvP()
        {
            return (!MainSettings.KillCommand || Helpers.GetSpellCooldown("Kill Command").TotalSeconds > 2)
                    && (!Me.CurrentMap.IsArena || Helpers.GetSpellCooldown("Disengage").TotalSeconds > 2)
                    && (!TalentManager.HasTalent(4) || !Me.CurrentMap.IsArena || Helpers.GetSpellCooldown("Silencing Shot").TotalSeconds > 2)
                    && (!TalentManager.HasTalent(5) || !Me.CurrentMap.IsArena || Helpers.GetSpellCooldown("Wyvern Sting").TotalSeconds > 2)
                    && (!TalentManager.HasTalent(6) || !Me.CurrentMap.IsArena || Helpers.GetSpellCooldown("Binding Shot").TotalSeconds > 2)
                    && (!TalentManager.HasTalent(7) || Helpers.GetSpellCooldown("Exhilaration").TotalSeconds > 9)
                    && (!TalentManager.HasTalent(11) || Helpers.GetSpellCooldown("Dire Beast").TotalSeconds > 4)
                    && (!TalentManager.HasTalent(10) || Helpers.GetSpellCooldown("Fervor").TotalSeconds > 8)
                    && (!TalentManager.HasTalent(13) || Helpers.GetSpellCooldown("A Murder of Crows").TotalSeconds > 10)
                    && (!TalentManager.HasTalent(14) || Helpers.GetSpellCooldown("Blink Strike").TotalSeconds > 2)
                    && (!TalentManager.HasTalent(15) || Helpers.GetSpellCooldown("Lynx Rush").TotalSeconds > 10)
                    && (!TalentManager.HasTalent(16) || Helpers.GetSpellCooldown("Glaive Toss").TotalSeconds > 2)
                    && (!TalentManager.HasTalent(17) || Helpers.GetSpellCooldown("Powershot").TotalSeconds > 5)
                    && (!TalentManager.HasTalent(18) || Helpers.GetSpellCooldown("Barrage").TotalSeconds > 5)
                    && (MainSettings.Intimidation == Intimidate.Never || !Me.CurrentMap.IsArena || Helpers.GetSpellCooldown("Intimidation").TotalSeconds > 2);
        }
        private static bool NeedBestialWrathPvP()
        {
            if (Target == null)
                return false;
            return Me.GotAlivePet && Pet.Location.Distance(Target.Location) <= 25 && !Me.HasAura("The Beast Within") &&
                   BestialExtensionPvP();

        }
        private static bool BestialExtensionPvP()
        {
            if (Target == null)
                return false;
            return (!TalentManager.HasTalent(15) || Helpers.GetSpellCooldown("Lynx Rush").TotalSeconds > 5 || Helpers.GetSpellCooldown("Lynx Rush").TotalMilliseconds < 1500) &&
                   (!TalentManager.HasTalent(13) || Helpers.GetSpellCooldown("A Murder of Crows").TotalSeconds > 5 || Helpers.GetSpellCooldown("A Murder of Crows").TotalMilliseconds < 1500);

        }
        private static bool ShouldCobraShotPvP()
        {
            if (Target == null)
                return false;
            return /*(Helpers.GetSpellCooldown("Kill Command").TotalSeconds > Math.Round(((Spell.GetSpellCastTime("Cobra Shot").TotalSeconds / ((100 + Helpers.HasteBuffs()) / 100)) - 0.4), 2) || CLua.PlayerPower < 20 || (!Me.HasAura("The Beast Within") && CLua.PlayerPower < 40))
                    && (!TalentManager.HasTalent(10) || (Helpers.GetSpellCooldown("Fervor").TotalMilliseconds > 750 && Helpers.GetSpellCooldown("Fervor").TotalSeconds < 29));

        }
        private static bool CanBeTrapped(WoWUnit u)
        {
            if (u == null)
                return false;

            return Lists.Immune(u) <= 250 && Lists.ImmuneSpell(u) <= 250 && (TalentManager.HasGlyph("Solace") || Lists.Dotted(u) <= 500);

        }
        private static bool SafeToTrap(WoWUnit u)
        {
            if (u == null)
                return false;
            return Lists.Rooted(u) >= 500 || u.HasAnyAura("Scatter Shot", "Lullaby", "Bad Manners") || u.HasAura(136634) || (Lists.CCd(u) <= 1500 && !u.IsMoving);
        }
        private static bool ShouldMisdirect()
        {
            if (Target == null)
                return false;
            if (Me.Specialization == WoWSpec.HunterBeastMastery)
            {
                switch (MainSettings.MisdirectWhen)
                {
                    case MisdirectWhen.Never:
                        return false;
                    case MisdirectWhen.InsecurelyTanking:
                        var tankable = Helpers.NearbyUnfriendlyUnitsInCombatWithMe.FirstOrDefault(u => u.Location.Distance2DSqr(Target.Location) < 25 * 25);
                        return (tankable != null && tankable.ThreatInfo.RawPercent < 90) || MDTarget().GetThreatInfoFor(Target).RawPercent < 90;
                    case MisdirectWhen.Always:
                        return true;
                }
            }
            if (Me.Specialization == WoWSpec.HunterSurvival)
            {
                switch (MainSettingsSV.MisdirectWhen)
                {
                    case MisdirectWhen.Never:
                        return false;
                    case MisdirectWhen.InsecurelyTanking:
                        var tankable = Helpers.NearbyUnfriendlyUnitsInCombatWithMe.FirstOrDefault(u => u.Location.Distance2DSqr(Target.Location) < 25 * 25);
                        return (tankable != null && tankable.ThreatInfo.RawPercent < 90) || MDTarget().GetThreatInfoFor(Target).RawPercent < 90;
                    case MisdirectWhen.Always:
                        return true;
                }
            }
            return false;
        }

        private static bool ShouldFeignDeathPvP()
        {
            if (Unit.ValidUnit(Target, true, false) && Me.HealthPercent < MainSettings.FeignHealthPvP)
            {
                switch (MainSettings.PvPFeignDeath)
                {
                    case PvPFeignDeath.PetNear:
                        return Target.GotAlivePet && Target.Pet.CurrentTargetGuid == Me.Guid && Target.Pet.Distance < 10;

                    case PvPFeignDeath.Casting:
                        return Target.IsPlayer && Target.GotTarget && Target.CurrentTargetGuid == Me.Guid && Target.IsCasting && (Lists.InterruptCastNoChannel(Target) > 200 || Lists.InterruptCastChannel(Target) > 200) && WoWSpell.FromId(Target.CastingSpellId).SpellEffect1.EffectType != WoWSpellEffectType.Heal;

                    case PvPFeignDeath.Targetted:
                        return Target.GotTarget && Target.CurrentTargetGuid == Me.Guid
                            && ((Unit.GetClassSpec(Target) == ClassType.Melee && Target.Distance < 10)
                            || ((Unit.GetClassSpec(Target) == ClassType.PRanged || Unit.GetClassSpec(Target) == ClassType.MRanged) && Target.Distance < 40));

                    case PvPFeignDeath.Casting_Pet:
                        return ((Target.GotAlivePet && Target.Pet.CurrentTargetGuid == Me.Guid && Target.Pet.Distance < 10) || (Target.IsPlayer && Target.CurrentTargetGuid == Me.Guid && Target.IsCasting && (Lists.InterruptCastNoChannel(Target) > 200 || Lists.InterruptCastChannel(Target) > 200) && WoWSpell.FromId(Target.CastingSpellId).SpellEffect1.EffectType != WoWSpellEffectType.Heal));

                    case PvPFeignDeath.Casting_Targetted:
                        return (Target.IsPlayer && Target.GotTarget && Target.CurrentTargetGuid == Me.Guid && (Lists.InterruptCastNoChannel(Target) > 200 || Lists.InterruptCastChannel(Target) > 200) && WoWSpell.FromId(Target.CastingSpellId).SpellEffect1.EffectType != WoWSpellEffectType.Heal)
                            ||
                            (Target.CurrentTargetGuid == Me.Guid
                            && ((Unit.GetClassSpec(Target) == ClassType.Melee && Target.Distance < 10)
                            || ((Unit.GetClassSpec(Target) == ClassType.PRanged || Unit.GetClassSpec(Target) == ClassType.MRanged) && Target.Distance < 40)));

                    case PvPFeignDeath.Pet_Targetted:
                        return (Target.GotAlivePet && Target.Pet.CurrentTargetGuid == Me.Guid && Target.Pet.Distance < 10)
                            ||
                            (Target.GotTarget && Target.CurrentTargetGuid == Me.Guid
                            && ((Unit.GetClassSpec(Target) == ClassType.Melee && Target.Distance < 10)
                            || ((Unit.GetClassSpec(Target) == ClassType.PRanged || Unit.GetClassSpec(Target) == ClassType.MRanged) && Target.Distance < 40)));

                    case PvPFeignDeath.Casting_Pet_Targetted:
                        return (Target.IsPlayer && Target.GotTarget && Target.CurrentTargetGuid == Me.Guid && (Lists.InterruptCastNoChannel(Target) > 200 || Lists.InterruptCastChannel(Target) > 200) && WoWSpell.FromId(Target.CastingSpellId).SpellEffect1.EffectType != WoWSpellEffectType.Heal)
                            ||
                            (Target.GotAlivePet && Target.Pet.CurrentTargetGuid == Me.Guid && Target.Pet.Distance < 10)
                            ||
                            (Target.GotTarget && Target.CurrentTargetGuid == Me.Guid
                            && ((Unit.GetClassSpec(Target) == ClassType.Melee && Target.Distance < 10)
                            || ((Unit.GetClassSpec(Target) == ClassType.PRanged || Unit.GetClassSpec(Target) == ClassType.MRanged) && Target.Distance < 40)));

                    case PvPFeignDeath.Never:
                        return false;
                }
            }
            return false;
        }

        private static bool ShouldSilencePvP(WoWUnit unit)
        {
            switch (MainSettings.SilencePvP)
            {
                case SilencePvP.Healer:
                    return unit.IsPlayer && Unit.GetClassSpec(unit) == ClassType.Healer;

                case SilencePvP.Casting:
                    return unit.IsPlayer && (Lists.InterruptCastNoChannel(Target) > 200 || Lists.InterruptCastChannel(Target) > 200);

                case SilencePvP.Casting_Healer:
                    return unit.IsPlayer && (Unit.GetClassSpec(unit) == ClassType.Healer || (Lists.InterruptCastNoChannel(Target) > 200 || Lists.InterruptCastChannel(Target) > 200));

                case SilencePvP.Never:
                    return false;
            }
            return false;
        }
        #endregion

        */

        #region Custom Units
        private static WoWUnit MDTarget()
        {
            switch (MainSettings.MisdirectOn)
            {
                case MisdirectOn.Me:
                    return StyxWoW.Me;
                case MisdirectOn.Focus:
                    return Unit.ValidFocus(false, true) ? Focus : null;
                case MisdirectOn.Pet:
                    return Me.GotAlivePet ? Pet : null;
                case MisdirectOn.Nobody:
                    return null;

            }
            return null;
        }
        #endregion

        #region Advanced Composites
        /*//Credits to Singular
        private static WoWUnit _ccUnit;
        internal static Composite FreezingTrapCrowdControl()
        {
            return new PrioritySelector(
                new Action(ret =>
                    {
                        _ccUnit = Focus;
                    return RunStatus.Failure;
                }),
                new Decorator(ret => Focus != null,
                    new PrioritySelector(
                //Blow him away if Glyphed
                     Spell.CreateTrapBehavior("Explosive Trap", true, on => Focus, ret => TalentManager.HasGlyph("Explosive Trap") && Focus.HasMyCachedAura("Freezing Trap")),
                //Lullaby + Freezing Trap
                    new Sequence(
                        PetManager.CreateCastPetActionOn("Lullaby", ret => Focus, ret => !Focus.HasAnyCachedAura(new[] {"Lullaby", "Scatter Shot"})),
                        Spell.CreateTrapBehavior("Freezing Trap", true, ret => Focus, ret => SafeToTrap(Focus))),
                //Wyvern Sting + Freezing Trap
                    new Sequence(
                        Spell.Buff("Wyvern Sting", false, ret => Focus, ret => Lists.Dotted(Focus) <= 0, "Lullaby", "Scatter Shot"),
                        Spell.CreateTrapBehavior("Freezing Trap", true, ret => Focus, ret => SafeToTrap(Focus))),
                //Scatter Shot + Freezing Trap
                    new Sequence(
                        Spell.Buff("Scatter Shot", false, ret => Focus, ret => !Focus.HasAnyAura("Lullaby", "Scatter Shot", "Binding Shot")),
                        Spell.CreateTrapBehavior("Freezing Trap", true, ret => Focus, ret => SafeToTrap(Focus))),
                //Blow him away if Glyphed
                     Spell.CreateTrapBehavior("Explosive Trap", true, on => Focus, ret => TalentManager.HasGlyph("Explosive Trap") && Focus.HasMyAura("Freezing Trap"))
                        ))
                   
                );
        }       
        */

        #endregion

        #region LocationContext
        internal static bool ContextEventSubscribed;

        internal static WoWContext LastWoWContext { get; set; }

        public static WoWContext CurrentWoWContext
        {
            get
            {
                Map map = StyxWoW.Me.CurrentMap;

                if (!StyxWoW.IsInGame)
                    return WoWContext.None;

                if ((map.IsBattleground || map.IsArena || MainSettings.Instance.ForcePvP
                || (MainSettings.Instance.PlayerPvP && StyxWoW.Me.GotTarget && StyxWoW.Me.CurrentTarget.IsPlayer && Unit.ValidUnit(StyxWoW.Me.CurrentTarget, true, false))) && StyxWoW.Me.Specialization != WoWSpec.HunterSurvival)
                {
                    return WoWContext.PvP;
                }

                return WoWContext.PvE;
            }
        }

        public static event EventHandler<WoWContextEventArg> OnWoWContextChanged;

        public static void UpdateContext()
        {
            // Subscribe to the map change event, so we can automatically update the context.
            if (!ContextEventSubscribed)
            {
                // Subscribe to OnBattlegroundEntered. Just 'cause.
                BotEvents.Battleground.OnBattlegroundEntered += e => UpdateContext();
                ContextEventSubscribed = true;
            }

            WoWContext current = CurrentWoWContext;

            // Can't update the context when it doesn't exist.
            if (current == WoWContext.None)
            {
                 Logger.DebugLog("WowContext doesn't exist.");
                return;
            }

            if (current == LastWoWContext || OnWoWContextChanged == null) return;
            try
            {
                OnWoWContextChanged(Instance, new WoWContextEventArg(current, LastWoWContext));
            }
            catch (Exception e)
            {
                // Eat any exceptions thrown.
                 Logger.DebugLog(e.ToString());
            }
            LastWoWContext = current;
        }
        #endregion

        #region RotationSelector
        public class RotationSelector
        {
            internal static Composite CreateCombat;
            internal static Composite CreatePreCombat;

            /// <summary>
            /// Creates our behaviors
            /// </summary>
            public static void RebuildBehaviors()
            {
                switch (CurrentWoWContext)
                {
                    #region WoWContext.PvP

                    case WoWContext.PvP:
                        switch (Me.Specialization)
                        {
                            case WoWSpec.HunterBeastMastery:
                                CreateCombat = new BeastMasteryPvP().CreateCombat;
                                CreatePreCombat = new BeastMasteryPvP().CreatePrecombat;
                                if (CreateCombat != null && CreatePreCombat != null)
                                    Logger.InitLog("PvP Specialization - Beast Mastery Activated");
                                break;
                            case WoWSpec.HunterSurvival:
                                CreateCombat = new SurvivalPvP().CreateCombat;
                                CreatePreCombat = new SurvivalPvP().CreatePrecombat;
                                if (CreateCombat != null && CreatePreCombat != null)
                                    Logger.InitLog("PvP Specialization - Survival Activated");
                                break;
                            default:
                                throw new ArgumentOutOfRangeException();
                        }
                        break;

                        #endregion

                        #region WoWContext.PvE

                    case WoWContext.PvE:
                        switch (Me.Specialization)
                        {
                            case WoWSpec.HunterBeastMastery:
                                CreateCombat = new BeastMasteryPvE().CreateCombat;
                                CreatePreCombat = new BeastMasteryPvE().CreatePrecombat;
                                if (CreateCombat != null && CreatePreCombat != null)
                                    Logger.InitLog("PvE Specialization - Beast Mastery Activated");
                                break;
                            case WoWSpec.HunterSurvival:
                                CreateCombat = new SurvivalPvE().CreateCombat;
                                CreatePreCombat = new SurvivalPvE().CreatePrecombat;
                                if (CreateCombat != null && CreatePreCombat != null)
                                    Logger.InitLog("PvE Specialization - Survival Activated");
                                break;
                            default:
                                throw new ArgumentOutOfRangeException();
                        }
                        break;
                }

                #endregion
            }
        }
        #endregion
    }
}
/* Mirabises Changes on V2.0.0  
 *  + Cached Auras on all units
 *  + New Cast Method for Beast Mastery to test
 *  + Overal code cleanup
 *  + Fixed broken GUI Settings
 *  + Made some methods to be called as extension
 *  + Removed the huge spell queeing system, to just use latency
 *  + Correct links to thread and change log in gui's now
 *  + Seperated the booleans & Removed redundant specializationchecks
 *  + Seperated
 *  - Broke ur Interface...
 *  - Still have to convert all strings to id's
 *  + AutoPause behavior
 *  + Removed redundant checks
 */


/* TO DO List
 * More intelligent Crowd Controlling, especially in arenas
 * Chain more CC's together.  
 * 
 * Smart Concussive Shot (PvP)
 * - Look at our Target and Focus and determine which is a better target for Conc shot.
 * 
 * Add features to Hotkeys
 * - Use cooldowns, Well timed silence, Trap on target location, Use Stun/Daze/Snare. Ask for communitys opinnion.
 * 
 * Add option to use AoE rotation in PvP
 * - Option for Never, Always and Only on NPC's
 * 
 * Improve requirments for all traps.
 * - Especially benefitial for Survival.
 * 
 * Research more spells to dispell with Tranq Shot (HoT's etc), and to Reflect with Deterrence
 * 
 * Improve support for some spells (such as Powershot, Barrage, Wyvern Sting)
 * 
 * Clean out and optimize code
 * - see what needs optimizing. -> Mirabiiiss
 * 
 * Premade presets
 * - Add premade presets
 * 
 * Add support for more pet spells (Hearth of the Phoenix, Bullheaded, Dash)
 */