﻿#region Author Information

// // Mark I - Items.cs
// // Created on 16/04/2013 at 13:42 Uhr
// // Author: Mirabis - Moreno Sint Hill
// 

#endregion

using System;
using System.Collections.Generic;
using System.Linq;
using Bestia_Magister.Managers;
using Bestia_Magister.Settings;
using Styx;
using Styx.CommonBot.Inventory;
using Styx.TreeSharp;
using Styx.WoWInternals;
using Styx.WoWInternals.WoWObjects;

namespace Bestia_Magister.Misc
{
    public static class Items
    {
        private static LocalPlayer Me
        {
            get { return StyxWoW.Me; }
        }

        private static WoWUnit Target
        {
            get { return StyxWoW.Me.CurrentTarget; }
        }

        /// <summary>
        /// Proffesions,Battlestandard,Healthstones,Potions
        /// </summary>
        public static PrioritySelector UseItems
        {
            get
            {
                return new PrioritySelector(
                    new Decorator(ret => Me.Specialization == WoWSpec.HunterBeastMastery,
                                  new PrioritySelector(
                                      UseEngineerGloves(),
                                      UseEngineerBelt(),
                                      CastManager.BuffSelf("Every Man for Himself",
                                                     ret =>
                                                     Me.Race == WoWRace.Human && CooldownSettings.Instance.Racial &&
                                                     Lists.CCd(Me) >= 900),
                                      UseTrinkets(),
                                      UsePotions(),
                                      UseBattleStandard())),
                    new Decorator(ret => Me.Specialization == WoWSpec.HunterSurvival,
                                  new PrioritySelector(
                                      UseEngineerGlovesSV(),
                                      UseEngineerBeltSV(),
                                      CastManager.BuffSelf("Every Man for Himself",
                                                     ret =>
                                                     Me.Race == WoWRace.Human && CooldownSettingsSV.Instance.Racial &&
                                                     Lists.CCd(Me) >= 900),
                                      UseTrinketsSV(),
                                      UsePotionsSV(),
                                      UseBattleStandardSV()))
                    );
            }
        }

        /// <summary>
        /// Uses the item specified (simulated click)
        /// </summary>
        /// <param name="item">The item to use</param>
        private static void UseItem(WoWItem item)
        {
            if (item != null)
            {
                item.Use();
            }
        }

        /// <summary>Creates a behavior to use an equipped item.</summary>
        /// <param name="slot">The slot number of the equipped item. </param>
        /// <returns>success or failure</returns>
        private static Composite UseEquippedItem(uint slot)
        {
            return new PrioritySelector(
                ctx => StyxWoW.Me.Inventory.GetItemBySlot(slot),
                new Decorator(
                    ctx => ctx != null && CanUseEquippedItem((WoWItem) ctx),
                    new Styx.TreeSharp.Action(ctx => UseItem((WoWItem) ctx))));
        }

        /// <summary>
        ///     Checks for items in the bag, and returns the first item that has an usable spell from the specified string array.
        /// </summary>
        /// <param name="spellNames"> Array of spell names to be check.</param>
        /// <returns></returns>
        private static WoWItem FindFirstUsableItemBySpell(params string[] spellNames)
        {
            List<WoWItem> carried = StyxWoW.Me.CarriedItems;
            // Yes, this is a bit of a hack. But the cost of creating an object each call, is negated by the speed of the Contains from a hash set.
            // So take your optimization bitching elsewhere.
            var spellNameHashes = new HashSet<string>(spellNames);


            return
                carried.Select(i => new {i, spells = i.ItemSpells})
                       .Where(
                           @t =>
                           t.i.ItemInfo != null && t.spells != null && t.spells.Count != 0 && t.i.Usable &&
                           t.i.Cooldown.Equals(0) && t.i.ItemInfo.RequiredLevel <= StyxWoW.Me.Level &&
                           t.spells.Any(
                               s => s.IsValid && s.ActualSpell != null && spellNameHashes.Contains(s.ActualSpell.Name)))
                       .OrderByDescending(@t => t.i.ItemInfo.Level)
                       .Select(t1 => t1.i)
                       .FirstOrDefault();
        }

        private static Composite UseEquippedTrinket(TrinketUsage usage)
        {
            return new PrioritySelector(
                new Decorator(
                    ret => usage == MainSettings.Instance.Trinket1Usage,
                    new PrioritySelector(
                        ctx => StyxWoW.Me.Inventory.GetItemBySlot((uint) WoWInventorySlot.Trinket1),
                        new Decorator(
                            ctx => ctx != null && CanUseEquippedItem((WoWItem) ctx),
                            new Styx.TreeSharp.Action(ctx => UseItem((WoWItem) ctx))
                            )
                        )
                    ),
                new Decorator(
                    ret => usage == MainSettings.Instance.Trinket2Usage,
                    new PrioritySelector(
                        ctx => StyxWoW.Me.Inventory.GetItemBySlot((uint) WoWInventorySlot.Trinket2),
                        new Decorator(
                            ctx => ctx != null && CanUseEquippedItem((WoWItem) ctx),
                            new Styx.TreeSharp.Action(ctx => UseItem((WoWItem) ctx))
                            )
                        )
                    )
                );
        }

        private static Composite UseEquippedTrinketSV(TrinketUsage usage)
        {
            return new PrioritySelector(
                new Decorator(
                    ret => usage == MainSettingsSV.Instance.Trinket1Usage,
                    new PrioritySelector(
                        ctx => StyxWoW.Me.Inventory.GetItemBySlot((uint) WoWInventorySlot.Trinket1),
                        new Decorator(
                            ctx => ctx != null && CanUseEquippedItem((WoWItem) ctx),
                            new Styx.TreeSharp.Action(ctx => UseItem((WoWItem) ctx))
                            )
                        )
                    ),
                new Decorator(
                    ret => usage == MainSettingsSV.Instance.Trinket2Usage,
                    new PrioritySelector(
                        ctx => StyxWoW.Me.Inventory.GetItemBySlot((uint) WoWInventorySlot.Trinket2),
                        new Decorator(
                            ctx => ctx != null && CanUseEquippedItem((WoWItem) ctx),
                            new Styx.TreeSharp.Action(ctx => UseItem((WoWItem) ctx))
                            )
                        )
                    )
                );
        }

        private static Composite UseRefreshment(Refreshment usage)
        {
            return new PrioritySelector(
                new Decorator(
                    ret => usage == Refreshment.Drink,
                    new PrioritySelector(
                        ctx => Consumable.GetBestDrink(true),
                        new Decorator(
                            ctx => ctx != null && CanUseEquippedItem((WoWItem) ctx),
                            new Styx.TreeSharp.Action(ctx => UseItem((WoWItem) ctx))
                            )
                        )
                    ),
                new Decorator(
                    ret => usage == Refreshment.Food,
                    new PrioritySelector(
                        ctx => Consumable.GetBestFood(true),
                        new Decorator(
                            ctx => ctx != null && CanUseEquippedItem((WoWItem) ctx),
                            new Styx.TreeSharp.Action(ctx => UseItem((WoWItem) ctx))
                            )
                        )
                    ),
                new Decorator(
                    ret => usage == Refreshment.Flask,
                    new PrioritySelector(
                        ctx => FindFirstUsableItemBySpell("TODO", "TODO"),
                        new Decorator(
                            ctx =>
                            ctx != null && CanUseEquippedItem((WoWItem) ctx) && !StyxWoW.Me.HasAura(ctx.ToString()),
                            new Styx.TreeSharp.Action(ctx => UseItem((WoWItem) ctx))
                            )
                        )
                    )
                );
        }

        private static Composite UsePotions()
        {
            return new PrioritySelector(
                new Decorator(
                    ret => StyxWoW.Me.HealthPercent < MainSettings.Instance.SHealthstone,
                    new PrioritySelector(
                        ctx => FindFirstUsableItemBySpell("Healthstone"),
                        new Decorator(
                            ret => ret != null,
                            new Sequence(
                                new Styx.TreeSharp.Action(
                                    ctx => Logger.ItemLog("Health Stone used  at {0}", StyxWoW.Me.HealthPercent)),
                                new Styx.TreeSharp.Action(ret => ((WoWItem) ret).UseContainerItem()))))),
                new Decorator(
                    ret => StyxWoW.Me.HealthPercent <= MainSettings.Instance.PotionHealth,
                    new PrioritySelector(
                        ctx => FindFirstUsableItemBySpell("Life Spirit", "Healing Potion", "Alchemist's Rejuvenation"),
                        new Decorator(
                            ctx => ctx != null,
                            new Sequence(
                                new Styx.TreeSharp.Action(
                                    ctx => Logger.ItemLog("Health Potion used at {0}", StyxWoW.Me.HealthPercent)),
                                new Styx.TreeSharp.Action(ctx => ((WoWItem) ctx).UseContainerItem()))))),
                new Decorator(
                    ret =>
                    Unit.ValidUnit(Target, true, false) && Target.HealthPercent <= MainSettings.Instance.VirmenHealth &&
                    MainSettings.Instance.UseVirminPot,
                    new PrioritySelector(
                        ctx => FindFirstUsableItemBySpell("Virmen's Bite"),
                        new Decorator(
                            ctx =>
                            ctx != null && (Lists.TargetIsBoss(Target) || Lists.TrainingDummies.Contains(Target.Entry)) &&
                            Me.HasAnyCachedAura(new[] {"Ancient Hysteria", "Time Warp", "Bloodlust", "Heroism"}),
                            new Sequence(
                                new Styx.TreeSharp.Action(
                                    ctx => Logger.ItemLog("Virmen's Bite used at {0}", StyxWoW.Me.HealthPercent)),
                                new Styx.TreeSharp.Action(ctx => ((WoWItem) ctx).UseContainerItem())))))
                );
        }

        private static Composite UsePotionsSV()
        {
            return new PrioritySelector(
                new Decorator(
                    ret => StyxWoW.Me.HealthPercent < MainSettingsSV.Instance.SHealthstone,
                    new PrioritySelector(
                        ctx => FindFirstUsableItemBySpell("Healthstone"),
                        new Decorator(
                            ret => ret != null,
                            new Sequence(
                                new Styx.TreeSharp.Action(
                                    ctx => Logger.ItemLog("Health Stone used  at {0}", StyxWoW.Me.HealthPercent)),
                                new Styx.TreeSharp.Action(ret => ((WoWItem) ret).UseContainerItem()))))),
                new Decorator(
                    ret => StyxWoW.Me.HealthPercent <= MainSettingsSV.Instance.PotionHealth,
                    new PrioritySelector(
                        ctx => FindFirstUsableItemBySpell("Life Spirit", "Healing Potion", "Alchemist's Rejuvenation"),
                        new Decorator(
                            ctx => ctx != null,
                            new Sequence(
                                new Styx.TreeSharp.Action(
                                    ctx => Logger.ItemLog("Health Potion used at {0}", StyxWoW.Me.HealthPercent)),
                                new Styx.TreeSharp.Action(ctx => ((WoWItem) ctx).UseContainerItem()))))),
                new Decorator(
                    ret =>
                    Unit.ValidUnit(Target, true, false) && Target.HealthPercent <= MainSettingsSV.Instance.VirmenHealth &&
                    MainSettingsSV.Instance.UseVirminPot,
                    new PrioritySelector(
                        ctx => FindFirstUsableItemBySpell("Virmen's Bite"),
                        new Decorator(
                            ctx =>
                            ctx != null && (Lists.TargetIsBoss(Target) || Lists.TrainingDummies.Contains(Target.Entry)) &&
                            Me.HasAnyCachedAura(new[] {"Ancient Hysteria", "Time Warp", "Bloodlust", "Heroism"}),
                            new Sequence(
                                new Styx.TreeSharp.Action(
                                    ctx => Logger.ItemLog("Virmen's Bite used at {0}", StyxWoW.Me.HealthPercent)),
                                new Styx.TreeSharp.Action(ctx => ((WoWItem) ctx).UseContainerItem())))))
                );
        }

        private static Composite UseEngineerGloves()
        {
            return new PrioritySelector(
                new Decorator(ret => CooldownSettings.Instance.Gloves, UseEquippedItem(9))
                );
        }

        private static Composite UseEngineerBelt()
        {
            return new PrioritySelector(
                new Decorator(ret => CooldownSettings.Instance.Belt && Target.CurrentHealth > 40000, UseEquippedItem(6))
                );
        }

        private static Composite UseTrinkets()
        {
            return new PrioritySelector(
                new Decorator(ret => true, UseEquippedTrinket(TrinketUsage.Burst)),
                new Decorator(ret => Me.HealthPercent < MainSettings.Instance.TrinketHealth,
                              UseEquippedTrinket(TrinketUsage.LowHealth)),
                new Decorator(ret => Lists.CCd(Me) > MainSettings.Instance.CCTrinketTime,
                              UseEquippedTrinket(TrinketUsage.CrowdControlled))
                );
        }

        private static Composite UseEngineerGlovesSV()
        {
            return new PrioritySelector(
                new Decorator(ret => CooldownSettingsSV.Instance.Gloves, UseEquippedItem(9))
                );
        }

        private static Composite UseEngineerBeltSV()
        {
            return new PrioritySelector(
                new Decorator(ret => CooldownSettingsSV.Instance.Belt && Target.CurrentHealth > 40000,
                              UseEquippedItem(6))
                );
        }

        private static Composite UseTrinketsSV()
        {
            return new PrioritySelector(
                new Decorator(ret => true, UseEquippedTrinket(TrinketUsage.Burst)),
                new Decorator(ret => Me.HealthPercent < MainSettingsSV.Instance.TrinketHealth,
                              UseEquippedTrinketSV(TrinketUsage.LowHealth)),
                new Decorator(ret => Lists.CCd(Me) > MainSettingsSV.Instance.CCTrinketTime,
                              UseEquippedTrinketSV(TrinketUsage.CrowdControlled))
                );
        }

        private static Composite UseBattleStandard()
        {
            return new PrioritySelector(
                );
        }

        private static Composite UseBattleStandardSV()
        {
            return new PrioritySelector(
                );
        }


        /// <summary>
        /// Checks the equiped item cooldown is 0 and its usable
        /// </summary>
        /// <param name="item">the item to query</param>
        /// <returns>true of false</returns>
        public static bool CanUseEquippedItem(WoWItem item)
        {
            try
            {
                // Check for engineering tinkers!
                using (StyxWoW.Memory.AcquireFrame())
                {
                    var itemSpell = Lua.GetReturnVal<string>("return GetItemSpell(" + item.Entry + ")", 0);
                    if (String.IsNullOrEmpty(itemSpell))
                        return false;

                    return item.Usable && item.Cooldown <= 0;
                }
            }
            catch
            {
                return false;
            }
        }
    }
}