﻿#region Author Information

// // Mark I - Unit.cs
// // Created on 16/04/2013 at 13:44 Uhr
// // Author: Mirabis - Moreno Sint Hill
// 

#endregion

using Styx;
using Styx.WoWInternals.WoWObjects;

namespace Bestia_Magister.Misc
{
    public static class Unit
    {
        private static LocalPlayer Me
        {
            get { return StyxWoW.Me; }
        }

        #region IsEnemy

        public static bool IsEnemy(WoWUnit u)
        {
            if (u != null)
            {
                //   if (!Me.IsFFAPvPFlagged)
                if (!StyxWoW.Me.CurrentMap.IsArena && !StyxWoW.Me.CurrentMap.IsBattleground)
                {
                    if (!u.IsFriendly)
                    {
                        return true;
                    }
                    return false;
                }
                //if (!UnitIsInMyPartyOrRaid(target))
                return !IsMyPartyRaidMember(u);
            }
            return false;
        }

        #endregion

        #region ValidUnit

        private static bool ValidUnit(this WoWUnit u)
        {
            return u != null && u.IsValid && !u.IsDisabled && u.IsAlive && u.CanSelect && !u.IsNonCombatPet;
        }

        public static bool ValidUnit(WoWUnit unit, bool enemy, bool friendly)
        {
            if (unit != null && ValidUnit(unit) && Lists.Immune(unit) <= 0)
            {
                if (enemy)
                {
                    return IsEnemy(unit);
                }
                if (friendly)
                {
                    return !IsEnemy(unit);
                }
            }
            return false;
        }

        public static bool ValidFocus(bool enemy, bool friendly)
        {
            var unit = Me.FocusedUnit;

            if (unit != null && ValidUnit(unit) && Lists.Immune(unit) <= 0)
            {
                if (enemy)
                {
                    return IsEnemy(unit);
                }
                if (friendly)
                {
                    return !IsEnemy(unit);
                }
            }
            return false;
        }

        #endregion

        #region GetClassSpec

        public static ClassType GetClassSpec(this WoWUnit target)
        {
            if (target == null)
            {
                return ClassType.Empty;
            }

            switch (target.Class)
            {
                case WoWClass.Monk: //TODO: check for monk
                    return ClassType.Healer;
                case WoWClass.DeathKnight:
                    return ClassType.Melee;
                case WoWClass.Hunter:
                    return ClassType.PRanged;
                case WoWClass.Mage:
                    return ClassType.MRanged;
                case WoWClass.Paladin:
                    return target.MaxMana >= 80000 ? ClassType.Healer : ClassType.Melee;
                case WoWClass.Priest:
                    return target.Shapeshift == ShapeshiftForm.Shadow ? ClassType.MRanged : ClassType.Healer;
                case WoWClass.Rogue:
                    return ClassType.Melee;
                case WoWClass.Warlock:
                    return ClassType.MRanged;
                case WoWClass.Warrior:
                    return ClassType.Melee;
                case WoWClass.Shaman:
                    if (target.MaxMana < 40000)
                        return ClassType.Melee;
                    if (target.Buffs.ContainsKey("Elemental Oath") &&
                        target.Buffs["Elemental Oath"].CreatorGuid == target.Guid)
                        return ClassType.MRanged;
                    return ClassType.Healer;
                case WoWClass.Druid:
                    if (target.Shapeshift == ShapeshiftForm.Moonkin)
                        return ClassType.MRanged;
                    if ((target.Buffs.ContainsKey("Leader of the Pack") &&
                         target.Buffs["Leader of the Pack"].CreatorGuid == target.Guid) || target.MaxMana < 40000)
                        return ClassType.Melee;
                    return ClassType.Healer;
            }
            return ClassType.Empty;
        }

        #endregion

        #region IsTargetingUs

        /// <summary>
        /// checks if unit is targeting you, your minions, a group member, or group pets
        /// </summary>
        /// <param name="u">unit</param>
        /// <returns>true if targeting your guys, false if not</returns>
        public static bool IsTargetingUs(this WoWUnit u)
        {
            return u.IsTargetingMeOrPet || u.IsTargetingAnyMinion || u.IsTargetingMyPartyMember ||
                   u.IsTargetingMyRaidMember;
        }

        #endregion

        #region IsMyPartyRaidMember

        /// <summary>
        /// Returns whether the unit is in my party or their owner is.
        /// </summary>
        /// <value>My Party/Raid Members.</value>
        public static bool IsMyPartyRaidMember(this WoWUnit u)
        {
            if (u == null || !u.IsValid)
            {
                return false;
            }
            if (u.IsPlayer)
            {
                var player = u.ToPlayer();
                if (player.IsInMyPartyOrRaid || u.IsMe)
                    return true;
            }
            else if (u.IsPet)
            {
                var owner = u.OwnedByUnit.ToPlayer();
                if (owner.IsInMyPartyOrRaid || u.IsMe)
                    return true;
            }
            return false;
        }

        #endregion
    }
}