﻿#region Author Information

// // Mark I - BeastMasteryPvP.cs
// // Created on 12/04/2013 at 19:05 Uhr
// // Author: Mirabis - Moreno Sint Hill
// 

#endregion

using System;
using System.Linq;
using System.Windows.Forms;
using Bestia_Magister.Managers;
using Bestia_Magister.Misc;
using Bestia_Magister.Settings;
using CommonBehaviors.Actions;
using Styx;
using Styx.CommonBot;
using Styx.TreeSharp;
using Styx.WoWInternals;
using Styx.WoWInternals.WoWObjects;
using Action = Styx.TreeSharp.Action;
using CM = Bestia_Magister.Managers.CastManager;

namespace Bestia_Magister.Rotations
{
    public class BeastMasteryPvP
    {
        #region Priority : Combat Selector

        private static Composite CombatSelector
        {
            get
            {
                return new PrioritySelector(
                    Helpers.TreePerformance("BeastMasteryPvP"),
                    AutoPause(),
                    new Decorator(ret => !HotkeyManager.Registered, new Action(ret => HotkeyManager.RegisterKeys())),
                    new Throttle(1, TimeSpan.FromSeconds(1), Helpers.AcquireTargets),
                    SetPetActions, //Macro Disablers etc
                    CM.WaitForGcdOrCastOrChannel(), // Return if Casting/Channeling ( Traverse Successor )
                    new Decorator(req => Unit.ValidUnit(Target, true, false), //Valid Target
                                  new PrioritySelector(
                                      new Throttle(Items.UseItems),
                                      PetManager.CreateHunterCallPetBehavior(),
                                      DefensiveCooldowns,
                                      CrowdControl(when => true),
                                      HotkeyBehavior(when =>(Helpers.IsKeyDown(MainSettings.Hotkeys.Instance.CCModifierChoice) ||MainSettings.Hotkeys.Instance.CCModifierChoice == Keys.None) &&
                                          Helpers.IsKeyDown(MainSettings.Hotkeys.Instance.CCModeKeyChoice)),
                                      PetManager.CreateCastPetActionOn("Attack", on => Target),
                                      Interrupt(req => true),
                                      FocusRegen(when => HotkeyManager.RegenMode && Lists.BreakOnDamage(Target) < 250),
                                      //Normally we go down here ; )
                                      new Decorator(when => !HotkeyManager.RegenMode,
                                                    new PrioritySelector(
                                                        OffensiveCooldowns(when => Lists.BreakOnDamage(Target) < 250),
                                                        TrapBehavior(when => true),
                                                        AreaofEffect(when =>HotkeyManager.AoEMode || Helpers.AddCount(25) >= MainSettings.AddCount),
                                                        SingleTarget))
                                      )));
            }
        }

        #endregion

        #region Priority : Pre Combat Selector

        private static Composite PreCombatSelector
        {
            get
            {
                return new PrioritySelector(
                    PetManager.CreateHunterCallPetBehavior(),
                    CM.Cast("Hunter's Mark", ret => Target,
                               ret =>
                               MainSettings.HuntersMarkOOC && Target.Distance > 40 && Target.Distance < 100 &&
                               (Me.CurrentMap.IsArena || Me.CurrentMap.IsBattleground) &&
                               !Target.HasMyCachedAura("Hunter's Mark"))
                    );
            }
        }

        #endregion

        //Finished

        #region Rotation Handler : Defensive Cooldowns

        private static Composite DefensiveCooldowns
        {
            get
            {
                return new PrioritySelector(
                    CM.BuffSelf("Gift of the Naaru",
                                   ret => CooldownSettings.Racial && Me.Race == WoWRace.Draenei && Me.HealthPercent < 68),
                    // Dranei racial Heal
                    CM.BuffSelf("Exhilaration",
                                   ret =>
                                   TalentManager.HasTalent(7) && MainSettings.Talent45 &&
                                   (Me.HealthPercent < 67 || (Me.GotAlivePet && Pet.HealthPercent < 15))),
                    CM.BuffSelf("Mend Pet",
                                   ret =>
                                   PetSettings.MendPet && Me.GotAlivePet && !Pet.HasAura("Mend Pet") &&
                                   Pet.HealthPercent < PetSettings.PetHealth),
                    CM.BuffSelf("Master's Call",
                                   ret =>
                                   Me.GotAlivePet &&
                                   (MainSettings.MastersCall == MastersCall.Me ||
                                    MainSettings.MastersCall == MastersCall.Me_Partner) && Lists.Snared(Me) >= 250 ||
                                   Lists.Rooted(Me) >= 250),
                    CM.BuffSelf("Deterrence",
                                   ret =>
                                   MainSettings.DeterrenceReflect && TalentManager.HasGlyph("Mirrored Blades") &&
                                   Helpers.UnitsCastingReflectableSpell()),
                    CM.Buff("Master's Call", ret => Helpers.ArenaPartner,
                               ret =>
                               Me.CurrentMap.IsArena && Helpers.ArenaPartner != null && Me.GotAlivePet &&
                               (MainSettings.MastersCall == MastersCall.Partner ||
                                MainSettings.MastersCall == MastersCall.Me_Partner) &&
                               Lists.Snared(Helpers.ArenaPartner) >= 250 || Lists.Rooted(Helpers.ArenaPartner) >= 250),
                    PetManager.CreateCastPetActionOn("Roar of Sacrifice", ret => Me,
                                                     ret =>
                                                     (Me.HealthPercent < PetSettings.RoSMe || Lists.CCd(Me) > 500) &&
                                                     (PetSettings.RoarOfSacrifice == RoarOfSacrifice.Me ||
                                                      PetSettings.RoarOfSacrifice == RoarOfSacrifice.Me_Partner)),
                    PetManager.CreateCastPetActionOn("Roar of Sacrifice", ret => Helpers.ArenaPartner,
                                                     ret =>
                                                     Me.CurrentMap.IsArena && Helpers.ArenaPartner != null &&
                                                     (Helpers.ArenaPartner.HealthPercent < PetSettings.RoSPartner ||
                                                      Lists.CCd(Helpers.ArenaPartner) > 500) &&
                                                     (PetSettings.RoarOfSacrifice == RoarOfSacrifice.Partner ||
                                                      PetSettings.RoarOfSacrifice == RoarOfSacrifice.Me_Partner)),
                    PetManager.CreateCastPetActionOn("Spirit Mend", ret => Me,
                                                     ret =>
                                                     Me.HealthPercent < PetSettings.SpiritHealMe &&
                                                     (PetSettings.SpiritMend == SpiritMend.Me ||
                                                      PetSettings.SpiritMend == SpiritMend.Me_Focus ||
                                                      PetSettings.SpiritMend == SpiritMend.Me_Focus_Pet)),
                    PetManager.CreateCastPetActionOn("Spirit Mend", ret => Focus,
                                                     ret =>
                                                     Unit.ValidFocus(false, true) &&
                                                     Focus.HealthPercent < PetSettings.SpiritHealFocus &&
                                                     (PetSettings.SpiritMend == SpiritMend.Focus ||
                                                      PetSettings.SpiritMend == SpiritMend.Me_Focus ||
                                                      PetSettings.SpiritMend == SpiritMend.Me_Focus_Pet)),
                    PetManager.CreateCastPetActionOn("Spirit Mend", ret => Pet,
                                                     ret =>
                                                     Pet.HealthPercent < PetSettings.SpiritHealPet &&
                                                     (PetSettings.SpiritMend == SpiritMend.Pet ||
                                                      PetSettings.SpiritMend == SpiritMend.Me_Pet ||
                                                      PetSettings.SpiritMend == SpiritMend.Me_Focus_Pet)),
                    PetManager.CreateCastPetActionOn("Spirit Mend", ret => Helpers.ArenaPartner,
                                                     ret =>
                                                     Me.CurrentMap.IsArena && Helpers.ArenaPartner != null &&
                                                     Helpers.ArenaPartner.HealthPercent < PetSettings.SpiritHealPartner &&
                                                     (PetSettings.SpiritMend == SpiritMend.Partner ||
                                                      PetSettings.SpiritMend == SpiritMend.Me_Partner ||
                                                      PetSettings.SpiritMend == SpiritMend.Pet_Partner ||
                                                      PetSettings.SpiritMend == SpiritMend.Me_Pet_Partner)),
                    CM.BuffSelf("Feign Death", ret => NeedFeignDeath()),
                    CM.BuffSelf("Will of The Forsaken",
                                   ret =>
                                   Me.Race == WoWRace.Undead && CooldownSettings.Racial &&
                                   Helpers.isForsaken(Me).TotalSeconds > 1),
                    CM.BuffSelf("Stoneform",
                                   ret =>
                                   CooldownSettings.Racial && Me.Race == WoWRace.Dwarf &&
                                   StyxWoW.Me.GetAllAuras()
                                          .Any(
                                              a =>
                                              a.Spell.Mechanic == WoWSpellMechanic.Bleeding ||
                                              a.Spell.DispelType == WoWDispelType.Disease ||
                                              a.Spell.DispelType == WoWDispelType.Poison)) // Defensive cooldown
                    );
            }
        }

        #endregion

        //Finished

        #region Rotation Handler : OffensiveCooldowns

        private static Composite OffensiveCooldowns(Main.Selection<bool> reqs = null)
        {
            return new Decorator(when => reqs == null || reqs(when),
                                 new PrioritySelector(
                                     CM.BuffSelf("Focus Fire",
                                                    ret =>
                                                    Me.CachedStackCount("Frenzy") >= 5 &&
                                                    !Me.HasAura("The Beast Within")),
                                     CM.BuffSelf("Aspect of the Hawk",
                                                    ret =>
                                                    MainSettings.AspectHawk && !Me.IsMoving && !Me.HasAura(109260) &&
                                                    !Me.HasAura(13165)),
                                     CM.BuffSelf("Berserking",
                                                    ret => CooldownSettings.Racial && Me.Race == WoWRace.Troll),
                                     CM.BuffSelf("Blood Fury",
                                                    ret => CooldownSettings.Racial && Me.Race == WoWRace.Orc),
                                     CM.BuffSelf("Fervor",
                                                    ret =>
                                                    SpellManager.HasSpell("Fervor") && MainSettings.Talent60_FV &&
                                                    (Helpers.PlayerPower < 60 ||
                                                     Me.HasAura("The Beast Within") && Helpers.PlayerPower < 40)),
                                     CM.BuffSelf("Rapid Fire",
                                                    ret =>
                                                    CooldownSettings.RapidFire &&
                                                    ((Me.GotTarget && Target.CurrentHealth > 22000 && Target.IsPlayer) ||
                                                     Lists.TrainingDummies.Contains(Target.Entry))),
                                     CM.Cast("Stampede",
                                                ret =>
                                                CooldownSettings.Stampede && Target.Distance <= 40 &&
                                                ((Target.CurrentHealth > 22000 && Target.IsPlayer) ||
                                                 Lists.TrainingDummies.Contains(Target.Entry))),
                                     CM.BuffSelf("Bestial Wrath",
                                                    ret => CooldownSettings.BestialWrath && NeedBestialWrath()),
                                     CM.BuffSelf("Readiness", ret => NeedReadiness()),
                                     CM.BuffSelf("Lifeblood", ret => CooldownSettings.Lifeblood)
                                     ));
        }

        #endregion

        //Finished

        #region Rotation Handler : HotkeyBehavior

        private static Composite HotkeyBehavior(Main.Selection<bool> reqs = null)
        {
            return new Decorator(when => reqs == null || reqs(when),
                                 new PrioritySelector(
                                     //Focused Unit
                                     new Decorator(ret => Unit.ValidFocus(true, false),
                                                   new PrioritySelector(
                                                       CM.CreateTrapBehavior("Freezing Trap", true, ret => Focus,
                                                                                ret =>
                                                                                CanBeTrapped(Focus) &&
                                                                                (SafeToTrap(Focus) ||
                                                                                 Helpers.IsKeyDown(
                                                                                     MainSettings.Hotkeys.Instance
                                                                                                 .ForceTrapChoice)) &&
                                                                                Focus.HasCachedAuraTimeLeft(
                                                                                    "Freezing Trap", true).TotalSeconds <=
                                                                                1),
                                                       CM.Buff("Wyvern Sting", ret => Focus,
                                                                  ret =>
                                                                  TalentManager.HasTalent(5) &&
                                                                  !Focus.HasAura("Freezing Trap") &&
                                                                  Helpers.GetSpellCooldown("Freezing Trap").TotalSeconds >
                                                                  5),
                                                       CM.Cast("Scatter Shot", ret => Focus,
                                                                  ret =>
                                                                  !Helpers.SelfControl(Focus) && Lists.CCd(Focus) < 500 &&
                                                                  Lists.Rooted(Focus) < 500),
                                                       PetManager.CreateCastPetActionOn("Lullaby", ret => Focus,
                                                                                        ret =>
                                                                                        PetSettings.Lullaby &&
                                                                                        !Focus.HasAnyCachedAura(new[]
                                                                                            {"Lullaby", "Scatter Shot"}) &&
                                                                                        ((Helpers.GetSpellCooldown(
                                                                                            "Scatter Shot").TotalSeconds <
                                                                                          27 &&
                                                                                          Helpers.GetSpellCooldown(
                                                                                              "Scatter Shot")
                                                                                                 .TotalSeconds > 1)) &&
                                                                                        Pet.Location.Distance(
                                                                                            Focus.Location) <= 20 &&
                                                                                        Lists.CCd(Focus) < 500 &&
                                                                                        Lists.Rooted(Focus) < 500)
                                                       )),
                                     //Current Target
                                     new Decorator(
                                         ret => Unit.ValidUnit(Target, true, false) && !Unit.ValidFocus(true, false),
                                         new PrioritySelector(
                                             CM.CreateTrapBehavior("Freezing Trap", true, ret => Target,
                                                                      ret =>
                                                                      CanBeTrapped(Target) &&
                                                                      (SafeToTrap(Target) ||
                                                                       Helpers.IsKeyDown(
                                                                           MainSettings.Hotkeys.Instance.ForceTrapChoice)) &&
                                                                      Target.HasCachedAuraTimeLeft("Freezing Trap", true)
                                                                            .TotalSeconds <= 1),
                                             CM.Buff("Wyvern Sting", ret => Target,
                                                        ret =>
                                                        TalentManager.HasTalent(5) && !Target.HasAura("Freezing Trap") &&
                                                        Helpers.GetSpellCooldown("Freezing Trap").TotalSeconds > 5),
                                             CM.Cast("Scatter Shot", ret => Target,
                                                        ret =>
                                                        !Helpers.SelfControl(Target) && Lists.CCd(Target) < 500 &&
                                                        Lists.Rooted(Target) < 500),
                                             PetManager.CreateCastPetActionOn("Lullaby", ret => Target,
                                                                              ret =>
                                                                              PetSettings.Lullaby &&
                                                                              !Target.HasAnyCachedAura(new[]
                                                                                  {"Lullaby", "Scatter Shot"}) &&
                                                                              ((Helpers.GetSpellCooldown("Scatter Shot")
                                                                                       .TotalSeconds < 27 &&
                                                                                Helpers.GetSpellCooldown("Scatter Shot")
                                                                                       .TotalSeconds > 1)) &&
                                                                              Pet.Location.Distance(Target.Location) <=
                                                                              20 && Lists.CCd(Target) < 500 &&
                                                                              Lists.Rooted(Target) < 500)
                                             ))));
        }

        #endregion

        //Finished

        #region Rotation Handler : Crowd Control

        private static Composite CrowdControl(Main.Selection<bool> reqs = null)
        {
            return new Decorator(when => reqs == null || reqs(when),
                                 new PrioritySelector(
                                     //Current Target
                                     new Decorator(
                                         ret =>
                                         Unit.ValidUnit(Target, true, false) && Me.CurrentPendingCursorSpell == null,
                                         new PrioritySelector(
                                             CM.Cast("Tranquilizing Shot", ret => Target,
                                                        ret =>
                                                        MainSettings.Tranquilizing && Lists.CanBeTranquilized(Target, 2) &&
                                                        (Helpers.PlayerPower > 60 ||
                                                         Me.HasAnyCachedAura(new[]
                                                             {"The Beast Within", "Thrill of the Hunt"}) ||
                                                         TalentManager.HasGlyph("Tranquilizing Shot")) &&
                                                        Target.Distance <= 40),
                                             PetManager.CreateCastPetActionOn("Growl", ret => Helpers.EnemyPet,
                                                                              ret =>
                                                                              MainSettings.GrowlPets &&
                                                                              Helpers.EnemyPet != null),
                                             CM.Cast("Arcane Shot", ret => Helpers.ShamanTotem,
                                                        ret =>
                                                        Me.CurrentMap.IsArena && Helpers.ShamanTotem != null &&
                                                        MainSettings.KillTotems)
                                             )),
                                     //Focused Unit
                                     new Decorator(ret => Unit.ValidFocus(true, false),
                                                   new PrioritySelector(
                                                       new Decorator(
                                                           ret =>
                                                           MainSettings.CCFreezingTrap && CanBeTrapped(Focus) &&
                                                           Helpers.GetSpellCooldown("Freezing Trap").TotalSeconds < 1,
                                                           CM.CreateTrapBehavior("Freezing Trap", true, ret => Focus,
                                                                                    ret =>
                                                                                    SafeToTrap(Focus) &&
                                                                                    Focus.HasCachedAuraTimeLeft(
                                                                                        "Freezing Trap", true)
                                                                                         .TotalSeconds <= 1)),
                                                       CM.Buff("Wyvern Sting", ret => Focus,
                                                                  ret =>
                                                                  MainSettings.CCWyvernSting &&
                                                                  TalentManager.HasTalent(5) &&
                                                                  !Focus.HasAura("Freezing Trap")),
                                                       CM.Cast("Scatter Shot", ret => Focus,
                                                                  ret =>
                                                                  MainSettings.FocusScatter &&
                                                                  !Helpers.SelfControl(Focus) && Lists.CCd(Focus) < 500 &&
                                                                  Lists.Rooted(Focus) < 500),
                                                       PetManager.CreateCastPetActionOn("Lullaby", ret => Focus,
                                                                                        ret =>
                                                                                        PetSettings.Lullaby &&
                                                                                        !Focus.HasAnyCachedAura(new[]
                                                                                            {"Lullaby", "Scatter Shot"}) &&
                                                                                        ((Helpers.GetSpellCooldown(
                                                                                            "Scatter Shot").TotalSeconds <
                                                                                          27 &&
                                                                                          Helpers.GetSpellCooldown(
                                                                                              "Scatter Shot")
                                                                                                 .TotalSeconds > 1) ||
                                                                                         !MainSettings.FocusScatter) &&
                                                                                        Pet.Location.Distance(
                                                                                            Focus.Location) <= 20 &&
                                                                                        Lists.CCd(Focus) < 500),
                                                       CM.Cast("Tranquilizing Shot", ret => Focus,
                                                                  ret =>
                                                                  MainSettings.FocusTranq &&
                                                                  Lists.CanBeTranquilized(Focus, 2) &&
                                                                  (Helpers.PlayerPower > 60 ||
                                                                   Me.HasAnyCachedAura(new[]
                                                                       {"The Beast Within", "Thrill of the Hunt"}) ||
                                                                   TalentManager.HasGlyph("Tranquilizing Shot")) &&
                                                                  Focus.Distance <= 40 &&
                                                                  Me.CurrentPendingCursorSpell == null)
                                                       ))));
        }

        #endregion

        //Finished

        #region Rotation Handler : Interrupt

        private static Composite Interrupt(Main.Selection<bool> reqs = null)
        {
            return new Decorator(when => reqs == null || reqs(when),
                                 new PrioritySelector(
                                     new Decorator(ret => Unit.ValidFocus(true, false),
                                                   new PrioritySelector(
                                                       CM.Cast("Silencing Shot", ret => Focus,
                                                                  ret =>
                                                                  TalentManager.HasTalent(4) &&
                                                                  MainSettings.FocusSilencingShot && NeedSilence(Focus) &&
                                                                  Focus.HealthPercent < MainSettings.SilenceHealth &&
                                                                  Lists.BreakOnDamage(Focus) < 250 &&
                                                                  Lists.Silenced(Focus) < 250 && Lists.CCd(Focus) < 250),
                                                       CM.Cast("Wyvern Sting", ret => Focus,
                                                                  ret =>
                                                                  MainSettings.CCWyvernSting &&
                                                                  !Focus.HasAura("Freezing Trap") &&
                                                                  Focus.Distance <= 35),
                                                       CM.Cast("Tranquilizing Shot", ret => Focus,
                                                                  ret =>
                                                                  MainSettings.FocusTranq &&
                                                                  Lists.CanBeTranquilized(Focus, 2) &&
                                                                  (Helpers.PlayerPower > 60 ||
                                                                   Me.HasAura("The Beast Within") &&
                                                                   Focus.Distance <= 40)),
                                                       CM.Buff("Concussive Shot", ret => Focus,
                                                                  ret =>
                                                                  MainSettings.FocusConcussive &&
                                                                  Lists.Snared(Focus) <= 500 && Focus.Distance <= 40 &&
                                                                  !Helpers.SpellOnGCD("Kill Command") &&
                                                                  Me.CurrentPendingCursorSpell == null &&
                                                                  Lists.BreakOnDamage(Focus) < 250),
                                                       CM.BuffSelf("War Stomp",
                                                                      ret =>
                                                                      CooldownSettings.Racial &&
                                                                      Me.Race == WoWRace.Tauren && Focus.Distance <= 8 &&
                                                                      Focus.IsCasting &&
                                                                      Focus.CurrentCastTimeLeft.TotalMilliseconds > 850 &&
                                                                      Lists.CCd(Focus) < 750))),
                                     new Decorator(ret => Unit.ValidUnit(Target, true, false),
                                                   new PrioritySelector(
                                                       CM.Cast("Tranquilizing Shot", ret => Target,
                                                                  ret =>
                                                                  MainSettings.Tranquilizing &&
                                                                  Lists.CanBeTranquilized(Target, 2) &&
                                                                  (Helpers.PlayerPower > 60 ||
                                                                   Me.HasAura("The Beast Within") ||
                                                                   Target.HasAura("Freezing Trap")) &&
                                                                  Target.Distance <= 40),
                                                       CM.Cast("Intimidation",
                                                                  ret =>
                                                                  NeedIntimidate() && Lists.BreakOnDamage(Target) < 250 &&
                                                                  Lists.CCd(Target) < 500),
                                                       CM.Cast("Silencing Shot", ret => Target,
                                                                  ret =>
                                                                  TalentManager.HasTalent(4) && MainSettings.Talent30_SS &&
                                                                  NeedSilence(Target) &&
                                                                  Target.HealthPercent < MainSettings.SilenceHealth &&
                                                                  Lists.BreakOnDamage(Target) < 250 &&
                                                                  Lists.Silenced(Target) < 250),
                                                       CM.Buff("Concussive Shot", ret => Target,
                                                                  ret =>
                                                                  MainSettings.ConcussiveShot &&
                                                                  Lists.BreakOnDamage(Target) < 250 &&
                                                                  Lists.Snared(Target) <= 0 && Target.Distance <= 40 &&
                                                                  !Helpers.SpellOnGCD("Kill Command")),
                                                       CM.BuffSelf("War Stomp",
                                                                      ret =>
                                                                      CooldownSettings.Racial &&
                                                                      Me.Race == WoWRace.Tauren && Target.Distance <= 8 &&
                                                                      Lists.BreakOnDamage(Target) < 250 &&
                                                                      Lists.CCd(Target) < 750 && Target.IsCasting &&
                                                                      Target.CurrentCastTimeLeft.TotalMilliseconds > 850),
                                                       CM.Cast("Arcane Torrent", ret => Target,
                                                                  ret =>
                                                                  CooldownSettings.Racial && Me.Race == WoWRace.BloodElf &&
                                                                  Lists.BreakOnDamage(Target) < 250 &&
                                                                  Lists.Silenced(Target) < 250 &&
                                                                  Me.CurrentTarget.IsPlayer &&
                                                                  (Lists.InterruptCastNoChannel(Target) > 200 ||
                                                                   Lists.InterruptCastChannel(Target) > 200))))
                                     ));
        }

        #endregion

        //Finished

        #region Rotation Handler : FocusRegen

        private static Composite FocusRegen(Main.Selection<bool> reqs = null)
        {
            return new Decorator(when => reqs == null || reqs(when),
                                 new PrioritySelector(
                                     CM.CancelCast("Cobra Shot",
                                                      ret =>
                                                      Helpers.PlayerPower > 100 &
                                                      Me.CurrentCastTimeLeft.TotalMilliseconds > 500, RunStatus.Failure),
                                     CM.Cast("Kill Shot", ret => Target.HealthPercent <= 20),
                                     CM.Cast("Dire Beast",
                                                ret =>
                                                MainSettings.Talent60_DB && TalentManager.HasTalent(11) &&
                                                CM.PredicFocus() < 70),
                                     CM.Cast("Blink Strike",
                                                ret => MainSettings.Talent75_BS && TalentManager.HasTalent(14)),
                                     CM.Cast("Glaive Toss",
                                                ret => MainSettings.Talent90_GT && TalentManager.HasTalent(16)),
                                     CM.CastCobraShot(ret => Target, ret => CM.PredicFocus() < 100),
                                     CM.Cast("Arcane Shot",
                                                ret => MainSettings.ArcaneShot && CM.PredicFocus() >= 100)
                                     ));
        }

        #endregion

        //Finished

        #region Rotation Handler : TrapBehavior

        private static Composite TrapBehavior(Main.Selection<bool> reqs = null)
        {
            return new Decorator(when => reqs == null || reqs(when),
                                 new PrioritySelector(
                                     new Decorator(ret => Target.Distance > 5 && Target.Distance < 40,
                                                   new PrioritySelector(
                                                       CM.CreateTrapBehavior("Ice Trap", true, ret => Target,
                                                                                ret =>
                                                                                MainSettings.IceTrapLauncher &&
                                                                                Target.GetClassSpec() == ClassType.Melee),
                                                       CM.CreateTrapBehavior("Snake Trap", true, ret => Target,
                                                                                ret =>
                                                                                MainSettings.SnakeTrapLauncher &&
                                                                                Target.GetClassSpec() == ClassType.Melee),
                                                       CM.CreateTrapBehavior("Snake Trap", true, ret => Target,
                                                                                ret =>
                                                                                MainSettings.SnakeTrapLauncher &&
                                                                                Target.GetClassSpec() ==
                                                                                ClassType.MRanged),
                                                       CM.CreateTrapBehavior("Freezing Trap", true, ret => Target,
                                                                                ret =>
                                                                                MainSettings.FreezingTrapLauncher &&
                                                                                Target.GetClassSpec() ==
                                                                                ClassType.PRanged),
                                                       CM.CreateTrapBehavior("Explosive Trap", true, ret => Target,
                                                                                ret =>
                                                                                MainSettings.ExplosiveTrapLauncher &&
                                                                                Target.GetClassSpec() == ClassType.Melee)
                                                       )),
                                     new Decorator(ret => Target.Distance <= 5,
                                                   new PrioritySelector(
                                                       CM.CreateTrapBehavior("Ice Trap", true, ret => Target,
                                                                                ret =>
                                                                                MainSettings.IceTrap &&
                                                                                Target.GetClassSpec() == ClassType.Melee),
                                                       CM.CreateTrapBehavior("Snake Trap", true, ret => Target,
                                                                                ret =>
                                                                                MainSettings.SnakeTrap &&
                                                                                Target.GetClassSpec() == ClassType.Melee),
                                                       CM.CreateTrapBehavior("Snake Trap", true, ret => Target,
                                                                                ret =>
                                                                                MainSettings.SnakeTrap &&
                                                                                Target.GetClassSpec() ==
                                                                                ClassType.MRanged),
                                                       CM.CreateTrapBehavior("Freezing Trap", true, ret => Target,
                                                                                ret =>
                                                                                MainSettings.FreezingTrap &&
                                                                                Target.GetClassSpec() ==
                                                                                ClassType.PRanged),
                                                       CM.CreateTrapBehavior("Explosive Trap", true, ret => Target,
                                                                                ret =>
                                                                                MainSettings.ExplosiveTrap &&
                                                                                Me.CurrentTarget.GetClassSpec() ==
                                                                                ClassType.Melee)
                                                       ))));
        }

        #endregion

        //Finished

        #region Rotation Handler : Single Target

        private static Composite SingleTarget
        {
            get
            {
                return new Decorator(req => Lists.BreakOnDamage(Target) < 250,
                                     new PrioritySelector(
                                         PetManager.CreateCastPetActionOn("Venom Web Spray", ret => Target,
                                                                          ret =>
                                                                          Target.CurrentTargetGuid == Me.Guid &&
                                                                          Lists.Rooted(Target) <= 0),
                                         CM.Cast("Hunter's Mark",
                                                    ret => MainSettings.HuntersMark && !Target.HasAura("Hunter's Mark")),
                                         CM.Cast("Kill Shot", ret => Target.HealthPercent <= 20),
                                         new Decorator(ret => NeedCobraShot(),
                                                       new PrioritySelector(
                                                           CM.CastCobraShot(ret => Target,
                                                                               ret =>
                                                                               MainSettings.KillCommand &&
                                                                               Me.GotAlivePet &&
                                                                               CM.PredicFocus() < 56 &&
                                                                               Helpers.PlayerPower > 40 &&
                                                                               !Me.HasAura("The Beast Within") &&
                                                                               Helpers.GetSpellCooldown("Kill Command")
                                                                                      .TotalMilliseconds > 1000 &&
                                                                               Helpers.GetSpellCooldown("Kill Command")
                                                                                      .TotalMilliseconds < 3000),
                                                           CM.CastCobraShot(ret => Target,
                                                                               ret =>
                                                                               MainSettings.KillCommand &&
                                                                               Me.GotAlivePet &&
                                                                               (CM.PredicFocus() < 18 ||
                                                                                (!Me.HasAura("The Beast Within") &&
                                                                                 CM.PredicFocus() < 38)) &&
                                                                               Helpers.GetSpellCooldown("Kill Command")
                                                                                      .TotalMilliseconds < 3000)
                                                           )),
                                         new Decorator(ret => Me.GotAlivePet,
                                                       new PrioritySelector( // Pet dependand, check only once for speed
                                                           CM.Cast("Kill Command",
                                                                      ret =>
                                                                      Pet.Location.Distance(Target.Location) <= 25 &&
                                                                      Pet.InLoSSS(Target)),
                                                           CM.Cast("Blink Strike",
                                                                      ret =>
                                                                      MainSettings.Talent75_BS &&
                                                                      TalentManager.HasTalent(14) &&
                                                                      Pet.Location.Distance(Target.Location) <= 40 &&
                                                                      Pet.InLoSSS(Target)),
                                                           CM.Cast("Lynx Rush",
                                                                      ret =>
                                                                      MainSettings.Talent75_LXR &&
                                                                      TalentManager.HasTalent(15) &&
                                                                      Pet.Location.Distance(Target.Location) <= 10 &&
                                                                      Target.HasCachedAuraTimeLeft("Lynx Rush", true)
                                                                            .TotalSeconds < 4 && Pet.InLoSSS(Target))
                                                           )),
                                         CM.PreventDoubleCast("Serpent Sting", 2, ret => Target,
                                                                 ret =>
                                                                 NeedSerpentSting() &&
                                                                 Target.HasCachedAuraTimeLeft("Serpent Sting", true)
                                                                       .TotalSeconds < 1),
                                         CM.Cast("Dire Beast",
                                                    ret => MainSettings.Talent60_DB && TalentManager.HasTalent(15)),
                                         CM.Buff("A Murder of Crows", ret => Target,
                                                    ret =>
                                                    MainSettings.Talent75_AMOC && TalentManager.HasTalent(13) &&
                                                    Target.IsPlayer && !Target.HasMyCachedAura("A Murder of Crows") &&
                                                    (!CooldownSettings.BestialWrath || Me.HasAura("The Beast Within") ||
                                                     Helpers.GetSpellCooldown("Bestial Wrath").TotalSeconds >= 10 ||
                                                     !Me.GotAlivePet || HasSpeedBuff())),
                                         CM.Buff("Widow Venom", true, ret => Target,
                                                    ret =>
                                                    CM.PredicFocus() >= 55 &&
                                                    Target.HasCachedAuraTimeLeft("Widow Venom", true).TotalSeconds <= 1 &&
                                                    (!MainSettings.KillCommand ||
                                                     Helpers.GetSpellCooldown("Kill Command").TotalSeconds > 2) && !Target.HasAnyCachedAura(new[] {"Widow Venom","Wild Strike", "Mortal Strike", "Wound Poison"})),
                                                    
                                         CM.Cast("Glaive Toss",
                                                    ret =>
                                                    MainSettings.Talent90_GT && TalentManager.HasTalent(16) &&
                                                    (!Me.CurrentMap.IsArena || !Unit.ValidFocus(true, false) ||
                                                     Lists.BreakOnDamage(Focus) < 500 ||
                                                     Target.Location.Distance(Focus.Location) > 7)),
                                         CM.Cast("Powershot",
                                                    ret =>
                                                    MainSettings.Talent90_PWRS && TalentManager.HasTalent(17) &&
                                                    Me.HasAnyCachedAura(new[]
                                                        {
                                                            "Rapid Fire", "Focus Fire", "Bloodlust", "Heroism",
                                                            "Ancient Hysteria", "Time Warp"
                                                        }) &&
                                                    (Helpers.GetSpellCooldown("Kill Command").TotalMilliseconds > 2500 ||
                                                     !Me.GotAlivePet || !MainSettings.KillCommand)),
                                         CM.Cast("Barrage",
                                                    ret => MainSettings.Talent90_BRG && TalentManager.HasTalent(18)),
                                         CM.CastCobraShot(ret => Target,
                                                             ret => CM.PredicFocus() < 90 && Lists.HasteBuffs() >= 50
                                                                    &&
                                                                    (!TalentManager.HasTalent(10) ||
                                                                     (Helpers.GetSpellCooldown("Fervor")
                                                                             .TotalMilliseconds > 750 &&
                                                                      Helpers.GetSpellCooldown("Fervor").TotalSeconds <
                                                                      29))),
                                         CM.Cast("Arcane Shot",
                                                    ret =>
                                                    MainSettings.ArcaneShot &&
                                                    (Helpers.PlayerPower > 62 ||
                                                     (Me.HasAura("The Beast Within") && Helpers.PlayerPower > 42) ||
                                                     Me.HasAura("Thrill of the Hunt")
                                                     ||
                                                     (TalentManager.HasTalent(10) &&
                                                      Helpers.GetSpellCooldown("Fervor").TotalMilliseconds < 500) ||
                                                     !Me.GotAlivePet || !MainSettings.KillCommand ||
                                                     !Pet.InLoSSS(Target))),
                                         new Decorator(ret => NeedCobraShot(),
                                                       CM.CastCobraShot(ret => Target,
                                                                           ret =>
                                                                           CM.PredicFocus() <
                                                                           MainSettings.FocusAmount ||
                                                                           (Me.HasAura("The Beast Within") &&
                                                                            CM.PredicFocus() <
                                                                            MainSettings.FocusAmount/2)))
                                         ));
            }
        }

        #endregion

        //No AoE Rotation

        #region Rotation Handler : Area of Effect

        private static Composite AreaofEffect(Main.Selection<bool> reqs = null)
        {
            return new Decorator(when => reqs == null || reqs(when),
                                 new PrioritySelector(
                                     ));
        }

        #endregion

        #region Ext :

        public  Composite CreateCombat
        {
            get { return CombatSelector; }
        }

        public  Composite CreatePrecombat
        {
            get { return PreCombatSelector; }
        }

        public static Composite HoldOn(WoWUnit u, bool enemy, bool friendly)
        {
            return new Decorator(ret => !Unit.ValidUnit(u, enemy, friendly) || Lists.ShouldHold(u),
                                 new ActionAlwaysSucceed());
        }

        public static Composite AutoPause()
        {
            return new Decorator(ret => Me.HasAnyCachedAura(new[] {"Drink", "Food", "First Aid", "Feign Death"}),
                                 new ActionAlwaysSucceed());
        }

/*
        private static WoWUnit MDTarget()
        {
            switch (MainSettings.MisdirectOn)
            {
                case MisdirectOn.Me:
                    return StyxWoW.Me;
                case MisdirectOn.Focus:
                    return Unit.ValidFocus(false, true) ? Focus : null;
                case MisdirectOn.Pet:
                    return Me.GotAlivePet ? Pet : null;
                case MisdirectOn.Nobody:
                    return null;

            }
            return null;
        }
*/

        #endregion

        #region Tidy : Bools

        private static bool CanBeTrapped(WoWUnit u)
        {
            if (u == null)
                return false;

            return Lists.Immune(u) <= 250 && Lists.ImmuneSpell(u) <= 250 &&
                   (TalentManager.HasGlyph("Solace") || Lists.Dotted(u) <= 500);
        }

        private static bool SafeToTrap(WoWUnit u)
        {
            if (u == null)
                return false;
            return Lists.Rooted(u) >= 500 || u.HasAnyCachedAura(new[] {"Scatter Shot", "Lullaby", "Bad Manners"}) ||
                   u.HasAura(136634) || (Lists.CCd(u) <= 1500 && !u.IsMoving);
        }


        private static bool NeedSilence(WoWUnit unit)
        {
            switch (MainSettings.SilencePvP)
            {
                case SilencePvP.Healer:
                    return unit.IsPlayer && unit.GetClassSpec() == ClassType.Healer;

                case SilencePvP.Casting:
                    return unit.IsPlayer &&
                           (Lists.InterruptCastNoChannel(Target) > 200 || Lists.InterruptCastChannel(Target) > 200);

                case SilencePvP.Casting_Healer:
                    return unit.IsPlayer &&
                           (unit.GetClassSpec() == ClassType.Healer ||
                            (Lists.InterruptCastNoChannel(Target) > 200 || Lists.InterruptCastChannel(Target) > 200));

                case SilencePvP.Never:
                    return false;
            }
            return false;
        }


        private static bool NeedFeignDeath()
        {
            if (Unit.ValidUnit(Target, true, false) && Me.HealthPercent < MainSettings.FeignHealthPvP)
            {
                switch (MainSettings.PvPFeignDeath)
                {
                    case PvPFeignDeath.PetNear:
                        return Target.GotAlivePet && Target.Pet.CurrentTargetGuid == Me.Guid && Target.Pet.Distance < 10;

                    case PvPFeignDeath.Casting:
                        return Target.IsPlayer && Target.GotTarget && Target.CurrentTargetGuid == Me.Guid &&
                               Target.IsCasting &&
                               (Lists.InterruptCastNoChannel(Target) > 200 || Lists.InterruptCastChannel(Target) > 200) &&
                               WoWSpell.FromId(Target.CastingSpellId).SpellEffect1.EffectType != WoWSpellEffectType.Heal;

                    case PvPFeignDeath.Targetted:
                        return Target.GotTarget && Target.CurrentTargetGuid == Me.Guid
                               && ((Target.GetClassSpec() == ClassType.Melee && Target.Distance < 10)
                                   ||
                                   ((Target.GetClassSpec() == ClassType.PRanged ||
                                     Target.GetClassSpec() == ClassType.MRanged) && Target.Distance < 40));

                    case PvPFeignDeath.Casting_Pet:
                        return ((Target.GotAlivePet && Target.Pet.CurrentTargetGuid == Me.Guid &&
                                 Target.Pet.Distance < 10) ||
                                (Target.IsPlayer && Target.CurrentTargetGuid == Me.Guid && Target.IsCasting &&
                                 (Lists.InterruptCastNoChannel(Target) > 200 || Lists.InterruptCastChannel(Target) > 200) &&
                                 WoWSpell.FromId(Target.CastingSpellId).SpellEffect1.EffectType !=
                                 WoWSpellEffectType.Heal));

                    case PvPFeignDeath.Casting_Targetted:
                        return (Target.IsPlayer && Target.GotTarget && Target.CurrentTargetGuid == Me.Guid &&
                                (Lists.InterruptCastNoChannel(Target) > 200 || Lists.InterruptCastChannel(Target) > 200) &&
                                WoWSpell.FromId(Target.CastingSpellId).SpellEffect1.EffectType !=
                                WoWSpellEffectType.Heal)
                               ||
                               (Target.CurrentTargetGuid == Me.Guid
                                && ((Target.GetClassSpec() == ClassType.Melee && Target.Distance < 10)
                                    ||
                                    ((Target.GetClassSpec() == ClassType.PRanged ||
                                      Target.GetClassSpec() == ClassType.MRanged) && Target.Distance < 40)));

                    case PvPFeignDeath.Pet_Targetted:
                        return (Target.GotAlivePet && Target.Pet.CurrentTargetGuid == Me.Guid &&
                                Target.Pet.Distance < 10)
                               ||
                               (Target.GotTarget && Target.CurrentTargetGuid == Me.Guid
                                && ((Target.GetClassSpec() == ClassType.Melee && Target.Distance < 10)
                                    ||
                                    ((Target.GetClassSpec() == ClassType.PRanged ||
                                      Target.GetClassSpec() == ClassType.MRanged) && Target.Distance < 40)));

                    case PvPFeignDeath.Casting_Pet_Targetted:
                        return (Target.IsPlayer && Target.GotTarget && Target.CurrentTargetGuid == Me.Guid &&
                                (Lists.InterruptCastNoChannel(Target) > 200 || Lists.InterruptCastChannel(Target) > 200) &&
                                WoWSpell.FromId(Target.CastingSpellId).SpellEffect1.EffectType !=
                                WoWSpellEffectType.Heal)
                               ||
                               (Target.GotAlivePet && Target.Pet.CurrentTargetGuid == Me.Guid &&
                                Target.Pet.Distance < 10)
                               ||
                               (Target.GotTarget && Target.CurrentTargetGuid == Me.Guid
                                && ((Target.GetClassSpec() == ClassType.Melee && Target.Distance < 10)
                                    ||
                                    ((Target.GetClassSpec() == ClassType.PRanged ||
                                      Target.GetClassSpec() == ClassType.MRanged) && Target.Distance < 40)));

                    case PvPFeignDeath.Never:
                        return false;
                }
            }
            return false;
        }

        private static bool NeedIntimidate()
        {
            if (Target == null || !Me.GotAlivePet)
                return false;
            switch (MainSettings.Intimidation)
            {
                case Intimidate.Interrupt:
                    return Pet.Location.Distance(Target.Location) < 9 &&
                           (Lists.InterruptCastNoChannel(Target) > 200 || Lists.InterruptCastChannel(Target) > 200);
                case Intimidate.Defend:
                    return Pet.Location.Distance(Target.Location) < 9 && Target.Distance <= 20 &&
                           Target.CurrentTargetGuid == Me.Guid;
                case Intimidate.Interrupt_Defend:
                    return (Target.Distance <= 20 && Target.CurrentTargetGuid == Me.Guid ||
                            (Lists.InterruptCastNoChannel(Target) > 200 || Lists.InterruptCastChannel(Target) > 200));
                case Intimidate.Interrupt_Defend_Health:
                    return Target.HealthPercent < MainSettings.IntimidationHealth &&
                           (Target.Distance <= 20 && Target.CurrentTargetGuid == Me.Guid ||
                            (Lists.InterruptCastNoChannel(Target) > 200 || Lists.InterruptCastChannel(Target) > 200));
                case Intimidate.Interrupt_Health:
                    return Target.HealthPercent < MainSettings.IntimidationHealth &&
                           (Lists.InterruptCastNoChannel(Target) > 200 || Lists.InterruptCastChannel(Target) > 200);
                case Intimidate.LowHealth:
                    return Target.HealthPercent < MainSettings.IntimidationHealth;
                case Intimidate.Never:
                    return false;
            }
            return false;
        }

        private static bool NeedReadiness()
        {
            if (Target == null)
                return false;
            return CooldownSettings.Readiness &&
                   Helpers.GetSpellCooldown("Rapid Fire").TotalSeconds > 10 &&
                   Helpers.GetSpellCooldown("Bestial Wrath").TotalSeconds > 8 &&
                   ExtReadiness();
        }

        private static bool ExtReadiness()
        {
            return (!MainSettings.KillCommand || Helpers.GetSpellCooldown("Kill Command").TotalSeconds > 2)
                   && (!Me.CurrentMap.IsArena || Helpers.GetSpellCooldown("Disengage").TotalSeconds > 2)
                   &&
                   (!TalentManager.HasTalent(4) || !Me.CurrentMap.IsArena ||
                    Helpers.GetSpellCooldown("Silencing Shot").TotalSeconds > 2)
                   &&
                   (!TalentManager.HasTalent(5) || !Me.CurrentMap.IsArena ||
                    Helpers.GetSpellCooldown("Wyvern Sting").TotalSeconds > 2)
                   &&
                   (!TalentManager.HasTalent(6) || !Me.CurrentMap.IsArena ||
                    Helpers.GetSpellCooldown("Binding Shot").TotalSeconds > 2)
                   && (!TalentManager.HasTalent(7) || Helpers.GetSpellCooldown("Exhilaration").TotalSeconds > 9)
                   && (!TalentManager.HasTalent(11) || Helpers.GetSpellCooldown("Dire Beast").TotalSeconds > 4)
                   && (!TalentManager.HasTalent(10) || Helpers.GetSpellCooldown("Fervor").TotalSeconds > 8)
                   && (!TalentManager.HasTalent(13) || Helpers.GetSpellCooldown("A Murder of Crows").TotalSeconds > 10)
                   && (!TalentManager.HasTalent(14) || Helpers.GetSpellCooldown("Blink Strike").TotalSeconds > 2)
                   && (!TalentManager.HasTalent(15) || Helpers.GetSpellCooldown("Lynx Rush").TotalSeconds > 10)
                   && (!TalentManager.HasTalent(16) || Helpers.GetSpellCooldown("Glaive Toss").TotalSeconds > 2)
                   && (!TalentManager.HasTalent(17) || Helpers.GetSpellCooldown("Powershot").TotalSeconds > 5)
                   && (!TalentManager.HasTalent(18) || Helpers.GetSpellCooldown("Barrage").TotalSeconds > 5)
                   &&
                   (MainSettings.Intimidation == Intimidate.Never || !Me.CurrentMap.IsArena ||
                    Helpers.GetSpellCooldown("Intimidation").TotalSeconds > 2);
        }

        private static bool NeedCobraShot()
        {
            if (Target == null)
                return false;
            return
                /*(Helpers.GetSpellCooldown("Kill Command").TotalSeconds > Math.Round(((CM.GetSpellCastTime("Cobra Shot").TotalSeconds / ((100 + Helpers.HasteBuffs()) / 100)) - 0.4), 2) || CLua.PlayerPower < 20 || (!Me.HasAura("The Beast Within") && CLua.PlayerPower < 40))
                    && */ (!TalentManager.HasTalent(10) ||
                                                                                                                                                                                                                                                                                                                 (Helpers
                                                                                                                                                                                                                                                                                                                      .GetSpellCooldown
                                                                                                                                                                                                                                                                                                                      ("Fervor")
                                                                                                                                                                                                                                                                                                                      .TotalMilliseconds >
                                                                                                                                                                                                                                                                                                                  750 &&
                                                                                                                                                                                                                                                                                                                  Helpers
                                                                                                                                                                                                                                                                                                                      .GetSpellCooldown
                                                                                                                                                                                                                                                                                                                      ("Fervor")
                                                                                                                                                                                                                                                                                                                      .TotalSeconds <
                                                                                                                                                                                                                                                                                                                  29));
        }

        private static bool NeedBestialWrath()
        {
            if (Target == null)
                return false;
            return Me.GotAlivePet && Pet.Location.Distance(Target.Location) <= 25 && !Me.HasAura("The Beast Within") &&
                   ExtBestial();
        }

        private static bool ExtBestial()
        {
            if (Target == null)
                return false;
            return (!TalentManager.HasTalent(15) || Helpers.GetSpellCooldown("Lynx Rush").TotalSeconds > 5 ||
                    Helpers.GetSpellCooldown("Lynx Rush").TotalMilliseconds < 1500) &&
                   (!TalentManager.HasTalent(13) || Helpers.GetSpellCooldown("A Murder of Crows").TotalSeconds > 5 ||
                    Helpers.GetSpellCooldown("A Murder of Crows").TotalMilliseconds < 1500);
        }

        private static bool NeedSerpentSting()
        {
            if (Target == null)
                return false;
            switch (MainSettings.SerpentSting)
            {
                case SerpentSting.Always:
                    return true;
                case SerpentSting.Sometimes:
                    return Target.HealthPercent > 25;
                case SerpentSting.Never:
                    return false;
            }
            return false;
        }

        private static bool HasSpeedBuff()
        {
            return
                Me.HasAnyCachedAura(new[]
                    {"Bloodlust", "Heroism", "Ancient Hysteria", "Time Warp", "Rapid Fire", "Focus Fire"});
        }

        #endregion

        #region Tidy: Calls

        private static WoWUnit Pet
        {
            get { return StyxWoW.Me.Pet; }
        }

        private static WoWUnit Focus
        {
            get { return StyxWoW.Me.FocusedUnit; }
        }

        private static WoWUnit Target
        {
            get { return StyxWoW.Me.CurrentTarget; }
        }

        private static LocalPlayer Me
        {
            get { return StyxWoW.Me; }
        }

        private static MainSettings MainSettings
        {
            get { return MainSettings.Instance; }
        }

        private static CooldownSettings CooldownSettings
        {
            get { return CooldownSettings.Instance; }
        }

        private static PetSettings PetSettings
        {
            get { return PetSettings.Instance; }
        }

        #endregion

        #region Tidy : PetMacro's

        private static Composite SetPetActions
        {
            get
            {
                return new Decorator(ret => MainSettings.DisablePetActions,
                                     new Action(delegate
                                         {
                                             if (MainSettings.GrowlPets)
                                             {
                                                 Lua.DoString("RunMacroText('/petautocastoff Growl');");
                                             }
                                             if (PetSettings.SpiritMend != SpiritMend.Never)
                                             {
                                                 Lua.DoString("RunMacroText('/petautocastoff Spirit Mend');");
                                             }
                                             if (PetSettings.RoarOfSacrifice != RoarOfSacrifice.Never)
                                             {
                                                 Lua.DoString("RunMacroText('/petautocastoff Roar of Sacrifice');");
                                             }
                                             return RunStatus.Failure;
                                         }));
            }
        }

        #endregion
    }
}