﻿#region Author Information

// // Mark I - SurvivalPvE.cs
// // Created on 12/04/2013 at 19:05 Uhr
// // Author: Mirabis - Moreno Sint Hill
// 

#endregion

using System;
using System.Linq;
using Bestia_Magister.Managers;
using Bestia_Magister.Misc;
using Bestia_Magister.Settings;
using CommonBehaviors.Actions;
using Styx;
using Styx.CommonBot;
using Styx.TreeSharp;
using Styx.WoWInternals;
using Styx.WoWInternals.WoWObjects;
using Action = Styx.TreeSharp.Action;
using CM = Bestia_Magister.Managers.CastManager;

namespace Bestia_Magister.Rotations
{
    public class SurvivalPvE
    {
        #region Priority : Combat Selector

        private static Composite CombatSelector
        {
            get
            {
                return new PrioritySelector(
                    Helpers.TreePerformance("SurvivalPvE"),
                    AutoPause(),
                    new Decorator(ret => !HotkeyManager.Registered, new Action(ret => HotkeyManager.RegisterKeys())),
                    new Throttle(1, TimeSpan.FromSeconds(1), Helpers.AcquireTargets),
                    //SetPetActions, 
                    CM.CancelCast("Cobra Shot",ret => Helpers.PlayerPower > 90 & Me.CurrentCastTimeLeft.TotalMilliseconds > 700,RunStatus.Failure),
                    CM.WaitForGcdOrCastOrChannel(), // Return if Casting/Channeling ( Traverse Successor )
                    new Decorator(req => Unit.ValidUnit(Target, true, false), //Valid Target
                                  new PrioritySelector(
                                      new Throttle(Items.UseItems),
                                      PetManager.CreateHunterCallPetBehavior(),
                                      DefensiveCooldowns,
                                      CrowdControl(when => true),
                                      HotkeyBehavior(when => true),
                                      PetManager.CreateCastPetActionOn("Attack", on => Target),
                                      Interrupt(req => true),
                                      FocusRegen(when => HotkeyManager.RegenMode),
                                      //Normally we go down here ; )
                                      new Decorator(when => !HotkeyManager.RegenMode,
                                                    new PrioritySelector(
                                                        OffensiveCooldowns(when => true),
                                                        TrapBehavior(when => true),
                                                        AreaofEffect(when =>HotkeyManager.AoEMode ||Helpers.AddCount(25) >= MainSettings.AddCount),
                                                        SingleTarget))
                                      )));
            }
        }

        #endregion

        #region Priority : Pre Combat Selector

        private static Composite PreCombatSelector
        {
            get
            {
                return new PrioritySelector(
                    PetManager.CreateHunterCallPetBehavior(),
                    CM.Cast("Hunter's Mark", ret => Target,
                               ret =>
                               Settings.MainSettings.HuntersMarkOOC && Target.Distance > 40 && Target.Distance < 100 &&
                               (Me.CurrentMap.IsArena || Me.CurrentMap.IsBattleground) &&
                               !Target.HasMyCachedAura("Hunter's Mark"))
                    );
            }
        }

        #endregion

        //Finished

        #region Rotation Handler : Defensive Cooldowns

        private static Composite DefensiveCooldowns
        {
            get
            {
                return new PrioritySelector(
                    CM.BuffSelf("Exhilaration",
                                   ret => TalentManager.HasTalent(7) && MainSettings.Talent45 && NeedExhilaration()),
                    CM.BuffSelf("Feign Death", ret => ShouldFeignDeath()),
                    CM.BuffSelf("Master's Call",
                                   ret =>
                                   MainSettings.MastersCallPvE && Me.GotAlivePet && Lists.Snared(Me) >= 250 ||
                                   Lists.Rooted(Me) >= 250),
                    //CM.BuffSelf("Detterence", ret => ShouldDetterence() ),
                    CM.BuffSelf("Mend Pet",
                                   ret =>
                                   PetSettings.MendPet && Me.GotAlivePet && Pet.HealthPercent <= PetSettings.PetHealth &&
                                   !Pet.HasMyCachedAura("Mend Pet")),
                    PetManager.CreateCastPetAction("Heart of the Phoenix",
                                                   ret =>
                                                   PetSettings.RevivePet &&
                                                   (Me.Pet == null || (Me.Pet != null && !Pet.IsAlive))),
                    CM.BuffSelf("Will of The Forsaken",
                                   ret =>
                                   Me.Race == WoWRace.Undead && CooldownSettings.Racial &&
                                   Helpers.isForsaken(Me).TotalSeconds > 1),
                    CM.BuffSelf("Stoneform",
                                   ret =>
                                   CooldownSettings.Racial && Me.Race == WoWRace.Dwarf &&
                                   StyxWoW.Me.GetAllAuras()
                                          .Any(
                                              a =>
                                              a.Spell.Mechanic == WoWSpellMechanic.Bleeding ||
                                              a.Spell.DispelType == WoWDispelType.Disease ||
                                              a.Spell.DispelType == WoWDispelType.Poison)) // Defensive cooldown
                    );
            }
        }

        #endregion

        //Finished

        #region Rotation Handler : OffensiveCooldowns

        private static Composite OffensiveCooldowns(Main.Selection<bool> reqs = null)
        {
            return new Decorator(when => reqs == null || reqs(when),
                                 new PrioritySelector(
                                     CM.BuffSelf("Aspect of the Hawk",
                                                    ret =>
                                                    MainSettings.AspectHawk && !Me.IsMoving && !Me.HasAura(109260) &&
                                                    !Me.HasAura(13165)),
                                     PetManager.CreateCastPetActionOn("Rabid", ret => Me,
                                                                      ret =>
                                                                      PetSettings.Rabid && Lists.TargetIsBoss(Target)),
                                     CM.BuffSelf("Berserking",
                                                    ret => CooldownSettings.Racial && Me.Race == WoWRace.Troll),
                                     CM.BuffSelf("Blood Fury",
                                                    ret => CooldownSettings.Racial && Me.Race == WoWRace.Orc),
                                     CM.Buff("Hunter's Mark", ret => Target,
                                                ret =>
                                                MainSettings.HuntersMark &&
                                                (Target.HealthPercent > 20 || Lists.TargetIsBoss(Target))),
                                     CM.BuffSelf("Fervor",
                                                    ret =>
                                                    SpellManager.HasSpell("Fervor") && MainSettings.Talent60_FV &&
                                                    Helpers.PlayerPower < 40),
                                     CM.BuffSelf("Rapid Fire",
                                                    ret =>
                                                    CooldownSettings.RapidFire &&
                                                    (Lists.TargetIsBoss(Target) &&
                                                     !Me.HasAnyCachedAura(new[]
                                                         {"Bloodlust", "Heroism", "Time Warp", "Ancient Hysteria"})) ||
                                                    Lists.TrainingDummies.Contains(Target.Entry)),
                                     CM.Cast("Stampede",
                                                ret =>
                                                Target != null && CooldownSettings.Stampede &&
                                                Lists.TargetIsBoss(Target) &&
                                                (Target.CurrentHealth > 1000000 ||
                                                 Lists.TrainingDummies.Contains(Target.Entry))),
                                     CM.BuffSelf("Readiness",
                                                    ret =>
                                                    Lists.TargetIsBoss(Target) && NeedReadinessPvESV() &&
                                                    CooldownSettings.Readiness),
                                     CM.BuffSelf("Lifeblood",
                                                    ret => CooldownSettings.Lifeblood && Target.MaxHealth > Me.MaxHealth)
                                     ));
        }

        #endregion

        //Not Implemented

        #region Rotation Handler : HotkeyBehavior

        private static Composite HotkeyBehavior(Main.Selection<bool> reqs = null)
        {
            return new Decorator(when => reqs == null || reqs(when),
                                 new PrioritySelector(
                                     //TODO: Nothing here?
                                     ));
        }

        #endregion

        //Finished

        #region Rotation Handler : Crowd Control

        private static Composite CrowdControl(Main.Selection<bool> reqs = null)
        {
            return new Decorator(when => reqs == null || reqs(when),
                                 new PrioritySelector(
                                     CM.CastOnGround("Binding Shot", ret => Target.Location,
                                                        ret =>
                                                        MainSettings.Talent30_BS &&
                                                        Helpers.AddCount(15) >= MainSettings.BindingCount),
                                     CM.Buff("Distracting Shot", ret => Target,
                                                ret =>
                                                MainSettings.DistractingShot && Me.GotAlivePet &&
                                                TalentManager.HasGlyph("Distracting Shot") &&
                                                Target.CurrentTargetGuid == Me.Guid),
                                     CM.Cast("Blink Strike",
                                                ret =>
                                                MainSettings.Talent75_BS && Me.GotAlivePet &&
                                                Pet.Location.Distance(Target.Location) > 25 && Pet.InLoSSS(Target)),
                                     PetManager.CreateCastPetActionOn("Lullaby", ret => Focus,
                                                                      ret =>
                                                                      PetSettings.Lullaby &&
                                                                      Unit.ValidFocus(true, false) &&
                                                                      !Focus.HasAnyCachedAura(new[]
                                                                          {"Lullaby", "Scatter Shot"}) &&
                                                                      Pet.Location.Distance(Focus.Location) <= 20 &&
                                                                      Lists.CCd(Focus) < 250),
                                     CM.Cast("Misdirection", ret => MDTarget(),
                                                ret => !Me.HasAura("Misdirection") && ShouldMisdirect())
                                     ));
        }

        #endregion

        //Finished

        #region Rotation Handler : Interrupt

        private static Composite Interrupt(Main.Selection<bool> reqs = null)
        {
            return new Decorator(when => reqs == null || reqs(when),
                                 new PrioritySelector(
                                     CM.Cast("Scatter Shot", ret => ShouldScatterShot()),
                                     CM.Cast("Silencing Shot",
                                                ret =>
                                                TalentManager.HasTalent(4) && MainSettings.Talent30_SS &&
                                                (Lists.InterruptCastNoChannel(Target) > 200 ||
                                                 Lists.InterruptCastChannel(Target) > 200)),
                                     CM.Cast("Wyvern Sting",
                                                ret =>
                                                TalentManager.HasTalent(5) && MainSettings.Talent30_WS &&
                                                (Lists.InterruptCastNoChannel(Target) > 200 ||
                                                 Lists.InterruptCastChannel(Target) > 200)),
                                     CM.Buff("Concussive Shot", ret => Target,
                                                ret =>
                                                MainSettings.ConcussiveShot && Lists.Snared(Target) <= 0 &&
                                                !Lists.TargetIsBoss(Target) && Target.Distance <= 40),
                                     PetManager.CreateCastPetActionOn("Lullaby", ret => Target,
                                                                      ret =>
                                                                      PetSettings.LullabyInt && Pet != null &&
                                                                      !Target.HasAnyCachedAura(new[]
                                                                          {"Lullaby", "Scatter Shot"})
                                                                      &&
                                                                      ((Helpers.GetSpellCooldown("Scatter Shot")
                                                                               .TotalSeconds < 27 &&
                                                                        Helpers.GetSpellCooldown("Scatter Shot")
                                                                               .TotalSeconds > 1)
                                                                       ||
                                                                       (MainSettings.ScatterShot == ScatterShot.Never ||
                                                                        MainSettings.ScatterShot == ScatterShot.Defend))
                                                                      && Pet.Location.Distance(Target.Location) <= 20 &&
                                                                      (Lists.InterruptCastNoChannel(Target) > 200 ||
                                                                       Lists.InterruptCastChannel(Target) > 200))
                                     ));
        }

        #endregion

        //Finished - Unused

        #region Rotation Handler : FocusRegen

        private static Composite FocusRegen(Main.Selection<bool> reqs = null)
        {
            return new Decorator(when => reqs == null || reqs(when),
                                 new PrioritySelector(
                                     CM.CancelCast("Cobra Shot",
                                                      ret =>
                                                      Helpers.PlayerPower > 80 &
                                                      Me.CurrentCastTimeLeft.TotalMilliseconds > 500, RunStatus.Failure),
                                     CM.Cast("Kill Shot", ret => Target.HealthPercent <= 20),
                                     CM.Cast("Explosive Shot",
                                                ret =>
                                                MainSettings.ExplosiveShot &&
                                                Me.ActiveAuras.ContainsKey("Lock and Load")),
                                     CM.Cast("Dire Beast",
                                                ret =>
                                                MainSettings.Talent60_DB && TalentManager.HasTalent(11) &&
                                                CM.PredicFocus() < 60),
                                     CM.Cast("Blink Strike",
                                                ret => MainSettings.Talent75_BS && TalentManager.HasTalent(14)),
                                     CM.Cast("Glaive Toss",
                                                ret => MainSettings.Talent90_GT && TalentManager.HasTalent(16)),
                                     CM.CastCobraShot(ret => Target, ret => CM.PredicFocus() < 80),
                                     CM.Cast("Arcane Shot",
                                                ret => MainSettings.ArcaneShot && CM.PredicFocus() >= 80)
                                     ));
        }

        #endregion

        //Finished

        #region Rotation Handler : TrapBehavior

        private static Composite TrapBehavior(Main.Selection<bool> reqs = null)
        {
            return new Decorator(when => reqs == null || reqs(when),
                                 new PrioritySelector(
                                     new Decorator(ret => Target.Distance > 5,
                                                   new PrioritySelector(
                                                       CM.CreateTrapBehavior("Explosive Trap", true, ret => Target,
                                                                                ret => MainSettings.STEXPLTrapLauncher),
                                                       CM.CreateTrapBehavior("Explosive Trap", true, ret => Target,
                                                                                ret =>
                                                                                MainSettings.ExplosiveTrapLauncher &&
                                                                                Helpers.AddCount(10) > 2),
                                                       CM.CreateTrapBehavior("Ice Trap", true, ret => Target,
                                                                                ret => MainSettings.IceTrapLauncher),
                                                       CM.CreateTrapBehavior("Snake Trap", true, ret => Target,
                                                                                ret => MainSettings.SnakeTrapLauncher),
                                                       CM.CreateTrapBehavior("Freezing Trap", true, ret => Target,
                                                                                ret => MainSettings.FreezingTrapLauncher)
                                                       )),
                                     new Decorator(ret => Target.Distance <= 5,
                                                   new PrioritySelector(
                                                       CM.CreateTrapBehavior("Explosive Trap", false, ret => Target,
                                                                                ret => MainSettings.STEXPLTrapLauncher),
                                                       CM.CreateTrapBehavior("Explosive Trap", false, ret => Target,
                                                                                ret =>
                                                                                MainSettings.ExplosiveTrap &&
                                                                                Helpers.AddCount(10) > 2),
                                                       CM.CreateTrapBehavior("Ice Trap", false, ret => Target,
                                                                                ret => MainSettings.IceTrap),
                                                       CM.CreateTrapBehavior("Snake Trap", false, ret => Target,
                                                                                ret => MainSettings.SnakeTrap),
                                                       CM.CreateTrapBehavior("Freezing Trap", false, ret => Target,
                                                                                ret => MainSettings.FreezingTrap)
                                                       ))));
        }

        #endregion

        //Finished + Error

        #region Rotation Handler : Single Target

        private static Composite SingleTarget
        {
            get
            {
                return new PrioritySelector(
                    //Cancel Spell Casts
                    CM.CastCobraShot(ret => Target,
                                        ret =>
                                        !Me.IsCasting && Target.HasMyCachedAura("Serpent Sting") &&
                                        Target.HasCachedAuraTimeLeft("Serpent Sting", true).TotalSeconds < 4 &&
                                        Helpers.PlayerPower < 80),
                    
                    //Single Target Rotation
                    CM.Cast("Kill Shot", ret => Target.HealthPercent <= 20),
                    new Decorator(
                        ret =>
                        (!TalentManager.HasTalent(10) ||
                         (Helpers.GetSpellCooldown("Fervor").TotalMilliseconds > 750 &&
                          Helpers.GetSpellCooldown("Fervor").TotalSeconds < 29)),
                        new PrioritySelector(
                            CM.CastCobraShot(ret => Target,
                                                ret =>
                                                Helpers.GetSpellCooldown("Black Arrow").TotalMilliseconds < 3000 &&
                                                MainSettings.BlackArrow && CM.PredicFocus() < 33),
                            CM.CastCobraShot(ret => Target,
                                                ret =>
                                                Helpers.GetSpellCooldown("Explosive Shot").TotalMilliseconds < 3000 &&
                                                MainSettings.ExplosiveShot && CM.PredicFocus() < 23)
                            )),
                    CM.Cast("Explosive Shot",
                               ret => MainSettings.ExplosiveShot && Me.ActiveAuras.ContainsKey("Lock and Load")),
                    CM.Cast("Black Arrow",
                               ret =>
                               MainSettings.BlackArrow &&
                               Target.HasCachedAuraTimeLeft("Black Arrow", true).TotalSeconds < 1 &&
                               (DpsMeter.GetCombatTimeLeft(Target, 3).TotalSeconds > 20 || Lists.TargetIsBoss(Target))),
                    CM.Cast("A Murder of Crows", ret => Target,
                               ret =>
                               TalentManager.HasTalent(13) && MainSettings.Talent75_AMOC &&
                               !Target.HasMyCachedAura("A Murder of Crows") &&
                               (Lists.TargetIsBoss(Target) || DpsMeter.GetCombatTimeLeft(Target, 5).TotalSeconds > 60)),
                    CM.Cast("Explosive Shot", ret => MainSettings.ExplosiveShot),
                    CM.PreventDoubleCast("Serpent Sting", 2, ret => Target,
                                            ret =>
                                            ShouldSerpentStingPvE() &&
                                            Target.HasCachedAuraTimeLeft("Serpent Sting", true).TotalSeconds < 1),
                    CM.Cast("Dire Beast",
                               ret =>
                               MainSettings.Talent60_DB && TalentManager.HasTalent(11) &&
                               (Lists.TargetIsBoss(Target) || CM.PredicFocus() < 60 || Target.CurrentHealth > 150000)),
                    CM.Cast("Glaive Toss", ret => MainSettings.Talent90_GT && TalentManager.HasTalent(16)),
                    CM.Cast("Blink Strike", ret => MainSettings.Talent75_BS && TalentManager.HasTalent(14)),
                    CM.Cast("Lynx Rush", ret => Target,
                               ret =>
                               MainSettings.Talent75_LXR && TalentManager.HasTalent(15) && Me.GotAlivePet &&
                               Target.HasCachedAuraTimeLeft("Lynx Rush", true).TotalSeconds < 4 &&
                               Pet.Location.Distance(Target.Location) < 25 &&
                               (Lists.TargetIsBoss(Target) || DpsMeter.GetCombatTimeLeft(Target, 3).TotalSeconds > 20)),
                    CM.Cast("Powershot",
                               ret =>
                               MainSettings.Talent90_PWRS && TalentManager.HasTalent(17) &&
                               Me.HasAnyCachedAura(new[]
                                   {"Rapid Fire", "Bloodlust", "Heroism", "Ancient Hysteria", "Time Warp"}) &&
                               (Helpers.GetSpellCooldown("Explosive Shot").TotalMilliseconds > 2500 ||
                                !MainSettings.ExplosiveShot)),
                    CM.Cast("Barrage",
                               ret =>
                               MainSettings.Talent90_BRG && TalentManager.HasTalent(18) && Lists.TargetIsBoss(Target)),
                    CM.CastCobraShot(ret => Target, ret => CM.PredicFocus() < 70 && Lists.HasteBuffs() >= 40
                                                              &&
                                                              (!TalentManager.HasTalent(10) ||
                                                               (Helpers.GetSpellCooldown("Fervor").TotalMilliseconds >
                                                                750 &&
                                                                Helpers.GetSpellCooldown("Fervor").TotalSeconds < 29))),
                    CM.Cast("Arcane Shot",
                               ret =>
                               MainSettings.ArcaneShot && !Helpers.SpellOnGCD("Explosive Shot") &&
                               !Me.HasAura("Thrill of the Hunt") &&
                               (Helpers.PlayerPower >= 60 ||
                                (Helpers.GetSpellCooldown("Explosive Shot").TotalSeconds > 2.5 &&
                                 Helpers.PlayerPower >= 50))),
                    CM.Cast("Arcane Shot",
                               ret =>
                               MainSettings.ArcaneShot && Me.HasAura("Thrill of the Hunt") &&
                               (!MainSettings.ExplosiveShot || !Helpers.SpellOnGCD("Explosive Shot"))),
                    new Decorator(ret => ShouldCobraShotPvESV(),
                                  CM.CastCobraShot(ret => Target,
                                                      ret => CM.PredicFocus() < MainSettings.FocusAmount))
                    );
            }
        }

        #endregion

        //Finished

        #region Rotation Handler : Area of Effect

        private static Composite AreaofEffect(Main.Selection<bool> reqs = null)
        {
            return new Decorator(when => reqs == null || reqs(when),
                                 new PrioritySelector(
                                     CM.CreateTrapBehavior("Explosive Trap", true, ret => Target,
                                                              ret => Target.Distance >= 5 && MainSettings.AoEExpTrap),
                                     CM.CreateTrapBehavior("Explosive Trap", false, ret => Target,
                                                              ret => Target.Distance < 5 && MainSettings.AoEExpTrap),
                                     CM.Cast("Kill Shot",
                                                ret =>
                                                MainSettings.KillShot && Target.HealthPercent < 20 &&
                                                CM.PredicFocus() < 40),
                                     CM.Cast("Dire Beast",
                                                ret => MainSettings.AoEDireBeast && CM.PredicFocus() < 70),
                                     CM.Cast("Glaive Toss",
                                                ret => MainSettings.Talent90_GT && TalentManager.HasTalent(16)),
                                     CM.PreventDoubleCast("Explosive Shot", 1, ret => Target,
                                                             ret =>
                                                             MainSettings.ExplosiveShot &&
                                                             Me.ActiveAuras.ContainsKey("Lock and Load")),
                                     CM.Cast("Barrage",
                                                ret => MainSettings.Talent90_BRG && TalentManager.HasTalent(18)),
                                     CM.Cast("Multi-Shot",
                                                ret =>
                                                MainSettings.MultiShot &&
                                                (Me.CurrentFocus >= 40 ||
                                                 (Me.HasAura("Thrill of the Hunt") && Me.CurrentFocus >= 20) ||
                                                 (TalentManager.HasTalent(10) && !Helpers.SpellOnCooldown("Fervor")))),
                                     CM.Cast("Blink Strike",
                                                ret =>
                                                MainSettings.Talent75_BS && TalentManager.HasTalent(14) &&
                                                Me.GotAlivePet && CM.PredicFocus() < 40),
                                     CM.Buff("Lynx Rush", true, ret => Target, 
                                                ret =>
                                                MainSettings.AoELynxRush && Me.GotAlivePet && CM.PredicFocus() < 40 &&
                                                DpsMeter.GetCombatTimeLeft(Target, 3).TotalSeconds > 15),
                                     new Decorator(
                                         ret =>
                                         CM.PredicFocus() < 20 ||
                                         (!Me.HasAura("Thrill of the Hunt") && CM.PredicFocus() < 40),
                                         CM.Cast("Steady Shot",
                                                    ret =>
                                                    !TalentManager.HasTalent(10) || Helpers.SpellOnCooldown("Fervor")))
                                     ));
        }

        #endregion

        #region Ext :

        public Composite CreateCombat
        {
            get { return CombatSelector; }
        }

        public Composite CreatePrecombat
        {
            get { return PreCombatSelector; }
        }

        public static Composite HoldOn(WoWUnit u, bool enemy, bool friendly)
        {
            return new Decorator(ret => !Unit.ValidUnit(u, enemy, friendly) || Lists.ShouldHold(u),
                                 new ActionAlwaysSucceed());
        }

        public static Composite AutoPause()
        {
            return new Decorator(ret => Me.HasAnyCachedAura(new[] {"Drink", "Food", "First Aid", "Feign Death"}),
                                 new ActionAlwaysSucceed());
        }

        private static WoWUnit MDTarget()
        {
            switch (MainSettings.MisdirectOn)
            {
                case MisdirectOn.Me:
                    return StyxWoW.Me;
                case MisdirectOn.Focus:
                    return Unit.ValidFocus(false, true) ? Focus : null;
                case MisdirectOn.Pet:
                    return Me.GotAlivePet ? Pet : null;
                case MisdirectOn.Nobody:
                    return null;
            }
            return null;
        }

        #endregion

        #region Tidy : Bools

        private static bool NeedExhilaration()
        {
            if (Target == null)
                return false;
            return Me.HealthPercent < 68 ||
                   (Me.GotAlivePet && Pet.HealthPercent < 15 && PetManager.CanCastPetAction("Heart of the Phoenix") &&
                    Helpers.GetSpellCooldown("Heart of the Phoenix").TotalSeconds > 10) ||
                   (Me.GotAlivePet && Pet.HealthPercent < 15 && !PetManager.CanCastPetAction("Heart of the Phoenix"));
        }

        private static bool ShouldFeignDeath()
        {
            switch (MainSettings.FeignDeath)
            {
                case FeignDeath.Aggro_Health:
                    return Target.ThreatInfo.RawPercent > 90 && Target.CurrentTargetGuid == Me.Guid &&
                           (Target.IsCasting || Target.Distance < 10) && Me.HealthPercent < MainSettings.FeignHealth;
                case FeignDeath.HighThreat:
                    return Target.ThreatInfo.RawPercent > 90;
                case FeignDeath.LowHealth:
                    return Target.CurrentTargetGuid == Me.Guid && Me.HealthPercent < MainSettings.FeignHealth;
                case FeignDeath.OnAggro:
                    return Target.ThreatInfo.RawPercent > 90 && Target.CurrentTargetGuid == Me.Guid &&
                           (Target.IsCasting || Target.Distance < 10);
                case FeignDeath.Threat_Health:
                    return Target.ThreatInfo.RawPercent > 90 && Me.HealthPercent < MainSettings.FeignHealth;
                case FeignDeath.Never:
                    return false;
            }
            return false;
        }

        private static bool NeedReadinessPvESV()
        {
            if (Target == null)
                return false;
            return Helpers.GetSpellCooldown("Rapid Fire").TotalSeconds > 20 &&
                   (Helpers.GetSpellCooldown("Black Arrow").TotalSeconds > 5 || !MainSettings.BlackArrow) &&
                   (Helpers.GetSpellCooldown("Explosive Shot").TotalSeconds > 2 || !MainSettings.ExplosiveShot) &&
                   ReadinessExtensionPvE();
        }

        private static bool ReadinessExtensionPvE()
        {
            return (TalentManager.HasTalent(15) && Helpers.GetSpellCooldown("Lynx Rush").TotalSeconds > 9 ||
                    !TalentManager.HasTalent(15))
                   &&
                   (TalentManager.HasTalent(16) && Helpers.GetSpellCooldown("Glaive Toss").TotalSeconds > 2 ||
                    !TalentManager.HasTalent(16))
                   &&
                   (TalentManager.HasTalent(17) && Helpers.GetSpellCooldown("Powershot").TotalSeconds > 5 ||
                    !TalentManager.HasTalent(17))
                   &&
                   (TalentManager.HasTalent(18) && Helpers.GetSpellCooldown("Barrage").TotalSeconds > 5 ||
                    !TalentManager.HasTalent(18))
                   &&
                   (TalentManager.HasTalent(10) && Helpers.GetSpellCooldown("Fervor").TotalSeconds > 9 ||
                    !TalentManager.HasTalent(10))
                   &&
                   (TalentManager.HasTalent(13) && Helpers.GetSpellCooldown("A Murder of Crows").TotalSeconds > 9 ||
                    !TalentManager.HasTalent(13))
                   &&
                   (TalentManager.HasTalent(14) && Helpers.GetSpellCooldown("Blink Strike").TotalSeconds > 2 ||
                    !TalentManager.HasTalent(14))
                   &&
                   (TalentManager.HasTalent(11) && Helpers.GetSpellCooldown("Dire Beast").TotalSeconds > 4 ||
                    !TalentManager.HasTalent(11));
        }

        private static bool ShouldMisdirect()
        {
            if (Target == null)
                return false;

            switch (MainSettings.MisdirectWhen)
            {
                case MisdirectWhen.Never:
                    return false;
                case MisdirectWhen.InsecurelyTanking:
                    var tankable =
                        Helpers.NearbyUnfriendlyUnitsInCombatWithMe.FirstOrDefault(
                            u => u.Location.Distance2DSqr(Target.Location) < 25*25);
                    return (tankable != null && tankable.ThreatInfo.RawPercent < 90) ||
                           MDTarget().GetThreatInfoFor(Target).RawPercent < 90;
                case MisdirectWhen.Always:
                    return true;
            }
            return false;
        }

        private static bool ShouldScatterShot()
        {
            if (Target == null)
                return false;

            switch (MainSettings.ScatterShot)
            {
                case ScatterShot.Defend:
                    return Target.Distance <= 20 && Target.CurrentTargetGuid == Me.Guid;
                case ScatterShot.Interrupt:
                    return Target.Distance <= 20 &&
                           (Lists.InterruptCastNoChannel(Target) > 200 || Lists.InterruptCastChannel(Target) > 200);
                case ScatterShot.Interrupt_Defend:
                    return Target.Distance <= 20 &&
                           (Target.CurrentTargetGuid == Me.Guid ||
                            (Lists.InterruptCastNoChannel(Target) > 200 || Lists.InterruptCastChannel(Target) > 200));
                case ScatterShot.Never:
                    return false;
            }
            return false;
        }

        private static bool ShouldSerpentStingPvE()
        {
            if (Target == null)
                return false;


            switch (MainSettings.SerpentSting)
            {
                case SerpentSting.Always:
                    return true;
                case SerpentSting.Sometimes:
                    return Lists.TargetIsBoss(Target) || DpsMeter.GetCombatTimeLeft(Target, 2).TotalSeconds > 12;
                case SerpentSting.Never:
                    return false;
            }
            return false;
        }

        private static bool ShouldCobraShotPvESV()
        {
            if (Target == null)
                return false;
            return (!TalentManager.HasTalent(10) ||
                    (Helpers.GetSpellCooldown("Fervor").TotalMilliseconds > 750 &&
                     Helpers.GetSpellCooldown("Fervor").TotalSeconds < 29));
        }

        #endregion

        #region Tidy: Calls

        private static WoWUnit Pet
        {
            get { return StyxWoW.Me.Pet; }
        }

        private static WoWUnit Focus
        {
            get { return StyxWoW.Me.FocusedUnit; }
        }

        private static WoWUnit Target
        {
            get { return StyxWoW.Me.CurrentTarget; }
        }

        private static LocalPlayer Me
        {
            get { return StyxWoW.Me; }
        }

        private static MainSettingsSV MainSettings
        {
            get { return MainSettingsSV.Instance; }
        }

        private static CooldownSettingsSV CooldownSettings
        {
            get { return CooldownSettingsSV.Instance; }
        }

        private static PetSettingsSV PetSettings
        {
            get { return PetSettingsSV.Instance; }
        }

        #endregion

        #region Tidy : PetMacro's

        /* private Composite SetPetActions
        {
            get
            {
                return new Decorator(ret => MarkI.Main.DisablePetActions,
                                     new Action(delegate
                                     {
                                         if (MainSettings.GrowlPets)
                                         {
                                             Lua.DoString("RunMacroText('/petautocastoff Growl');");
                                         }
                                         if (PetSettings.SpiritMend != SpiritMend.Never)
                                         {
                                             Lua.DoString("RunMacroText('/petautocastoff Spirit Mend');");
                                         }
                                         if (PetSettings.RoarOfSacrifice != RoarOfSacrifice.Never)
                                         {
                                             Lua.DoString("RunMacroText('/petautocastoff Roar of Sacrifice');");
                                         }
                                         return RunStatus.Failure;
                                     }));
            }
        }
        */

        #endregion
    }
}