﻿using System.ComponentModel;
using Bestia_Magister.Misc;
using JetBrains.Annotations;
using Styx;
using Styx.Common;
using Styx.Helpers;
using System.Windows.Forms;

namespace Bestia_Magister.Settings
{
    internal class MainSettings : Styx.Helpers.Settings
    {
        private static MainSettings _instance;
        public static MainSettings Instance
        {
            get { return _instance ?? (_instance = new MainSettings()); }
        }

        public MainSettings()
            : base(SettingsPath + ".config")
        {
        }

        private static string SettingsPath
        {
            get
            {
                return string.Format("{0}\\Settings\\MarkI\\MarkISettings_Main_{1}", Utilities.AssemblyDirectory,
                                     StyxWoW.Me.Name);
            }
        }

/*
        private static bool IsTrinketUsageWanted(TrinketUsage usage)
        {
            return usage == Instance.Trinket1Usage
                || usage == Instance.Trinket2Usage;
        }
*/

        #region Category: Hotkeys

        /// <summary>
        ///     Contains all Hotkey Settings
        /// </summary>
        public class Hotkeys : Styx.Helpers.Settings
        {
            [CanBeNull] private static Hotkeys _instance;

            public Hotkeys()
                : base(Helpers.SettingsPath("Hotkeys") + ".config")
            {
            }

            [CanBeNull]
            public static Hotkeys Instance
            {
                get { return _instance ?? (_instance = new Hotkeys()); }
            }

            [Setting]
            [Styx.Helpers.DefaultValue(Keys.E)]
            public Keys FocusKeyChoice { get; set; }

            [Setting]
            [Styx.Helpers.DefaultValue(Keys.D1)]
            public Keys CCModeKeyChoice { get; set; }

            [Setting]
            [Styx.Helpers.DefaultValue(Keys.D1)]
            public Keys ForceTrapChoice { get; set; }

            [Setting]
            [Styx.Helpers.DefaultValue(Keys.F)]
            public Keys AoEKeyChoice { get; set; }

            [Setting]
            [Styx.Helpers.DefaultValue(Keys.LShiftKey)]
            public Keys CCModifierChoice { get; set; }

            [Setting]
            [Styx.Helpers.DefaultValue(ModifierKeys.Alt)]
            public ModifierKeys ModifierChoice { get; set; }

            [Setting]
            [Styx.Helpers.DefaultValue(ModifierKeys.Shift)]
            public ModifierKeys FocusModifierChoice { get; set; }
        }

        #endregion

        #region Category: Talents

        [Setting]
        [Styx.Helpers.DefaultValue(false)]
        [Category("Talents")]
        [DisplayName("Silencing Shot")]
        [Description("Use Silencing Shot")]
        public bool Talent30_SS { get; set; }

        [Setting]
        [Styx.Helpers.DefaultValue(false)]
        [Category("Talents")]
        [DisplayName("Wyvern Sting")]
        [Description("Use Wyvern Sting")]
        public bool Talent30_WS { get; set; }

        [Setting]
        [Styx.Helpers.DefaultValue(false)]
        [Category("Talents")]
        [DisplayName("Binding Shot")]
        [Description("Use Binding Shot")]
        public bool Talent30_BS { get; set; }

        [Setting]
        [Styx.Helpers.DefaultValue(true)]
        [Category("Talents")]
        [DisplayName("Exhilaration")]
        [Description("Use Exhilaration")]
        public bool Talent45 { get; set; }

        [Setting]
        [Styx.Helpers.DefaultValue(true)]
        [Category("Talents")]
        [DisplayName("Fervor")]
        [Description("Use Fervor")]
        public bool Talent60_FV { get; set; }

        [Setting]
        [Styx.Helpers.DefaultValue(true)]
        [Category("Talents")]
        [DisplayName("Dire Beast")]
        [Description("Use Dire Beast")]
        public bool Talent60_DB { get; set; }

        [Setting]
        [Styx.Helpers.DefaultValue(true)]
        [Category("Talents")]
        [DisplayName("A Murder of Crows")]
        [Description("Use A Murder of Crows")]
        public bool Talent75_AMOC { get; set; }

        [Setting]
        [Styx.Helpers.DefaultValue(true)]
        [Category("Talents")]
        [DisplayName("Blink Strike")]
        [Description("Use Blink Strike")]
        public bool Talent75_BS { get; set; }

        [Setting]
        [Styx.Helpers.DefaultValue(true)]
        [Category("Talents")]
        [DisplayName("Lynx Rush")]
        [Description("Use Lynx Rush")]
        public bool Talent75_LXR { get; set; }

        [Setting]
        [Styx.Helpers.DefaultValue(true)]
        [Category("Talents")]
        [DisplayName("Glaive Toss")]
        [Description("Use Glaive Toss")]
        public bool Talent90_GT { get; set; }

        [Setting]
        [Styx.Helpers.DefaultValue(true)]
        [Category("Talents")]
        [DisplayName("Powershot")]
        [Description("Use Powershot")]
        public bool Talent90_PWRS { get; set; }

        [Setting]
        [Styx.Helpers.DefaultValue(true)]
        [Category("Talents")]
        [DisplayName("Barrage")]
        [Description("Use Barrage")]
        public bool Talent90_BRG { get; set; }
        #endregion

        #region Category: Items

        //Trinket Pop Health %
        [Setting]
        [Styx.Helpers.DefaultValue(30)]
        [Category("Items")]
        [DisplayName("Health Trinket")]
        [Description("Less than % Health to use a Health Trinket")]
        public int TrinketHealth { get; set; }

        //Trinket 1
        [Setting]
        [Styx.Helpers.DefaultValue(TrinketUsage.Never)]
        [Category("Items")]
        [DisplayName("Trinket 1")]
        [Description("When to use Trinket 1")]
        public TrinketUsage Trinket1Usage { get; set; }

        //Trinket 2
        [Setting]
        [Styx.Helpers.DefaultValue(TrinketUsage.Never)]
        [Category("Items")]
        [DisplayName("Trinket 2")]
        [Description("When to use Trinket 2")]
        public TrinketUsage Trinket2Usage { get; set; }

        [Setting]
        [Styx.Helpers.DefaultValue(5400)]
        [Category("Items")]
        [DisplayName("CC Trinket Time")]
        [Description("More than X time left (ms) on a Stun / CC to use Trinket")]
        public int CCTrinketTime { get; set; }

        // Healthstone - Health
        [Setting]
        [Styx.Helpers.DefaultValue(40)]
        [Category("Items")]
        [DisplayName("Healthstone")]
        [Description("Less than % health for using Healthstone")]
        public int SHealthstone { get; set; }

        //Health Potion
        [Setting]
        [Styx.Helpers.DefaultValue(30)]
        [Category("Items")]
        [DisplayName("Potion Health")]
        [Description("Less than % Health to use a Health potion")]
        public int PotionHealth { get; set; }

        //Health Potion
        [Setting]
        [Styx.Helpers.DefaultValue(95)]
        [Category("Items")]
        [DisplayName("Virmen Health")]
        [Description("Less than % Health to use a Virmen potion")]
        public int VirmenHealth { get; set; }

        //Health Potion
        [Setting]
        [Styx.Helpers.DefaultValue(false)]
        [Category("Items")]
        [DisplayName("Virmen Pot")]
        [Description("Enable/Disable use of this pot")]
        public bool UseVirminPot { get; set; }

        // Battlestandard - Health
        [Setting]
        [Styx.Helpers.DefaultValue(40)]
        [Category("Items")]
        [DisplayName("Battlestandard")]
        [Description("My HP Less than % health to use Horde/Alliance Battle Standard")]
        public int SbStandard { get; set; }
        #endregion

        #region Main

        [Setting]
        [Styx.Helpers.DefaultValue(65)]
        [Category("Main")]
        [DisplayName("Focus Amount")]
        [Description("Less than x Focus to cast Steady/Cobra Shot")]
        public int FocusAmount { get; set; }

        [Setting]
        [Styx.Helpers.DefaultValue(false)]
        [Category("Main")]
        [DisplayName("Hunter's Mark")]
        [Description("Use Hunter's Mark")]
        public bool HuntersMark { get; set; }

        [Setting]
        [Styx.Helpers.DefaultValue(true)]
        [Category("Main")]
        [DisplayName("Aspect of the Hawk")]
        [Description("Force Aspect of the Hawk in combat")]
        public bool AspectHawk { get; set; }

        [Setting]
        [Styx.Helpers.DefaultValue(true)]
        [Category("Main")]
        [DisplayName("Master's Call")]
        [Description("Use Master's Call When Snared / Rooted in PvE")]
        public bool MastersCallPvE { get; set; }

        [Setting]
        [Styx.Helpers.DefaultValue(FeignDeath.Never)]
        [Category("Main")]
        [DisplayName("Feign Death")]
        [Description("When to use Feign Death.")]
        public FeignDeath FeignDeath { get; set; }

        [Setting]
        [Styx.Helpers.DefaultValue(20)]
        [Category("Main")]
        [DisplayName("Feign Health")]
        [Description("Less than % Health to use Feign Death (PvE)")]
        public int FeignHealth { get; set; }

        [Setting]
        [Styx.Helpers.DefaultValue(Intimidate.Never)]
        [Category("Main")]
        [DisplayName("Intimidation")]
        [Description("When to use Intimidation.")]
        public Intimidate Intimidation { get; set; }

        [Setting]
        [Styx.Helpers.DefaultValue(40)]
        [Category("Main")]
        [DisplayName("Intimidation Health")]
        [Description("Less than % Health on target to use Intimidation")]
        public int IntimidationHealth { get; set; }


        [Setting]
        [Styx.Helpers.DefaultValue(MisdirectOn.Nobody)]
        [Category("Main")]
        [DisplayName("Misdirect On")]
        [Description("Who to use Misdirect on.")]
        public MisdirectOn MisdirectOn { get; set; }

        [Setting]
        [Styx.Helpers.DefaultValue(MisdirectWhen.Never)]
        [Category("Main")]
        [DisplayName("Misdirect")]
        [Description("When to use Misdirect.")]
        public MisdirectWhen MisdirectWhen { get; set; }

        [Setting]
        [Styx.Helpers.DefaultValue(true)]
        [Category("Main")]
        [DisplayName("TreePerformance")]
        [Description("Enable to measure Tree Performance")]
        public bool TreePerformance { get; set; }

        [Setting]
        [Styx.Helpers.DefaultValue(100)]
        [Category("Main")]
        [DisplayName("Time left on GCD to Cast")]
        [Description("Less than x ms left on Global Cooldown to allow casting")]
        public int GCDTimeLeft { get; set; }

        [Setting]
        [Styx.Helpers.DefaultValue(65)]
        [Category("Main")]
        [DisplayName("Time left on GCD to Traverse")]
        [Description("Less than x ms left on Global Cooldown to traverse through the tree")]
        public int GCDTraverse { get; set; }
        #endregion

        #region PvP

        [Setting]
        [Styx.Helpers.DefaultValue(false)]
        [Category("PvP")]
        [DisplayName("Force PvP")]
        [Description("Forces PvP rotation")]
        public bool ForcePvP { get; set; }

        [Setting]
        [Styx.Helpers.DefaultValue(true)]
        [Category("PvP")]
        [DisplayName("Always PvP against Players")]
        [Description("Forces PvP rotation when targetting an attackable player")]
        public bool PlayerPvP { get; set; }

        [Setting]
        [Styx.Helpers.DefaultValue(false)]
        [Category("PvP")]
        [DisplayName("Auto Disable Pet Actions")]
        [Description("Automatically disables pet actions you've chosen to let the CR use.")]
        public static bool DisablePetActions { get; set; }

        [Setting]
        [Styx.Helpers.DefaultValue(false)]
        [Category("PvP")]
        [DisplayName("Growl Pets")]
        [Description("Use Growl on Nearby pets in Arenas")]
        public bool GrowlPets { get; set; }

        [Setting]
        [Styx.Helpers.DefaultValue(true)]
        [Category("PvP")]
        [DisplayName("Kill Totems")]
        [Description("Kill Shaman Totems in Arenas, Communists must die")]
        public bool KillTotems { get; set; }

        [Setting]
        [Styx.Helpers.DefaultValue(PvPFeignDeath.Never)]
        [Category("PvP")]
        [DisplayName("Feign Death in PvP")]
        [Description("When to use Feign Death in PvP.")]
        public PvPFeignDeath PvPFeignDeath { get; set; }

        [Setting]
        [Styx.Helpers.DefaultValue(80)]
        [Category("PvP")]
        [DisplayName("Feign Health")]
        [Description("Less than % Health to use Feign Death (PvP)")]
        public int FeignHealthPvP { get; set; }

        [Setting]
        [Styx.Helpers.DefaultValue(MastersCall.Never)]
        [Category("PvP")]
        [DisplayName("Master's Call")]
        [Description("Who to use Master's Call on, in PvP.")]
        public MastersCall MastersCall { get; set; }

        [Setting]
        [Styx.Helpers.DefaultValue(true)]
        [Category("PvP")]
        [DisplayName("Deterrence Reflect")]
        [Description("Use Deterrence to reflect spells when glyphed.")]
        public bool DeterrenceReflect { get; set; }

        [Setting]
        [Styx.Helpers.DefaultValue(SilencePvP.Never)]
        [Category("PvP")]
        [DisplayName("Silencing Shot PvP")]
        [Description("When to use Silencing Shot in PvP.")]
        public SilencePvP SilencePvP { get; set; }

        [Setting]
        [Styx.Helpers.DefaultValue(50)]
        [Category("PvP")]
        [DisplayName("Silencing Shot Health")]
        [Description("Less than % health on Target to use Silencing Shot")]
        public int SilenceHealth { get; set; }

        [Setting]
        [Styx.Helpers.DefaultValue(true)]
        [Category("PvP")]
        [DisplayName("Tranquilizing Shot")]
        [Description("Use Tranquilizing Shot")]
        public bool Tranquilizing { get; set; }

        [Setting]
        [Styx.Helpers.DefaultValue(true)]
        [Category("PvP")]
        [DisplayName("Widow Venom")]
        [Description("Use Widow Venom")]
        public bool WidowVenom { get; set; }

        [Setting]
        [Styx.Helpers.DefaultValue(true)]
        [Category("PvP")]
        [DisplayName("Hunter's Mark > 40")]
        [Description("Use Hunter's Mark out of combat on targets that are over 40 yds away")]
        public static bool HuntersMarkOOC { get; set; }
        #endregion

        #region Attacks

        [Setting]
        [Styx.Helpers.DefaultValue(true)]
        [Category("Attacks")]
        [DisplayName("Kill Command")]
        [Description("Use Kill Command")]
        public bool KillCommand { get; set; }

        [Setting]
        [Styx.Helpers.DefaultValue(true)]
        [Category("Attacks")]
        [DisplayName("Kill Shot")]
        [Description("Use Kill Shot")]
        public bool KillShot { get; set; }

        [Setting]
        [Styx.Helpers.DefaultValue(true)]
        [Category("Attacks")]
        [DisplayName("Arcane Shot")]
        [Description("Use Arcane Shot")]
        public bool ArcaneShot { get; set; }

        [Setting]
        [Styx.Helpers.DefaultValue(SerpentSting.Sometimes)]
        [Category("Attacks")]
        [DisplayName("Serpent Sting")]
        [Description("When to use Serpent Sting.")]
        public SerpentSting SerpentSting { get; set; }

        [Setting]
        [Styx.Helpers.DefaultValue(ScatterShot.Never)]
        [Category("Attacks")]
        [DisplayName("Scatter Shot")]
        [Description("When to use Scatter Shot.")]
        public ScatterShot ScatterShot { get; set; }

        [Setting]
        [Styx.Helpers.DefaultValue(true)]
        [Category("Attacks")]
        [DisplayName("Concussive Shot")]
        [Description("Use Concussive Shot")]
        public bool ConcussiveShot { get; set; }


        [Setting]
        [Styx.Helpers.DefaultValue(true)]
        [Category("Attacks")]
        [DisplayName("Distracting Shot")]
        [Description("Use Distracting Shot")]
        public bool DistractingShot { get; set; }

        #endregion

        #region Traps

        [Setting]
        [Styx.Helpers.DefaultValue(false)]
        [Category("Traps")]
        [DisplayName("Ice Trap Launcher")]
        [Description("Use Ice Trap Launcher")]
        public bool IceTrapLauncher { get; set; }

        [Setting]
        [Styx.Helpers.DefaultValue(false)]
        [Category("Traps")]
        [DisplayName("Snake Trap Launcher")]
        [Description("Use Snake Trap Launcher")]
        public bool SnakeTrapLauncher { get; set; }

        [Setting]
        [Styx.Helpers.DefaultValue(false)]
        [Category("Traps")]
        [DisplayName("Freezing Trap Launcher")]
        [Description("Use Freezing Trap Launcher")]
        public bool FreezingTrapLauncher { get; set; }

        [Setting]
        [Styx.Helpers.DefaultValue(true)]
        [Category("Traps")]
        [DisplayName("Explosive Trap Launcher")]
        [Description("Use Explosive Trap Launcher")]
        public bool ExplosiveTrapLauncher { get; set; }

        [Setting]
        [Styx.Helpers.DefaultValue(true)]
        [Category("Traps")]
        [DisplayName("ST EXPL Trap Launcher")]
        [Description("Use Single Target Explosive Trap Launcher")]
        public bool STEXPLTrapLauncher { get; set; }

        [Setting]
        [Styx.Helpers.DefaultValue(false)]
        [Category("Traps")]
        [DisplayName("Ice Trap")]
        [Description("Use Ice Trap")]
        public bool IceTrap { get; set; }

        [Setting]
        [Styx.Helpers.DefaultValue(false)]
        [Category("Traps")]
        [DisplayName("Snake Trap")]
        [Description("Use Snake Trap")]
        public bool SnakeTrap { get; set; }

        [Setting]
        [Styx.Helpers.DefaultValue(false)]
        [Category("Traps")]
        [DisplayName("Freezing Trap")]
        [Description("Use Freezing Trap")]
        public bool FreezingTrap { get; set; }

        [Setting]
        [Styx.Helpers.DefaultValue(true)]
        [Category("Traps")]
        [DisplayName("Explosive Trap")]
        [Description("Use Explosive Trap")]
        public bool ExplosiveTrap { get; set; }

        #endregion

        #region Focused Unit

        [Setting]
        [Styx.Helpers.DefaultValue(true)]
        [Category("Focus Unit")]
        [DisplayName("Freezing Trap")]
        [Description("Crowd Control Focused Unit using Freezing Trap Launcher")]
        public bool CCFreezingTrap { get; set; }

        [Setting]
        [Styx.Helpers.DefaultValue(true)]
        [Category("Focus Unit")]
        [DisplayName("Scatter Shot")]
        [Description("Use Scatter Shot on Focused Unit")]
        public bool FocusScatter { get; set; }

        [Setting]
        [Styx.Helpers.DefaultValue(true)]
        [Category("Focus Unit")]
        [DisplayName("Binding Shot")]
        [Description("Use Binding Shot on Focused Unit")]
        public bool FocusBinding { get; set; }

        [Setting]
        [Styx.Helpers.DefaultValue(true)]
        [Category("Focus Unit")]
        [DisplayName("Wyvern Sting")]
        [Description("Crowd Control Focused Unit using Wyvern Sting")]
        public bool CCWyvernSting { get; set; }

        [Setting]
        [Styx.Helpers.DefaultValue(true)]
        [Category("Focus Unit")]
        [DisplayName("Silencing Shot")]
        [Description("Use Silencing Shot on Focused Unit")]
        public bool FocusSilencingShot { get; set; }

        [Setting]
        [Styx.Helpers.DefaultValue(true)]
        [Category("Focus Unit")]
        [DisplayName("Concussive Shot")]
        [Description("Use Concussive Shot on Focused Unit")]
        public bool FocusConcussive { get; set; }

        [Setting]
        [Styx.Helpers.DefaultValue(true)]
        [Category("Focus Unit")]
        [DisplayName("Tranquilizing Shot")]
        [Description("Use Tranquilizing Shot on Focused Unit")]
        public bool FocusTranq { get; set; }

        #endregion

        #region AoE

        [Setting]
        [Styx.Helpers.DefaultValue(3)]
        [Category("AoE")]
        [DisplayName("AoE Add Count")]
        [Description("Units needed to perform AoE Rotation")]
        public static int AddCount { get; set; }

        [Setting]
        [Styx.Helpers.DefaultValue(4)]
        [Category("AoE")]
        [DisplayName("Binding Shot Add Count")]
        [Description("Units near target to use Binding Shot")]
        public int BindingCount { get; set; }

        [Setting]
        [Styx.Helpers.DefaultValue(true)]
        [Category("AoE")]
        [DisplayName("Multi Shot")]
        [Description("Use Multi Shot")]
        public bool MultiShot { get; set; }

        [Setting]
        [Styx.Helpers.DefaultValue(true)]
        [Category("AoE")]
        [DisplayName("AoE Dire Beast")]
        [Description("Use Dire Beast in AoE")]
        public bool AoEDireBeast { get; set; }

        [Setting]
        [Styx.Helpers.DefaultValue(true)]
        [Category("AoE")]
        [DisplayName("AoE Explosive Trap")]
        [Description("Use Explosive Trap in AoE")]
        public bool AoEExpTrap { get; set; }

        [Setting]
        [Styx.Helpers.DefaultValue(true)]
        [Category("AoE")]
        [DisplayName("AoE Lynx Rush")]
        [Description("Use Lynx Rush in AoE")]
        public bool AoELynxRush { get; set; }

        #endregion

    }
}
